#############################################################################
# pgceditor.tcl Copyright 2004, 2005 r0lZ
# The PGC Header editor stuff.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::pgced {

namespace export edit_pgc dump_pgc

set play {
	#define play_bitmap_width 8
	#define play_bitmap_height 13
	static unsigned char play_bitmap_bits[] = {
	   0x01,
	   0x03,
	   0x07,
	   0x0f,
	   0x1f,
	   0x3f,
	   0x7f,
	   0x3f,
	   0x1f,
	   0x0f,
	   0x07,
	   0x03,
	   0x01};
}
image create bitmap ::preview::bitmapplay -background {} -data $play
unset play


proc BCDtime {bcd showfps} {
	set pbtimetxt           [format  %i [expr { [get_int bcd 0 1] / 16}]]
	set pbtimetxt $pbtimetxt[format  %i [expr { [get_int bcd 0 1] & 15}]]
	set pbtimetxt $pbtimetxt:[format %i [expr { [get_int bcd 1 1] / 16}]]
	set pbtimetxt $pbtimetxt[format  %i [expr { [get_int bcd 1 1] & 15}]]
	set pbtimetxt $pbtimetxt:[format %i [expr { [get_int bcd 2 1] / 16}]]
	set pbtimetxt $pbtimetxt[format  %i [expr { [get_int bcd 2 1] & 15}]]
	set pbtimetxt $pbtimetxt.[format %i [expr {([get_int bcd 3 1]& 0X30)/16}]]
	set pbtimetxt $pbtimetxt[format  %i [expr { [get_int bcd 3 1] & 15}]]
	if {$showfps} {
		set fpsmode [expr {[get_int bcd 3 1] & 0XC0}]
		switch $fpsmode {
			192 {set pbtimetxt "$pbtimetxt (at 30 fps)"}
			64  {set pbtimetxt "$pbtimetxt (at 25 fps)"}
			default {set pbtimetxt "$pbtimetxt (illegal FPS)"}
		}
	}
	return $pbtimetxt
}

proc short_BCDtime {bcd} {
	set t1 [format %i%i [expr { [get_int bcd 0 1] / 16}] [expr { [get_int bcd 0 1] & 15}]]:
	set t2 [format %i%i [expr { [get_int bcd 1 1] / 16}] [expr { [get_int bcd 1 1] & 15}]]:
	set t3 [format %i%i [expr { [get_int bcd 2 1] / 16}] [expr { [get_int bcd 2 1] & 15}]].
	set t4 [format %i%i [expr {([get_int bcd 3 1]& 0X30)/16}] [expr { [get_int bcd 3 1] & 15}]]

	if {$t1 == "00:"} {
		set t1 ""
		if {[string range $t2 0 0] == "0"} {
			set t2 [string range $t2 1 end]
		}
	} else {
		set t1 [string trimleft $t1 "0"]
	}

	return $t1$t2$t3$t4
}


proc AddBCDtime {bcd1 bcd2} {
	set fpsmode [expr {[get_int bcd1 3 1] & 0XC0}]
	switch $fpsmode {
		192 {set fps 30}
		64  {set fps 25}
		default {set fps 0}
	}
	if {$fps == 0} {return "\000\000\000\000"}

	set f1 [expr {((([get_int bcd1 3 1]& 0X30)/16)*10) + ([get_int bcd1 3 1] & 15)}]
	set s1 [expr { (([get_int bcd1 2 1] / 16) * 10) + ([get_int bcd1 2 1] & 15)}]
	set m1 [expr { (([get_int bcd1 1 1] / 16) * 10) + ([get_int bcd1 1 1] & 15)}]
	set h1 [expr { (([get_int bcd1 0 1] / 16) * 10) + ([get_int bcd1 0 1] & 15)}]

	set f2 [expr {((([get_int bcd2 3 1]& 0X30)/16)*10) + ([get_int bcd2 3 1] & 15)}]
	set s2 [expr { (([get_int bcd2 2 1] / 16) * 10) + ([get_int bcd2 2 1] & 15)}]
	set m2 [expr { (([get_int bcd2 1 1] / 16) * 10) + ([get_int bcd2 1 1] & 15)}]
	set h2 [expr { (([get_int bcd2 0 1] / 16) * 10) + ([get_int bcd2 0 1] & 15)}]

	set f [expr {$f1 + $f2}]
	set s [expr {$s1 + $s2}]
	set m [expr {$m1 + $m2}]
	set h [expr {$h1 + $h2}]

	while {$f >= $fps} {
		set f [expr {$f - $fps}]
		incr s
	}
	while {$s >= 60} {
		set s [expr {$s - 60}]
		incr m
	}
	while {$m >= 60} {
		set m [expr {$m - 60}]
		incr h
	}

	set f "[expr {((($f / 10) * 16) + ($f % 10)) | $fpsmode}]"
	set s "[expr {(($s / 10) * 16) + ($s % 10)}]"
	set m "[expr {(($m / 10) * 16) + ($m % 10)}]"
	set h "[expr {(($h / 10) * 16) + ($h % 10)}]"

	return "[format %c $h][format %c $m][format %c $s][format %c $f]"
}

# dump PGC contents as text
proc dump_pgc {pgctype index pgcheader numprograms programmap numcells cellpbtable cellpostable PTTlist} {

	set vts   [lindex [split $index ","] 0]

	# get editables values from the header's binary data
	set txt " Playback time:  [BCDtime [string range $pgcheader 4 7] true]\n"

	set ::pgced::pbmode    [get_int pgcheader 0XA2 1]
	if {$::pgced::pbmode == 0} {
		set pbmodetxt "sequential"
	} else {
		set ::pgced::pbprogs [expr {$::pgced::pbmode & 0X7F}]
		if {[expr {$::pgced::pbmode & 0X80}] == 0} {
			set pbmodetxt "random: $::pgced::pbprogs programs"
		} else {
			set pbmodetxt "shuffle: $::pgced::pbprogs programs"
		}
	}
	set txt "${txt} PG Playback mode:  $pbmodetxt\n"

	set puos      [get_int pgcheader 8 4]
	set txt "${txt} PUOs:           $puos (0x[format %08X $puos])\n"
	for {set p 0} {$p <= 24} {incr p} {
		set state [expr {$puos & round(pow(2,$p))}]
		if {$state > 0} {
			set txt "${txt}                 - [lindex $::pgced::puos_labels $p] (0x[format %08X [expr {round(pow(2,$p))}]])\n"
		}
	}

	set txt "${txt} NextPGCN:       [get_int pgcheader 0X9C 2]\n"
	set txt "${txt} PrevPGCN:       [get_int pgcheader 0X9E 2]\n"
	set txt "${txt} GoUpPGCN:       [get_int pgcheader 0XA0 2]\n"
	set txt "${txt} PGC Still Time: [get_int pgcheader 0XA3 1]\n"

	for {set a 1} {$a <= 8} {incr a} {
		set audio_status [get_int pgcheader [expr {(0X0A + ($a*2))}] 2]
		if {$audio_status != 0} {
			set txt "${txt} Audio  stream  $a status: [format 0x%08X $audio_status] (stream=[expr {($audio_status & 0x0700)/0x100}])\n"
		}
	}
	for {set s 1} {$s <= 32} {incr s} {
		binary scan [string range $pgcheader [expr {(0X18 + ($s*4))}] [expr {(0X18 + ($s*4) + 3)}]] cccc s1 s2 s3 s4
		if {$s1 != 0} {
			set txt "${txt} Subpic stream [format %2d $s] status: [format 0x%08X [get_int pgcheader [expr {(0X18) + ($s*4)}] 4]] (streams for 4:3=[expr {$s1 & 0X1F}], wide=$s2, letterbox=$s3, pan&scan=$s4)\n"
		}
	}

	set txt "${txt}\n"
	set end_time "\000\000\000\000"
	if {$numcells > 0} {
		if {$pgctype == "menu"} {set txt "${txt} But- "} else {set txt "${txt} Chap."}
		set	txt "${txt} Prog. Cell  Type  Layer Res-  Still Cell  Playback     End          Entry    First    Last     Last     VOB Cell\n"
		if {$pgctype == "menu"} {set txt "${txt} tons "} else {set txt "${txt} (PTT)"}
		set	txt "${txt}             Flags Break tric- Time  Cmd.  Time         Time         VOBU     ILVU     VOBU     VOBU     ID  ID\n"
		set	txt "${txt}                               ted.                                        sector   End      Start    End\n\n"

		for {set c 1} {$c <= $numcells} {incr c} {
			set cell_program($c) ""
			set prog_chapter($c) ""
		}
		for {set p 1} {$p <= $numprograms} {incr p} {
			set cell_program([get_int programmap [expr {$p-1}] 1]) $p
		}
		for {set c 1} {$c <= $numcells} {incr c} {
			set prog_chapter($c) 0
		}
		foreach pttpair $PTTlist {
			set pg  [lindex $pttpair 0]
			set ptt [lindex $pttpair 1]
			for {set c 1} {$c <= $numcells} {incr c} {
				if {$pg == $cell_program($c) && $prog_chapter($c) == 0} {
					set prog_chapter($c) $ptt
				}
			}
		}

		for {set c 1} {$c <= $numcells} {incr c} {
			set cellidx1 [expr {($c - 1) * 24}]
			set cellidx2 [expr {($c - 1) * 4}]

			set vobid  [get_int cellpostable $cellidx2 2]
			set cellid [get_int cellpostable [expr {$cellidx2 + 3}] 1]

			if {$pgctype == "menu"} {
				if {[info exists ::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)]} {
					set txt "$txt [format %-5s $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)] "
				} else {
					set txt "$txt ?     "
				}
			} else {
				if {[info exists ::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts)]} {
					set txt "$txt $::buttons::butsinfo(pgc,$vts,$vobid,$cellid,numbuts) "
				}
				if {$prog_chapter($c) == 0} {
					set txt "$txt       "
				} else {
					set txt "$txt [format %-5s $prog_chapter($c)] "
				}
			}
			set txt "$txt[format %-5s $cell_program($c)] "
			set txt "$txt[format %-5d $c] "

			set type [get_int cellpbtable $cellidx1 1]
			set txt "$txt[format %-5d $type] "
			if {[expr {$type & 8}] == 8} {
				set txt "${txt}no    "
			} else {
				set txt "${txt}yes   "
			}

			set var [expr {[get_int cellpbtable [expr {$cellidx1 + 1}] 1] & 64}]
			if {$var == 0} {
				set txt "${txt}no    "
			} else {
				set txt "${txt}yes   "
			}

			set txt "$txt[format %-5d [get_int cellpbtable [expr {$cellidx1 + 2}] 1]] "

			set txt "$txt[format %-5d [get_int cellpbtable [expr {$cellidx1 + 3}] 1]] "

			set cell_time [string range $cellpbtable [expr {$cellidx1 + 4}] [expr {$cellidx1 + 7}]]
			set txt "$txt[BCDtime $cell_time false]  "

			# Add cell time to total PGC time, but only if angle 1.
			if {[expr {$type & 0x10}] == 0 || [expr {$type & 0xC0}] == 0x40} {
				set end_time [AddBCDtime $cell_time $end_time]
			}
			set txt "$txt[BCDtime $end_time false]  "

			set txt "$txt[format %-8d [get_int cellpbtable [expr {$cellidx1 +  8}]  4]] "
			set txt "$txt[format %-8d [get_int cellpbtable [expr {$cellidx1 +  12}] 4]] "
			set txt "$txt[format %-8d [get_int cellpbtable [expr {$cellidx1 +  16}] 4]] "
			set txt "$txt[format %-8d [get_int cellpbtable [expr {$cellidx1 +  20}] 4]] "

			set txt "$txt[format %-3d $vobid] "
			set txt "$txt$cellid\n"
		}
	}
	return $txt
}
proc dump_pgc_top {type vts args} {
	if {$type == "menu"} {
		set lu  [lindex $args 0]
		set pgc [lindex $args 1]
		set index "$vts,$lu,$pgc"
		set label [::main::pgc_label "" menu $vts $lu $pgc]
	} else {
		set pgc [lindex $args 0]
		set index "$vts,$pgc"
		set label [::main::pgc_label "" pgc $vts $pgc]
	}
	set pgcheader    [subst \$::${type}s($index,header)]
	set numprograms  [subst \$::${type}s($index,numprograms)]
	set programmap   [subst \$::${type}s($index,programmap)]
	set numcells     [subst \$::${type}s($index,numcells)]
	set cellpbtable  [subst \$::${type}s($index,cellpbtable)]
	set cellpostable [subst \$::${type}s($index,cellpostable)]
	if {$type == "pgc" && [info exists ::${type}s($index,pttlist)]} {
		set pttlist [subst \$::${type}s($index,pttlist)]
	} else {
		set pttlist ""
	}
	set vtsnum [lindex [split $index ","] 0]

	return " ************ $label ************\n\n[::pgced::dump_pgc $type \
			$index $pgcheader $numprograms $programmap \
			$numcells $cellpbtable $cellpostable $pttlist]"
}

proc dump_all_pgcs {} {
	if {! [info exists ::dvddir] || $::scratchbook_active} {
		bell
		return
	}
	show_msg "Dumping VMG..."

	set out "DVD: \"$::dvddir\"\n"
	set out "${out}DVD-TEXT General Name: \"[::utils::get_General_Name]\"\n"
	set out "${out}Provider ID: \"[::utils::get_provider_ID]\"\n"
	set out "${out}Number of VTS: $::pgcs(numvts)\n\n"

	# VIDEO_TS.IFO
	set out "${out}VMG (VIDEO_TS.IFO)\n"
	set out "${out}------------------\n\n"
	set out "${out}[dump_pgc_top pgc 0 0]\n\n"

	if {$::menus(0,numlus)==0} {
		set out "${out}VMGM: no menu.\n\n"
	} else {
		for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
			set out "${out}VMGM , LU $lu ($::menus(0,$lu,lang)):\n\n"
			for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
				set out "$out[dump_pgc_top menu 0 $lu $pgc]\n\n"
			}
		}
		set out "$out\n"
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		show_msg "Dumping VTS $vts..."
		set out "${out}VTS $vts (VTS_[format %02i $vts]_0.IFO)\n"
		set out "${out}--------------------\n\n"

		if {$::menus($vts,numlus)==0} {
			set out "${out}VTSM $vts: no menu.\n\n"
		} else {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				set out "${out}VTSM $vts , LU $lu ($::menus($vts,$lu,lang)):\n\n"
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set out "$out[dump_pgc_top menu $vts $lu $pgc]\n\n"
				}
			}
			set out "$out\n"
		}

		set out "${out}VTST $vts:\n\n"
		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			set out "$out[dump_pgc_top pgc $vts $pgc]\n\n"
		}
	}
	::main::showtext .dumpallpgcs "DVD PGCs Info" [string trimright $out \n]\n
	show_msg ""
}


#### Main entry for PgcEditor
proc edit_pgc {pgctype index pgcheader numprograms programmap numcells cellpbtable cellpostable PTTlist} {
	set w ".pgceditor"
	if {[winfo exists $w]} {return ""}

	show_msg "Building PGC editor window..."

	unset -nocomplain ::pgced::__warning_delete_chapters
	unset -nocomplain ::pgced::__warning_toggle_chapters

	set ::pgced::__pgcheader    $pgcheader
	set ::pgced::__programmap   $programmap
	set ::pgced::__numcells     $numcells
	set ::pgced::__numprograms  $numprograms
	set ::pgced::__cellpbtable  $cellpbtable
	set ::pgced::__cellpostable $cellpostable
	set ::pgced::__PTTlist      $PTTlist
	set ::pgced::__pgctype		$pgctype
	set ::pgced::__index		$index

	set type [lindex $::editor_cmds_from 0]
	set vts [lindex [split $::pgced::__index ","] 0]

	# get editables values from the header's binary data

	set ::pgced::__pgcpbtime [string range $::pgced::__pgcheader 4 7]
	set ::pgced::pbtimetxt [BCDtime $::pgced::__pgcpbtime true]

	set ::pgced::pbmode    [get_int ::pgced::__pgcheader 0XA2 1]
	if {$::pgced::pbmode > 0} {
		set ::pgced::pbprogs [expr {$::pgced::pbmode & 0X7F}]
		if {[expr {$::pgced::pbmode & 0X80}] == 0} {
			set ::pgced::pbmode 1	;# random
		} else {
			set ::pgced::pbmode 2	;# shuffle
		}
	} else {
		set ::pgced::pbprogs 1
	}

	set ::pgced::puos      [get_int ::pgced::__pgcheader 8 4]
	set ::pgced::nextpgcn  [get_int ::pgced::__pgcheader 0X9C 2]
	set ::pgced::prevpgcn  [get_int ::pgced::__pgcheader 0X9E 2]

	set tmp [get_int ::pgced::__pgcheader 0xA0 2]

	if {$type == "menu"} {
		if {$tmp == 65535} {
			set ::pgced::goup_type 0
			set ::pgced::gouppgcn 0
		} else {
			set ::pgced::goup_type 1
			set ::pgced::gouppgcn $tmp
		}
	} else {
		set ::pgced::goup_type 1
		set ::pgced::gouppgcn $tmp
	}
	set ::pgced::stilltime [get_int ::pgced::__pgcheader 0XA3 1]

	# build interface
	toplevel $w -class Dialog
	wm withdraw $w
	catch {grab set $w}
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::done 0}
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	bind $w <Escape> {set ::pgced::done 0}

	frame $w.f1
	pack  $w.f1 -pady 2 -anchor w -fill x

	# misc PGC values
	frame $w.f1.fl
	pack  $w.f1.fl -fill x -expand 1 -padx 2 -pady 1 -side left
	frame $w.f1.fl.mf
	pack  $w.f1.fl.mf -padx 2 -fill x -expand 1
	label $w.f1.fl.mf.l0 -text "Playback Time "
	entry $w.f1.fl.mf.e0 -textvariable ::pgced::pbtimetxt -state readonly -justify center

	label $w.f1.fl.mf.l1 -text "PG Playback Mode "

#	entry $w.f1.fl.mf.e1 -textvariable ::pgced::pbmodetxt -state readonly -justify center
	frame $w.f1.fl.mf.f1 -bd 2 -relief groove
	radiobutton $w.f1.fl.mf.f1.r1 -pady 0 -bd 1 -variable ::pgced::pbmode -value 0 \
			-text "Sequential" -command "$w.f1.fl.mf.f1.f2.e configure -state disabled"
	frame $w.f1.fl.mf.f1.f1
	radiobutton $w.f1.fl.mf.f1.f1.r2 -pady 0 -bd 1 -variable ::pgced::pbmode -value 1 \
			-text "Random:" -command "$w.f1.fl.mf.f1.f2.e configure -state normal; focus $w.f1.fl.mf.f1.f2.e"
	radiobutton $w.f1.fl.mf.f1.f1.r3 -pady 0 -bd 1 -variable ::pgced::pbmode -value 2 \
			-text "Shuffle:"  -command "$w.f1.fl.mf.f1.f2.e configure -state normal; focus $w.f1.fl.mf.f1.f2.e"
	pack  $w.f1.fl.mf.f1.r1 -pady 0 -anchor w
	pack  $w.f1.fl.mf.f1.f1 -side left
	pack  $w.f1.fl.mf.f1.f1.r2 -pady 0 -anchor w
	pack  $w.f1.fl.mf.f1.f1.r3 -pady 0 -anchor w
	frame $w.f1.fl.mf.f1.f2
	pack  $w.f1.fl.mf.f1.f2 -side left
	entry $w.f1.fl.mf.f1.f2.e -width 3 -textvariable ::pgced::pbprogs -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 1 127 {%d} {%P}"
	pack  $w.f1.fl.mf.f1.f2.e -side left
	if {$::pgced::pbmode == 0} {
		$w.f1.fl.mf.f1.f2.e configure -state disabled
	} else {
		$w.f1.fl.mf.f1.f2.e configure -state normal
	}
	::balloon::set_balloon $w.f1.fl.mf.f1.f2.e "Number of programs to play - 1\n(min 1, max 127)"
	label $w.f1.fl.mf.f1.f2.l -text "programs"
	pack  $w.f1.fl.mf.f1.f2.l -side left

	label $w.f1.fl.mf.l2 -text "Prohibited User Operations "
	entry $w.f1.fl.mf.e2 -textvariable ::pgced::puos -width 10 -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0x1FFFFFF {%d} {%P}"
	button $w.f1.fl.mf.b2 -text "..." -pady 0 -width 2 \
			-command {::pgced::puos_editor "Prohibited User Operations" .pgceditor}
	bind   $w.f1.fl.mf.b2 <3> {set ::pgced::puos 0}
	label $w.f1.fl.mf.l3 -text "Previous PGCN link "
	entry $w.f1.fl.mf.e3 -textvariable ::pgced::prevpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFFFF {%d} {%P}"
	label $w.f1.fl.mf.l4 -text "Next PGCN link "
	entry $w.f1.fl.mf.e4 -textvariable ::pgced::nextpgcn  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFF {%d} {%P}"
	label $w.f1.fl.mf.l5 -text "Go Up (Return) link "
	frame $w.f1.fl.mf.e5 -bd 2 -relief groove

	frame $w.f1.fl.mf.e5.f1
	pack  $w.f1.fl.mf.e5.f1 -side top -fill x -pady 0
	radiobutton $w.f1.fl.mf.e5.f1.c -text "PGCN" -variable ::pgced::goup_type -value 1 \
			-command "$w.f1.fl.mf.e5.f1.e configure -state normal; focus $w.f1.fl.mf.e5.f1.e" -padx 0 -pady 0 -bd 1
	pack  $w.f1.fl.mf.e5.f1.c -side left -padx 0
	entry $w.f1.fl.mf.e5.f1.e -textvariable ::pgced::gouppgcn -justify center -width 16 \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0xFF {%d} {%P}"
	pack  $w.f1.fl.mf.e5.f1.e -side left -fill x -padx 2
	radiobutton $w.f1.fl.mf.e5.c -text "Resume (menus only)" -variable ::pgced::goup_type -value 0 \
			-command "$w.f1.fl.mf.e5.f1.e configure -state disabled" -padx 0 -pady 0 -bd 1
	pack  $w.f1.fl.mf.e5.c -side left -pady 0
	if {$type != "menu"} {
		$w.f1.fl.mf.e5.c configure -state disabled
	}
	if {! $::pgced::goup_type} {
		$w.f1.fl.mf.e5.f1.e configure -state disabled
	}

	label $w.f1.fl.mf.l6 -text "PGC Still Time (255=infinite) "
	entry $w.f1.fl.mf.e6 -textvariable ::pgced::stilltime  -justify center \
			-validate key -validatecommand "::pgced::value_entry_validate 0 0XFF {%d} {%P}"
	::balloon::set_balloon $w.f1.fl.mf.e6 "0 to 254 seconds\nor 255 = infinite"

	grid  $w.f1.fl.mf.l0 -row 0 -column 0 -sticky e ;#-ipady 1
	grid  $w.f1.fl.mf.e0 -row 0 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l1 -row 1 -column 0 -sticky e -ipady 1
	grid  $w.f1.fl.mf.f1 -row 1 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l2 -row 2 -column 0 -sticky e ;#-ipady 1
	grid  $w.f1.fl.mf.e2 -row 2 -column 1 -sticky ew
	grid  $w.f1.fl.mf.b2 -row 2 -column 2 -sticky ew
	grid  $w.f1.fl.mf.l3 -row 3 -column 0 -sticky e -ipady 1
	grid  $w.f1.fl.mf.e3 -row 3 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l4 -row 4 -column 0 -sticky e ;#-ipady 1
	grid  $w.f1.fl.mf.e4 -row 4 -column 1 -sticky ew -columnspan 2
	grid  $w.f1.fl.mf.l5 -row 5 -column 0 -sticky e ;#-ipady 1
	grid  $w.f1.fl.mf.e5 -row 5 -column 1 -sticky ew -columnspan 2 ;#-pady 2
	grid  $w.f1.fl.mf.l6 -row 6 -column 0 -sticky e ;#-ipady 1
	grid  $w.f1.fl.mf.e6 -row 6 -column 1 -sticky ew -columnspan 2
	grid columnconfigure $w.f1.fl.mf 1 -weight 100

	# audio streams control
	frame $w.f1.mfa -borderwidth 2 -relief groove
	pack  $w.f1.mfa -padx 2 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfa.tit -text "Audio\nStreams" -pady 0
	grid  $w.f1.mfa.tit -row 0 -column 0 -columnspan 2
	for {set a 1} {$a <= 8} {incr a} {
		label $w.f1.mfa.al$a -text "[expr {$a-1}]"
		button $w.f1.mfa.ab$a -textvariable ::pgced::audiostream$a -width 5 -pady 0 \
				-command "::pgced::set_audio ::pgced::audiostream$a $a"
		grid  $w.f1.mfa.al$a -row $a -column 0 -sticky e
		grid  $w.f1.mfa.ab$a -row $a -column 1 -sticky ew
	}
	bind  $w.f1.mfa     $::config(popups_button) "tk_popup $w.f1.mfa.menu \[winfo pointerx .\] \[winfo pointery .\] 0"
	bind  $w.f1.mfa.tit $::config(popups_button) "tk_popup $w.f1.mfa.menu \[winfo pointerx .\] \[winfo pointery .\] 0"
	if {! [winfo exists $w.f1.mfa.menu]} {
		menu $w.f1.mfa.menu -tearoff 0
		$w.f1.mfa.menu add command -label "Copy all"  -command "::pgced::copy_all_audio"
		$w.f1.mfa.menu add command -label "Paste all" -command "::pgced::paste_all_audio"
	}

	# sub-picture streams control
	frame $w.f1.mfs -borderwidth 2 -relief groove
	pack  $w.f1.mfs -padx 2 -pady 1 -ipadx 4 -ipady 2 -side left
	label $w.f1.mfs.tit -text "Sub-picture VOB's Decoding Streams\n(for 4:3, wide, letterbox, pan&scan)" -pady 0
	grid  $w.f1.mfs.tit -row 0 -column 0 -columnspan 8
	set numsubpics [::streams::get_pgc_num_subpic $type $index]
	for {set s 1} {$s <= 32} {incr s} {
		if {$s <= $numsubpics} {
			set descr [string map {" Off:" : " On:" :} [::streams::get_pgc_subpic_descr $type $index [expr {$s-1}]]]
		} else {
			set descr "[expr {$s-1}]: not used."
		}
		binary scan [string range $::pgced::__pgcheader [expr {(0X18 + ($s*4))}] [expr {(0X18 + ($s*4) + 3)}]] cccc s1 s2 s3 s4
		if {$s1 == 0} {
			set ::pgced::subpicstream$s "none"
		} else {
			set ::pgced::subpicstream$s [string map {{ } {  }} [format "%2d %2d %2d %2d" [expr {$s1 & 0X1F}] $s2 $s3 $s4]]
		}
		label $w.f1.mfs.al$s -text " [expr {$s-1}]"
		button $w.f1.mfs.ab$s -textvariable ::pgced::subpicstream$s -width 12 -pady 0 \
				-command "::pgced::set_subpic ::pgced::subpicstream$s \"$descr\""
		set row [expr {(($s-1) % 8) + 1}]
		set col [expr {(($s-1) / 8) * 2}]
		grid  $w.f1.mfs.al$s -row $row -column $col -sticky e
		grid  $w.f1.mfs.ab$s -row $row -column [expr {$col+1}] -sticky ew
		::balloon::set_balloon $w.f1.mfs.ab$s $descr
	}
	bind  $w.f1.mfs     $::config(popups_button) "tk_popup $w.f1.mfs.menu \[winfo pointerx .\] \[winfo pointery .\] 0"
	bind  $w.f1.mfs.tit $::config(popups_button) "tk_popup $w.f1.mfs.menu \[winfo pointerx .\] \[winfo pointery .\] 0"
	if {! [winfo exists $w.f1.mfs.menu]} {
		menu $w.f1.mfs.menu -tearoff 0
		$w.f1.mfs.menu add command -label "Copy all"  -command "::pgced::copy_all_subpic"
		$w.f1.mfs.menu add command -label "Paste all" -command "::pgced::paste_all_subpic"
	}

	# bottom frame
	frame  $w.f2
	pack   $w.f2 -expand 1 -fill both

	# color table
	frame $w.f2.mfc -borderwidth 2 -relief groove -pady 2
	pack  $w.f2.mfc -side left -fill x -ipadx 2 -padx 4 -anchor n
	label $w.f2.mfc.l0 -text "Sub-picture\nColor Lookup Table "
	pack  $w.f2.mfc.l0 -fill x -pady 4
	frame $w.f2.mfc.f1
	pack  $w.f2.mfc.f1
	set col 0
	set row 0
	for {set c 0} {$c < 16} {incr c} {
		button $w.f2.mfc.f1.b$c -width 2 -pady 0 -command "::pgced::set_color $c $w.f2.mfc.f1.b$c"
		grid   $w.f2.mfc.f1.b$c -row $row -column $col -ipadx 3 -ipady 3 -padx 1 -pady 1 -sticky ns
		incr col
		if {$col == 4} {
			set col 0
			grid rowconfigure $w.f2.mfc.f1 $row -weight 1
			incr row
		}
	}
	frame $w.f2.mfc.f2
	pack  $w.f2.mfc.f2 -pady 2
	button $w.f2.mfc.f2.bc -text "Copy all"  -width 8 -command ::pgced::copy_colors
	button $w.f2.mfc.f2.bp -text "Paste all" -width 8 -command ::pgced::paste_colors
	set txt "You may copy from one DVD to another\none by launching PgcEdit twice.\nYou must do the paste before closing the\nsource PgcEdit!"
	::balloon::set_balloon $w.f2.mfc.f2.bc $txt
	::balloon::set_balloon $w.f2.mfc.f2.bp $txt
	pack   $w.f2.mfc.f2.bc -side left -padx 2
	pack   $w.f2.mfc.f2.bp -side right -padx 2

	frame $w.f2.mfc.f3
	pack  $w.f2.mfc.f3 -pady 2
	button $w.f2.mfc.f3.bc -text "Export" -width 8 -command ::pgced::export_colors
	button $w.f2.mfc.f3.bp -text "Import" -width 8 -command ::pgced::import_colors
	set txt " the raw YcRcB color table as 64 binary bytes."
	::balloon::set_balloon $w.f2.mfc.f3.bc "Export$txt"
	::balloon::set_balloon $w.f2.mfc.f3.bp "Import$txt"
	pack   $w.f2.mfc.f3.bc -side left -padx 2
	pack   $w.f2.mfc.f3.bp -side right -padx 2


	### Cells table
	frame  $w.f2.fc -borderwidth 2 -relief groove
	pack   $w.f2.fc -side left -padx 2 -expand 1 -fill both -anchor w
	canvas $w.f2.fc.c -yscrollcommand "$w.f2.fc.vs set" -height 8 \
			-highlightcolor [$w.f2.fc cget -background]
	pack   $w.f2.fc.c -side left -expand 1 -fill both
	scrollbar $w.f2.fc.vs -command "$w.f2.fc.c yview" -orient vert
	pack   $w.f2.fc.vs -side left -fill y
	frame  $w.f2.fc.c.f	;# destroyed later

	### buttons
	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4 -fill x
	button $w.fb.ss -text " Current domain streams attributes " \
			-command "::streams::set_streams_attribs $w"
	pack   $w.fb.ss -side left -padx 2
	# new!
	button $w.fb.rc -text "            Remove Cell            " \
			-command "::pgced::remove_cell $w"
	pack   $w.fb.rc -side left -padx 2
	if {$type == "pgc" && $vts != 0} {
		button $w.fb.ptt -text "    Edit Chapter (PTT) table    " \
				-command "if \[::pgced::ptted::ptt_table_editor $w\] ::pgced::refresh_chapters"
		pack   $w.fb.ptt -side left -padx 2
	}
	frame  $w.fb.pad
	pack   $w.fb.pad -side left -fill x -expand 1
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::done 1}
	pack   $w.fb.ok -side left -padx 3
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::done 0}
	pack   $w.fb.cancel -side left -padx 2

	wm title $w "PGC Editor - [$::widget(main_cmd_lab) cget -text]"

	for {set c 0} {$c < 16} {incr c} {
		set yc_color [string range $::pgced::__pgcheader [expr {(0XA4 + ($c*4))}] [expr {(0XA4 + ($c*4) + 3)}]]
		$w.f2.mfc.f1.b$c configure -background [yc2rgb $yc_color]
	}
	set ::pgced::__color_table_modified false
	set numaudios [::streams::get_pgc_num_audio $type $index]
	for {set a 1} {$a <= 8} {incr a} {
		if {$a <= $numaudios} {
			set descr [::streams::get_pgc_audio_descr $type $index [expr {$a-1}]]
			::balloon::set_balloon $w.f1.mfa.ab$a $descr
		} else {
			::balloon::set_balloon $w.f1.mfa.ab$a "[expr {$a-1}]: not used."
		}
		binary scan [string range $::pgced::__pgcheader [expr {(0X0A + ($a*2))}] [expr {(0X0A + ($a*2))}]] c audio_status
		if {$audio_status == 0} {
			set ::pgced::audiostream$a "none"
		} else {
			set ::pgced::audiostream$a [expr {$audio_status & 7}]
		}
	}

	# window geometry
	#set maxheight [winfo height $w.f2.fc.c.f]
	#if {$maxheight > 200} {set maxheight 200}
	#$w.f2.fc.c configure -height $maxheight
	update
	if {! [winfo exists $w]} {return ""}
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-$ww)/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-$wh)/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	$w.f2.fc.c configure -height 1
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 1 1

	if {! [winfo ismapped $w]} {
		wm deiconify $w
	}
	grab set $w
	focus $w

	set end_time [build_cell_table $w]
	update idletasks
	if {$end_time != -1 && $end_time != $::pgced::__pgcpbtime} {
		set txt "WARNING: The PGC Playback Time ([BCDtime $::pgced::__pgcpbtime true]) doesn't match"
		set txt "$txt\nthe total of the cells times ([BCDtime $end_time true])."
		set txt "$txt\n\nFix it?"
		set r [tk_messageBox -title "Pgc Editor" -icon warning \
			-message $txt -parent . -type yesno -default yes]
		if {$r == "yes"} {
			set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$end_time[string range $::pgced::__pgcheader 8 end]"
			set ::pgced::pbtimetxt [BCDtime $end_time true]
			set ::pgced::__pgcpbtime $end_time
		}
	}

	show_msg ""

	set ::pgced::done -1
	tkwait variable ::pgced::done
	wm withdraw $w
	update idletasks

	if {$::pgced::done} {
		accept_modifs $w
		destroy $w
		return [list $::pgced::__pgcheader $::pgced::__cellpbtable $::pgced::__cellpostable $::pgced::__programmap $::pgced::__PTTlist $::pgced::__numcells]
	} else {
		destroy $w
		return ""
	}
}

proc accept_modifs {w} {
	switch $::pgced::pbmode {
		0 {
			set_int ::pgced::__pgcheader 0xA2 0 1
		}
		1 {
			set_int ::pgced::__pgcheader 0xA2 $::pgced::pbprogs 1
		}
		2 {
			set_int ::pgced::__pgcheader 0xA2 [expr {$::pgced::pbprogs | 0x80}] 1
		}
	}
	set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$::pgced::__pgcpbtime[string range $::pgced::__pgcheader 8 end]"

	if {$::pgced::puos      != ""} {set_int ::pgced::__pgcheader 8    [expr {$::pgced::puos}] 4}
	if {$::pgced::nextpgcn  != ""} {set_int ::pgced::__pgcheader 0X9C [expr {$::pgced::nextpgcn}] 2}
	if {$::pgced::prevpgcn  != ""} {set_int ::pgced::__pgcheader 0X9E [expr {$::pgced::prevpgcn}] 2}
	if {$::pgced::goup_type} {
		if {$::pgced::gouppgcn  != ""} {set_int ::pgced::__pgcheader 0XA0 [expr {$::pgced::gouppgcn}] 2}
	} else {
		set_int ::pgced::__pgcheader 0XA0 65535 2
	}
	if {$::pgced::stilltime != ""} {set_int ::pgced::__pgcheader 0XA3 [expr {$::pgced::stilltime}] 1}
	if {$::pgced::__color_table_modified} {
		for {set c 0} {$c < 16} {incr c} {
			set rgb_color [$w.f2.mfc.f1.b$c cget -background]
			set_int ::pgced::__pgcheader [expr {(0XA4 + (4*$c))}] [rgb2yc $rgb_color] 4
		}
	}
	for {set a 1} {$a <= 8} {incr a} {
		set varname ::pgced::audiostream$a
		if {[subst $$varname] == "none"} {
			set as 0
		} else {
			set as [expr {[subst $$varname] | 0X80}]
		}
		set_int ::pgced::__pgcheader [expr {(0X0A + ($a*2))}] $as 1
	}
	for {set s 1} {$s <= 32} {incr s} {
		set varname ::pgced::subpicstream$s
		if {[subst $$varname] == "none"} {
			set ss 0
		} else {
			set sl [subst $$varname]
			set ss [expr {(0X80000000 + \
					([lindex $sl 0] * 0X1000000) + \
					([lindex $sl 1] * 0X10000) + \
					([lindex $sl 2] * 0X100) + \
					[lindex $sl 3])}]
		}
		set_int ::pgced::__pgcheader [expr {(0X18 + ($s*4))}] $ss 4
	}

	for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
		set cellidx1 [expr {($c - 1) * 24}]
		set_int ::pgced::__cellpbtable $cellidx1 [$w.f2.fc.c.f.w_${c}_3 cget -text] 1
		set_int ::pgced::__cellpbtable [expr {$cellidx1 + 1}] [subst \$::pgced::var_${c}_5] 1
		set tmp [subst \$::pgced::var_${c}_6]
		if {$tmp == "" || $tmp == "0x" || $tmp == "0X"} {set tmp 0}
		set_int ::pgced::__cellpbtable [expr {$cellidx1 + 2}] [expr {$tmp}] 1
		set tmp [subst \$::pgced::var_${c}_7]
		if {$tmp == "" || $tmp == "0x" || $tmp == "0X"} {set tmp 0}
		set_int ::pgced::__cellpbtable [expr {$cellidx1 + 3}] [expr {$tmp}] 1
	}
}

### build Cells table (+ PTT and Programs infos)
proc build_cell_table {w} {

	destroy $w.f2.fc.c.f
	frame $w.f2.fc.c.f
	pack  $w.f2.fc.c.f -anchor nw
	$w.f2.fc.c create window 0 0 -window $w.f2.fc.c.f -anchor nw

	if {$::pgced::__numcells == 0} {
		label $w.f2.fc.c.f.w_0_0 -text "Dummy PGC: no cells."
		pack $w.f2.fc.c.f.w_0_0
		set end_time -1
	} else {
		show_msg "Building cell table..."
		if {$::pgced::__pgctype == "menu"} {
			label $w.f2.fc.c.f.w_0_0 -text "Num\nBut-\ntons"
		} else {
			label $w.f2.fc.c.f.w_0_0 -text "Chap.\n(PTT)"
		}
		label $w.f2.fc.c.f.w_0_1 -text "Prog."
		label $w.f2.fc.c.f.w_0_2 -text "Cell"
		label $w.f2.fc.c.f.w_0_3 -text "Type\nFlags"
		label $w.f2.fc.c.f.w_0_4 -text "Layer\nBreak"
		label $w.f2.fc.c.f.w_0_5 -text "Res-\ntricted"
		label $w.f2.fc.c.f.w_0_6 -text "Still\nTime"
		label $w.f2.fc.c.f.w_0_7 -text "Cell\nCmd"
		label $w.f2.fc.c.f.w_0_8 -text "Playback\nTime"
		label $w.f2.fc.c.f.w_0_9 -text "End\nTime"
		label $w.f2.fc.c.f.w_0_10 -text "Entry\nVOBU\nsector"
		label $w.f2.fc.c.f.w_0_11 -text "First\nILVU\nEnd"
		label $w.f2.fc.c.f.w_0_12 -text "Last\nVOBU\nStart"
		label $w.f2.fc.c.f.w_0_13 -text "Last\nVOBU\nEnd"
		label $w.f2.fc.c.f.w_0_14 -text "VOB/Cell\nID"
		label $w.f2.fc.c.f.w_0_15 -text ""

		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set cell_program($c) ""
			set prog_chapter($c) ""
		}
		for {set p 1} {$p <= $::pgced::__numprograms} {incr p} {
			set cell_program([get_int ::pgced::__programmap [expr {$p-1}] 1]) $p
		}
		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set prog_chapter($c) 0
		}
		foreach pttpair $::pgced::__PTTlist {
			set pg  [lindex $pttpair 0]
			set ptt [lindex $pttpair 1]
			for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
				if {$pg == $cell_program($c) && $prog_chapter($c) == 0} {
					set prog_chapter($c) $ptt
				}
			}
		}

		set vts [lindex [split $::pgced::__index ","] 0]
		set end_time "\000\000\000\000"
		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set cellidx1 [expr {($c - 1) * 24}]
			set cellidx2 [expr {($c - 1) * 4}]
			set vobid  [get_int ::pgced::__cellpostable $cellidx2 2]
			set cellid [get_int ::pgced::__cellpostable [expr {$cellidx2 + 3}] 1]
			if {$::pgced::__pgctype == "menu"} {
				if {[info exists ::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)]} {
					set txt $::buttons::butsinfo(menu,$vts,$vobid,$cellid,numbuts)
				} else {
					set txt "?"
				}
				label $w.f2.fc.c.f.w_${c}_0 -text $txt
			} else {
#				button $w.f2.fc.c.f.w_${c}_0 -text $prog_chapter($c) -anchor e -pady 0 -width 3 \
#					-command "::pgced::toggle_ptt $w"
				set txt $prog_chapter($c)
				if {$txt == "0"} {set txt ""}
				label $w.f2.fc.c.f.w_${c}_0 -text $txt
			}
			button $w.f2.fc.c.f.w_${c}_1 -text $cell_program($c) -anchor e -pady 0 -width 3 \
					-command "::pgced::toggle_program $w $c"
			if {$::pgced::__pgctype == "pgc" && $cell_program($c) == ""} {
				$w.f2.fc.c.f.w_${c}_0 configure -relief flat -state disabled -text ""
			}

			label $w.f2.fc.c.f.w_${c}_2 -text $c

			set type [get_int ::pgced::__cellpbtable $cellidx1 1]
			button $w.f2.fc.c.f.w_${c}_3 -text $type -width 3 -anchor e -pady 0 \
					-command "::pgced::cell_type_editor $w.f2.fc.c.f.w_${c}_3 ::pgced::var_${c}_4"

			checkbutton $w.f2.fc.c.f.w_${c}_4 -variable ::pgced::var_${c}_4 \
					-command "::pgced::set_layer_break ::pgced::var_${c}_4 $w.f2.fc.c.f.w_${c}_3"
			set entryvobu [get_int ::pgced::__cellpbtable [expr {$cellidx1 +  8}] 4]
			if {$c > 1} {
				set lastvobuend [get_int ::pgced::__cellpbtable [expr {$cellidx1 - 4}] 4]
			} else {
				set lastvobuend -2
			}
			if {[expr {$type & 8}] == 0} {
				set [subst ::pgced::var_${c}_4] 1
				 if {[expr {$lastvobuend+1}] != $entryvobu} {
				 	$w.f2.fc.c.f.w_${c}_4 configure -state disabled
				 }
			} else {
				set [subst ::pgced::var_${c}_4] 0
			}
			if {$c == 1} {$w.f2.fc.c.f.w_${c}_4 configure -state disabled}
			set  ::pgced::var_${c}_5 [expr {[get_int ::pgced::__cellpbtable [expr {$cellidx1 + 1}] 1] & 64}]
			checkbutton $w.f2.fc.c.f.w_${c}_5 -variable ::pgced::var_${c}_5 \
					-variable ::pgced::var_${c}_5 -onvalue 64 -offvalue 0
			set  ::pgced::var_${c}_6 [get_int ::pgced::__cellpbtable [expr {$cellidx1 + 2}] 1]
			entry $w.f2.fc.c.f.w_${c}_6 -textvariable ::pgced::var_${c}_6 -width 3 -justify right \
					-validate all -validatecommand "::ced::value_entry_validate 0 255 {%d} {%P}"
			::balloon::set_balloon $w.f2.fc.c.f.w_${c}_6 "0 to 254 seconds\nor 255 = infinite"
			set  ::pgced::var_${c}_7 [get_int ::pgced::__cellpbtable [expr {$cellidx1 + 3}] 1]
			entry $w.f2.fc.c.f.w_${c}_7 -textvariable ::pgced::var_${c}_7 -width 3 -justify right \
					-validate all -validatecommand "::ced::value_entry_validate 0 128 {%d} {%P}"
			set cell_time [string range $::pgced::__cellpbtable [expr {$cellidx1 + 4}] [expr {$cellidx1 + 7}]]
			label $w.f2.fc.c.f.w_${c}_8 -text [BCDtime $cell_time false]

			# Add cell time to total PGC time, but only if angle 1.
			if {[expr {$type & 0x10}] == 0 || [expr {$type & 0xC0}] == 0x40} {
				set end_time [AddBCDtime $cell_time $end_time]
			}
			label $w.f2.fc.c.f.w_${c}_9 -text [BCDtime $end_time false]
			label $w.f2.fc.c.f.w_${c}_10 -text $entryvobu
			label $w.f2.fc.c.f.w_${c}_11 -text [get_int ::pgced::__cellpbtable [expr {$cellidx1 + 12}] 4]
			label $w.f2.fc.c.f.w_${c}_12 -text [get_int ::pgced::__cellpbtable [expr {$cellidx1 + 16}] 4]
			label $w.f2.fc.c.f.w_${c}_13 -text [get_int ::pgced::__cellpbtable [expr {$cellidx1 + 20}] 4]
			button $w.f2.fc.c.f.w_${c}_14 -pady 0 \
					-text "$vobid / $cellid" \
					-command "::pgced::accept_modifs $w ; \
							::pgced::reassign_VOB_ID $w $c $vobid $cellid"
			if {$::config(use_preview)} {
				button $w.f2.fc.c.f.w_${c}_15 -image ::preview::bitmapplay -height 15 -command \
						"::preview::preview2 [lindex $::editor_cmds_from 0] [lindex $::editor_cmds_from 1] $c false false 0"
			}
		}
		if {$::config(use_preview)} {
			set lastitem 15
			grid columnconf $w.f2.fc.c.f 15 -pad 4
		} else {
			set lastitem 14
		}
		for {set c 0} {$c <= $::pgced::__numcells} {incr c} {
			for {set f 0} {$f <= $lastitem} {incr f} {
				grid $w.f2.fc.c.f.w_${c}_$f -row $c -column $f
			}
			grid configure $w.f2.fc.c.f.w_${c}_14 -sticky ew ;#-padx 2
		}
		#for {set f 0} {$f <= $lastitem} {incr f} {
		#	grid columnconf $w.f2.fc.c.f $f -minsize 28 -pad 2
		#}
	}

	# inform scrollbar of the height of the canvas's internal window
	if {[winfo exists $w.f2.fc.c.f]} {
		update
		$w.f2.fc.c configure -width [winfo width $w.f2.fc.c.f] \
				-scrollregion "0 0 0 [winfo height $w.f2.fc.c.f]" \
	}
	show_msg ""
	return $end_time
}

proc refresh_chapters {} {
	for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
		set prog_chapter($c) 0
		.pgceditor.f2.fc.c.f.w_${c}_0 configure -text ""
	}
	foreach pttpair $::pgced::__PTTlist {
		set pg  [lindex $pttpair 0]
		set ptt [lindex $pttpair 1]
		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set cell_program [.pgceditor.f2.fc.c.f.w_${c}_1 cget -text]
			if {$pg == $cell_program && $prog_chapter($c) == 0} {
				set prog_chapter($c) $ptt
				if {$ptt == 0} {set ptt ""}
				.pgceditor.f2.fc.c.f.w_${c}_0 configure -text $ptt
			}
		}
	}
}

proc value_entry_validate {min max action val} {
	if {$action != 0} {
		if {$val=="0"||$val=="0x"||$val=="0X"} {
			if {$min>0 && $val=="0"} bell
			return true
		}
		if {![string is integer $val]} {
			bell
			return false
		}
		if {$val < $min} {
			bell
			return false
		}
		if {$val > $max} {
			bell
			return false
		}
	}
	return true
}

proc rgb2yc {rgb_color} {
	if {[string range $rgb_color 0 0] == "#"} {
		set rgb_color [string range $rgb_color 1 end]
	}
	set r [x2d [string range $rgb_color 0 1]]
	set g [x2d [string range $rgb_color 2 3]]
	set b [x2d [string range $rgb_color 4 5]]

	set y  [expr {round(( 0.257*$r) + (0.504*$g) + (0.098*$b) + 16)}]
	set cr [expr {round(( 0.439*$r) - (0.368*$g) - (0.071*$b) + 128)}]
	set cb [expr {round((-0.148*$r) - (0.291*$g) + (0.439*$b) + 128)}]

	if {$y  < 0} {set y  0}
	if {$cr < 0} {set cr 0}
	if {$cb < 0} {set cb 0}
	if {$y  > 255} {set y  255}
	if {$cr > 255} {set cr 255}
	if {$cb > 255} {set cb 255}

	return [expr {($y*0X10000) + ($cr*0X100) + $cb}]
}

proc yc2rgb {yc_color} {
	if {[string length $yc_color] == 4} {
		set yc_color [string range $yc_color 1 end]
	}
	binary scan $yc_color ccc y cr cb
	set y  [expr {($y  + 0X100) % 0X100}]
	set cr [expr {($cr + 0X100) % 0X100}]
	set cb [expr {($cb + 0X100) % 0X100}]

	set r [expr {round((1.164*($y-16)) + (1.596*($cr-128)))}]
	set g [expr {round((1.164*($y-16)) - (0.813*($cr-128)) - (0.392*($cb-128)))}]
	set b [expr {round((1.164*($y-16)) + (2.017*($cb-128)))}]

	if {$r < 0} {set r 0}
	if {$g < 0} {set g 0}
	if {$b < 0} {set b 0}
	if {$r > 255} {set r 255}
	if {$g > 255} {set g 255}
	if {$b > 255} {set b 255}

	return [format "#%02X%02X%02X" $r $g $b]
}

proc set_color {colornum w} {
	set newcolor [tk_chooseColor -initialcolor [$w cget -background] -parent .pgceditor -title "Select color #$colornum"]
	if {$newcolor == ""} return
	$w configure -background $newcolor
	set ::pgced::__color_table_modified true
}

proc copy_colors {} {
	set colors {}
	for {set c 0} {$c < 16} {incr c} {
		lappend colors [.pgceditor.f2.mfc.f1.b$c cget -background]
	}
	clipboard clear
	clipboard append $colors
}

proc paste_colors {} {
	set err [catch {set colors [clipboard get]}]
	if {$err || [llength $colors] != 16} {
		bell
		return
	}
	for {set c 0} {$c < 16} {incr c} {
		.pgceditor.f2.mfc.f1.b$c configure -background [lindex $colors $c]
	}
	set ::pgced::__color_table_modified true
}

proc copy_all_audio {} {
	set streams pgcedit_audio_streams
	for {set a 1} {$a <= 8} {incr a} {
		set varname ::pgced::audiostream$a
		lappend streams [subst $$varname]
	}
	clipboard clear
	clipboard append $streams
}
proc paste_all_audio {} {
	set streams [clipboard get]
	if {[llength $streams] != 9 || [lindex $streams 0] != "pgcedit_audio_streams"} {bell;return}
	for {set a 1} {$a <= 8} {incr a} {
		set ::pgced::audiostream$a [lindex $streams $a]
	}
}
proc copy_all_subpic {} {
	set streams pgcedit_subpic_streams
	for {set s 1} {$s <= 32} {incr s} {
		set varname ::pgced::subpicstream$s
		lappend streams [subst $$varname]
	}
	clipboard clear
	clipboard append $streams
}
proc paste_all_subpic {} {
	set streams [clipboard get]
	if {[llength $streams] != 33 || [lindex $streams 0] != "pgcedit_subpic_streams"} {bell;return}
	for {set s 1} {$s <= 32} {incr s} {
		set ::pgced::subpicstream$s [lindex $streams $s]
	}
}

proc export_colors {} {
	set txt "Export YCrCb colors as 64 bytes raw data?"
	set txt "$txt\nAnswer 'No' to export RGB colors as ASCII text."
	set r [tk_messageBox -title "Export Color Lookup Table" -icon question \
			-message $txt -parent .pgceditor -type yesnocancel -default yes]
	if {$r == "cancel"} return
	if {$r == "yes"} {
		set extension .clb
		set filetypes {{"YCrCb raw color table" {*.clb}} {"all files" {*.*}}}
	} else {
		set extension .txt
		set filetypes {{"Text files with RGB colors" {*.txt}} {"all files" {*.*}}}
	}

	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set defaultfilename "[file rootname [::main::get_default_filename $::editor_cmds_from]]$extension"
	set filename [tk_getSaveFile -filetypes $filetypes \
		-parent .pgceditor -initialdir $::config(import_initialdir) \
		-initialfile $defaultfilename -title "PgcEdit: Export PGC color lookup table"]
	if {$filename == ""} {
		return
	}
	set ::config(import_initialdir) [file dirname $filename]
	if {[string toupper [file extension $filename]] != [string toupper $extension]} {
		set filename "${filename}$extension"
	}

	if {$r == "yes"} {
		saveFile $filename binary [string range $::pgced::__pgcheader 0xA4 0xE3]
	} else {
		set txt ""
		for {set c 0} {$c < 16} {incr c} {
			set rgb [.pgceditor.f2.mfc.f1.b$c cget -background]
			set r [x2d [string range $rgb 1 2]]
			set g [x2d [string range $rgb 3 4]]
			set b [x2d [string range $rgb 5 6]]
			set txt "${txt}\t\tColor $c=$r, $g, $b\n"
		}
		set txt [string trim $txt \n]
		saveFile $filename auto $txt
	}
}

proc import_colors {} {
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]
	set defaultfilename "[file rootname [::main::get_default_filename $::editor_cmds_from]].clb"
	set filename [tk_getOpenFile \
		-filetypes {{"All color lookup table formats" {*.clb *.txt *.mxp VTS_*_0.IFO}} {"YCrCb raw color table" {*.clb}} {"Text files with RGB colors" {*.txt}} {"VTS IFO files" {VTS_*_0.IFO}} {"Muxman project files (RGB)" {*.mxp}} {"all files" {*.*}}} \
		-parent .pgceditor -initialdir $::config(import_initialdir) \
		-initialfile $defaultfilename -title "PgcEdit: Import PGC color lookup table"]
	if {$filename == ""} {
		return
	}
	set ::config(import_initialdir) [file dirname $filename]
	if {![file exists $filename] && [string toupper [file extension $filename]] != ".CLB"} {
		set filename "${filename}.clb"
	}
	if {![file exists $filename]} {
		show_warning -icon error -title "PgcEdit $::program_version error" \
			-message "File not found:\n$filename" -parent .pgceditor
		return
	}
	set rawcolors [readFile $filename binary]
	if {[string length $rawcolors] == 64} {
		for {set c 0} {$c < 16} {incr c} {
			set yc_color [string range $rawcolors [expr {$c*4}] [expr {($c*4) + 3}]]
			.pgceditor.f2.mfc.f1.b$c configure -background [yc2rgb $yc_color]
		}
		set ::pgced::__color_table_modified true
	} elseif {[string toupper [file extension $filename]] == ".IFO"} {
		set pgcit_start [expr [get_int rawcolors 0xCC 4] * 0x800]
		set numsrp [get_int rawcolors $pgcit_start 2]
		set start [expr $pgcit_start + 8 + ($numsrp * 8) + 0xA4]
		set rawcolors [string range $rawcolors $start [expr $start + 63]]
		for {set c 0} {$c < 16} {incr c} {
			set yc_color [string range $rawcolors [expr {$c*4}] [expr {($c*4) + 3}]]
			.pgceditor.f2.mfc.f1.b$c configure -background [yc2rgb $yc_color]
		}
	} else {
		# try to load the color table from a TXT or a Muxman project file
		unset -nocomplain colors
		set fh [open $filename]
		while {1} {
			set line [string toupper [string trim [gets $fh] \t]]
			if {[eof $fh]} {
				break
			}
			if {[string first "COLOR" $line] == 0} {
				set line [string range $line 6 end]
				set line [string map {= " " , " "} $line]
				set ns [scan $line "%d %d %d %d" cn r g b]
				if {$ns != 4} continue
				set colors($cn) [format "#%02X%02X%02X" $r $g $b]
				if {$cn == 15} {
					set ok true
					for {set c 0} {$c < 16} {incr c} {
						if {![info exists colors($c)]} {
							set ok false
							break
						}
					}
					if {$ok} {
						close $fh
						break
					}
				}
			}
		}
		catch {close $fh}
		set ok true
		for {set c 0} {$c < 16} {incr c} {
			if {![info exists colors($c)]} {
				set ok false
				break
			}
		}
		if {$ok} {
			for {set c 0} {$c < 16} {incr c} {
				.pgceditor.f2.mfc.f1.b$c configure -background $colors($c)
			}
			set ::pgced::__color_table_modified true
		} else {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "File not of required type:\n$filename" -parent .pgceditor
		}
	}
}

set puos_labels {
	{Time play or search}
	{PTT play or search}
	{Title play}
	{Stop}
	{GoUp  (Not used in PGCs) }
	{Time or PTT search}
	{TopPG or PrevPG search}
	{NextPG search}
	{Forward scan}
	{Backward scan}
	{Menu call - Title}
	{Menu call - Root}
	{Menu call - Subpicture}
	{Menu call - Audio}
	{Menu call - Angle}
	{Menu call - PTT}
	{Resume}
	{Button select or activate}
	{Still off}
	{Pause on}
	{Audio stream change}
	{Subpicture stream change}
	{Angle change}
	{Karaoke audio mix change}
	{Video presentation mode change}
}

proc puos_editor {label parent} {
	# puos labels
	# note: GoUp (bit 4) is not used in PGC

	set w ".puoseditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::puosdone 0}
		bind $w <Escape> "set ::pgced::puosdone 0"
		wm transient $w $parent
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm title $w "Prohibited User Operations"

		label $w.lab -text $label
		pack  $w.lab -fill x -padx 4

		frame $w.f1
		pack  $w.f1 -fill x -padx 4

		set puon 0
		foreach p $::pgced::puos_labels {
			checkbutton $w.f1.puo$puon -text "$p ([expr {round(pow(2,$puon))}])" -anchor w -pady 0 -bd $::config(min_bd) -variable ::pgced::puo_$puon
			pack $w.f1.puo$puon -fill x -pady 0
			incr puon
		}
		#$w.f1.puo4 configure -state disabled	;# GoUp flag normally not used in PGC (?)
		frame  $w.fb
		pack   $w.fb -fill x -padx 2 -pady 4
		button $w.fb.clear -text "Clear all" -command {set ::pgced::tmppuos 0; ::pgced::set_puos_vars}
		pack   $w.fb.clear -side left -padx 5
		frame  $w.fb.pad
		pack   $w.fb.pad -side left -expand 1 -fill x
		button $w.fb.ok -text "OK" -width 6 -command {set ::pgced::puosdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 6 -command {set ::pgced::puosdone 0}
		pack   $w.fb.cancel -side left -padx 5

		update
		set wx [expr {[winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	set ::pgced::tmppuos $::pgced::puos
	#set ::pgced::tmppuos [expr {$::pgced::puos & 0X1FFFFEF}]	;# disable GoUp flag, not used in PGC (?)
	::pgced::set_puos_vars
	$w.lab configure -text $label

	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::puosdone -1
	tkwait variable ::pgced::puosdone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	wm withdraw $w
	if {$::pgced::puosdone} {
		set ::pgced::puos 0
		for {set p 0} {$p <= 24} {incr p} {
			set varname ::pgced::puo_$p
			if {[subst $$varname]} {
				set ::pgced::puos [expr {$::pgced::puos | round(pow(2,$p))}]
			}
		}
	}
	return $::pgced::puos
}

proc set_puos_vars {} {
	for {set p 0} {$p <= 24} {incr p} {
		if {[expr {$::pgced::tmppuos & round(pow(2,$p))}] == 0} {
			set ::pgced::puo_$p 0
		} else {
			set ::pgced::puo_$p 1
		}
	}
}

proc set_audio {streamvar streamnum} {
	set type [lindex $::editor_cmds_from 0]
	set numaudios [::streams::get_pgc_num_audio $type $::pgced::__index]
	if {$streamnum <= $numaudios} {
		set descr [::streams::get_pgc_audio_descr $type $::pgced::__index [expr {$streamnum-1}]]
		set ::pgced::startaudioid 0
		if {[string first (AC3)		  $descr]>=0} {set ::pgced::startaudioid [expr {(0x80)}]}
		if {[string first (MPEG)	  $descr]>=0} {set ::pgced::startaudioid [expr {(0xC0)}]}
		if {[string first (MPEG2-ext) $descr]>=0} {set ::pgced::startaudioid [expr {(0xC8)}]}
		if {[string first (LPCM)	  $descr]>=0} {set ::pgced::startaudioid [expr {(0xA0)}]}
		if {[string first (DTS)		  $descr]>=0} {set ::pgced::startaudioid [expr {(0x88)}]}
		if {[string first (SDDS)	  $descr]>=0} {set ::pgced::startaudioid [expr {(0x90)}]}
	} else {
		set descr "$streamnum: not used"
		set ::pgced::startaudioid 0
	}

	set w .audiosetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm transient $w .pgceditor
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::audiodone 0}
		bind $w <Escape> "set ::pgced::audiodone 0"
		wm title $w "Audio stream setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}

		label $w.lab -text "Audio $descr"
		pack  $w.lab -pady 4

		frame $w.f
		pack  $w.f
		label $w.f.l1 -text "VOB Audio stream number"
		label $w.f.l2 -text "ID"
		scale $w.f.s -from 0 -to 7 -variable "::pgced::audio_status" -length 170 -command {
				if {$::pgced::startaudioid == 0} {
					set ::pgced::audio_id "\n0x??"
				} else {
					set ::pgced::audio_id \n[format 0x%02X [expr {$::pgced::audio_status+$::pgced::startaudioid}]]
				}
				set tmp }
		label $w.f.id -textvariable ::pgced::audio_id -width 4 -anchor w
		grid  $w.f.l1 -row 0 -column 0
		grid  $w.f.l2 -row 0 -column 1
		grid  $w.f.s  -row 1 -column 0
		grid  $w.f.id -row 1 -column 1 -sticky se

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::audiodone 2}
		pack   $w.fb.dis -side left -padx 4
		::balloon::set_balloon $w.fb.dis \
				"When an audio stream is disabled,\nit will not show up anymore.\nSelecting this stream via a menu\nbutton will play stream 1!"
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::audiodone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::audiodone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr {[winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::audio_status 0
	} else {
		set ::pgced::audio_status [subst $$streamvar]
	}
	if {$::pgced::startaudioid == 0} {
		set ::pgced::audio_id "\n0x??"
	} else {
		set ::pgced::audio_id \n[format 0x%02X [expr {$::pgced::audio_status+$::pgced::startaudioid}]]
	}

	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::audiodone -1
	tkwait variable ::pgced::audiodone
	grab release $w
	grab set .pgceditor
	destroy $w
	switch $::pgced::audiodone {
		1 {set $streamvar $::pgced::audio_status}
		2 {set $streamvar "none"}
	}
	return
}

proc set_subpic {streamvar descr} {
	set w .subpicsetup
	if {! [winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm transient $w .pgceditor
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::subpicdone 0}
		bind $w <Escape> "set ::pgced::subpicdone 0"
		wm title $w "Subpic streams setup"
		wm resizable $w 0 0
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}

		label $w.lab -text "Subpic $descr"
		pack  $w.lab -pady 4

		frame $w.f
		pack  $w.f -padx 4

		label $w.f.tl0 -text "Video modes"
		label $w.f.tl1 -text "VOB Sub-picture stream numbers"
		label $w.f.tl2 -text "IDs"

		label $w.f.l1 -text "\n4:3"
		scale $w.f.s1 -from 0 -to 31 -variable "::pgced::subpic_1_status" -length 170 \
				-command {::pgced::validate_subpic_value 1}
		label $w.f.id1 -textvariable ::pgced::subpic_1_id -width 4 -anchor w
		label $w.f.l2 -text "\nwide"
		scale $w.f.s2 -from 0 -to 31 -variable "::pgced::subpic_2_status" \
				-command {::pgced::validate_subpic_value 2}
		label $w.f.id2 -textvariable ::pgced::subpic_2_id -width 4 -anchor w
		label $w.f.l3 -text "\nletterbox"
		scale $w.f.s3 -from 0 -to 31 -variable "::pgced::subpic_3_status" \
				-command {::pgced::validate_subpic_value 3}
		label $w.f.id3 -textvariable ::pgced::subpic_3_id -width 4 -anchor w
		label $w.f.l4 -text "\npan&scan"
		scale $w.f.s4 -from 0 -to 31 -variable "::pgced::subpic_4_status" \
				-command {::pgced::validate_subpic_value 4}
		label $w.f.id4 -textvariable ::pgced::subpic_4_id -width 4 -anchor w

		grid  $w.f.tl0 -row 0 -column 0 -sticky e
		grid  $w.f.tl1 -row 0 -column 1
		grid  $w.f.tl2 -row 0 -column 2
		grid  $w.f.l1  -row 1 -column 0 -sticky e
		grid  $w.f.s1  -row 1 -column 1 -sticky ew
		grid  $w.f.id1 -row 1 -column 2 -sticky sw
		grid  $w.f.l2  -row 2 -column 0 -sticky e
		grid  $w.f.s2  -row 2 -column 1 -sticky ew
		grid  $w.f.id2 -row 2 -column 2 -sticky sw
		grid  $w.f.l3  -row 3 -column 0 -sticky e
		grid  $w.f.s3  -row 3 -column 1 -sticky ew
		grid  $w.f.id3 -row 3 -column 2 -sticky sw
		grid  $w.f.l4  -row 4 -column 0 -sticky e
		grid  $w.f.s4  -row 4 -column 1 -sticky ew
		grid  $w.f.id4 -row 4 -column 2 -sticky sw

		frame  $w.fb
		pack   $w.fb -padx 2 -pady 4 -anchor e
		button $w.fb.dis -text "Disable" -width 7 -command {set ::pgced::subpicdone 2}
		pack   $w.fb.dis -side left -padx 4
		::balloon::set_balloon $w.fb.dis \
				"When a subpic stream is disabled,\nit will not show up anymore.\nSelecting this stream via a menu\nbutton will play stream 1!"
		button $w.fb.ok -text "Set" -width 7 -command {set ::pgced::subpicdone 1}
		pack   $w.fb.ok -side left
		button $w.fb.cancel -text "Cancel" -width 7 -command {set ::pgced::subpicdone 0}
		pack   $w.fb.cancel -side left -padx 4

		update
		set wx [expr {[winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}
	if {[subst $$streamvar] == "none"} {
		set ::pgced::subpic_1_status 0
		set ::pgced::subpic_2_status 0
		set ::pgced::subpic_3_status 0
		set ::pgced::subpic_4_status 0
	} else {
		set sublist [subst $$streamvar]
		set ::pgced::subpic_1_status [lindex $sublist 0]
		set ::pgced::subpic_2_status [lindex $sublist 1]
		set ::pgced::subpic_3_status [lindex $sublist 2]
		set ::pgced::subpic_4_status [lindex $sublist 3]
	}
	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::subpicdone -1
	tkwait variable ::pgced::subpicdone
	grab release $w
	grab set .pgceditor
	destroy $w
	switch $::pgced::subpicdone {
		1 {set $streamvar [string map {{ } {  }} [format "%2d %2d %2d %2d" $::pgced::subpic_1_status \
				$::pgced::subpic_2_status $::pgced::subpic_3_status $::pgced::subpic_4_status]]}
		2 {set $streamvar "none"}
	}
	return
}
proc validate_subpic_value {streamnum value} {
	set ::pgced::subpic_${streamnum}_id [format 0x%02X [expr {$value+32}]]
}

proc toggle_program {w cell} {
	set prog [$w.f2.fc.c.f.w_${cell}_1 cget -text]

	if {$prog == ""} {
		# create new program
		set newprog 0
		for {set c 1} {$c < $cell} {incr c} {
			set p [$w.f2.fc.c.f.w_${c}_1 cget -text]
			if {$p != ""} {
				set newprog $p
			}
		}
		incr newprog
		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set p [$w.f2.fc.c.f.w_${c}_1 cget -text]
			if {$p != ""} {
				if {$p >= $newprog} {
					incr p
					$w.f2.fc.c.f.w_${c}_1 configure -text $p
				}
			}
		}
		$w.f2.fc.c.f.w_${cell}_1 configure -text $newprog
		incr ::pgced::__numprograms
	} else {
		if {$::pgced::__numprograms == 1} {
			show_msg "Cannot remove the last Program!"
			bell
			return
		}
		# remove program
		$w.f2.fc.c.f.w_${cell}_1 configure -text ""
		set prog 1
		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set p [$w.f2.fc.c.f.w_${c}_1 cget -text]
			if {$p != ""} {
				$w.f2.fc.c.f.w_${c}_1 configure -text $prog
				incr prog
			}
		}
		incr ::pgced::__numprograms -1
	}

	# rebuild program map
	set ::pgced::__programmap {}
	for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
		set p [$w.f2.fc.c.f.w_${c}_1 cget -text]
		if {$p != ""} {
			set ::pgced::__programmap ${::pgced::__programmap}[d2c $c 1]
		}
	}

	# refresh PTT buttons
	if {$::pgced::__pgctype == "pgc"} {
		set error false
		foreach pttpair $::pgced::__PTTlist {
			set pg  [lindex $pttpair 0]
			set ptt [lindex $pttpair 1]
			if {! [info exists pg_ptt($pg)]} {
				set pg_ptt($pg) $ptt
				if {$pg > $::pgced::__numprograms} {
					set error true
				}
			}
		}

		for {set c 1} {$c <= $::pgced::__numcells} {incr c} {
			set p [$w.f2.fc.c.f.w_${c}_1 cget -text]
			if {$p != ""} {
				if {[info exists pg_ptt($p)]} {
					$w.f2.fc.c.f.w_${c}_0 configure -text $pg_ptt($p) -state normal
				} else {
					$w.f2.fc.c.f.w_${c}_0 configure -text "" -state normal
				}
			} else {
				$w.f2.fc.c.f.w_${c}_0 configure -text "" -state disabled
			}
		}

		if {$error} {
			.pgceditor.fb.ptt configure -foreground red
			if {! [info exists ::pgced::__warning_toggle_chapters]} {
				set txt "You must edit the chapter table to remove or reassign\nchapters pointing to removed programs!\n\nClick on the \"Edit Chapter (PTT) Table\" button.\n\n(This message will not be displayed anymore!)"
				show_warning -title "Toggle Program" -icon warning \
					-message $txt -parent .
				set ::pgced::__warning_toggle_chapters true
			}
		}
	}
	return ::pgced::__numprograms
}


namespace eval ::pgced::ptted {
proc ptt_table_editor {w} {
	set w .ptteditor
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Chapter (PTT) Table editor"
	wm transient $w .pgceditor
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::utils::pttedit_done 0}
	bind $w <Escape> "set ::utils::pttedit_done 0"

	frame $w.f
	pack  $w.f -padx 4 -pady 2 -fill both -expand 1

	listbox   $w.f.lb -exportselection 0 -selectmode extended \
			-listvariable ::pgced::ptted::lblist -width 33 -height 16 \
			-activestyle dotbox -yscrollcommand "$w.f.ys set" \
			-font $::config(fixed_width_font)
	scrollbar $w.f.ys -command "$w.f.lb yview"
	grid  $w.f.lb -column 0 -row 0 -sticky nesw
	grid  $w.f.ys -column 1 -row 0 -sticky ns
	grid columnconf $w.f 0 -weight 1000
	grid rowconf    $w.f 0 -weight 1000

	frame $w.b1
	pack  $w.b1 -padx 3 -pady 1 -fill x
	button $w.b1.badd -text "Add chapter" -width 20 -command {
				lappend ::pgced::ptted::ptts [expr {[llength $::pgced::ptted::ptts]+1}]
				lappend ::pgced::ptted::pgs  [expr {[lindex $::pgced::ptted::pgs end end]+1}]
				::pgced::ptted::refresh_listbox
			}
	pack  $w.b1.badd -fill x -padx 1 -side left -expand 1
	button $w.b1.incr -text "Increment programs" -width 20 -command {
				foreach e [.ptteditor.f.lb curselection] {
					lset ::pgced::ptted::pgs $e [expr {[lindex $::pgced::ptted::pgs $e]+1}]
				}
				::pgced::ptted::refresh_listbox
			}
	pack  $w.b1.incr -fill x -padx 1 -side left -expand 1
	frame $w.b2
	pack  $w.b2 -padx 3 -pady 1 -fill x
	button $w.b2.bdel -text "Delete last chapter" -width 20 -command {
				set ::pgced::ptted::ptts [lreplace $::pgced::ptted::ptts end end]
				set ::pgced::ptted::pgs  [lreplace $::pgced::ptted::pgs  end end]
				::pgced::ptted::refresh_listbox
			}
	pack  $w.b2.bdel -fill x -padx 1 -side left -expand 1
	button $w.b2.decr -text "Decrement programs" -width 20 -command {
				foreach e [.ptteditor.f.lb curselection] {
					set num [expr {[lindex $::pgced::ptted::pgs $e]-1}]
					if {$num > 0} {
						lset ::pgced::ptted::pgs $e [expr {[lindex $::pgced::ptted::pgs $e]-1}]
					}
				}
				::pgced::ptted::refresh_listbox
			}
	pack  $w.b2.decr -fill x -padx 1 -side left -expand 1

	button $w.eq -text "Set selected Programs = PTTs" -command {
				foreach e [.ptteditor.f.lb curselection] {
					lset ::pgced::ptted::pgs $e [expr {$e+1}]
				}
				::pgced::ptted::refresh_listbox
			}
	pack  $w.eq -fill x -padx 4 -pady 1

	frame $w.b3
	pack  $w.b3 -padx 3 -pady 1 -fill x
	label $w.b3.l -text "Set selected Programs to"
	pack  $w.b3.l -side left
	entry $w.b3.e -textvariable ::pgced::ptted::entry -width 3 \
			-validatecommand "::pgced::value_entry_validate 1 255 {%d} {%P}"
	pack  $w.b3.e -side left
	bind  $w.b3.e <Return> ::pgced::ptted::apply_pg_num
	frame $w.b3.pad -width 4
	pack  $w.b3.pad -side left
	button $w.b3.b -text "Apply" -width 6 -command ::pgced::ptted::apply_pg_num
	pack  $w.b3.b -fill x -padx 1 -side left -expand 1

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::utils::pttedit_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::utils::pttedit_done 0}
	pack   $w.fb.cancel -side left -padx 2

	set ::pgced::ptted::ptts {}
	set ::pgced::ptted::pgs {}
	foreach pttpair $::pgced::__PTTlist {
		lappend ::pgced::ptted::ptts [lindex $pttpair 1]
		lappend ::pgced::ptted::pgs  [lindex $pttpair 0]
	}
	refresh_listbox

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	focus $w
	grab set $w

	tkwait variable ::utils::pttedit_done

	unset ::pgced::ptted::lblist
	destroy $w
	grab set .pgceditor
	if {$::utils::pttedit_done} {
		if {$::pgced::__PTTlist != ""} {
			set ttu [lindex [lindex $::pgced::__PTTlist 0] 2]
		} else {
			set ttu $::pgcs($::pgced::__index,TTU)
		}
		set ::pgced::__PTTlist {}
		set error false
		for {set e 0} {$e < [llength $::pgced::ptted::ptts]} {incr e} {
			set pg  [lindex $::pgced::ptted::pgs $e]
			set ptt [lindex $::pgced::ptted::ptts $e]
			lappend ::pgced::__PTTlist [list $pg $ptt $ttu]
			if {$pg > [string length $::pgced::__programmap]} {
				set error true
			}
		}
		if {$error} {
			.pgceditor.fb.ptt configure -foreground red
			set txt "You must remove or reassign the chapters\npointing to out of range programs!\n\nClick on the \"Edit Chapter (PTT) Table\" button."
			show_warning -title "Remove Cell" -icon warning \
				-message $txt -parent .
		} else {
			.pgceditor.fb.ptt configure -foreground black
		}
		return true
	} else {
		return false
	}
}
proc refresh_listbox {} {
	set sel [.ptteditor.f.lb curselection]
	set ::pgced::ptted::lblist {}
	set idx 0
	foreach ptt $::pgced::ptted::ptts {
		lappend ::pgced::ptted::lblist " Chapter (PTT) [format %2d $ptt] : Program [format %3d [lindex $::pgced::ptted::pgs $idx]] "
		incr idx
	}
	foreach s $sel {
		.ptteditor.f.lb selection set $s
	}
}
proc apply_pg_num {} {
	if {$::pgced::ptted::entry == "" || $::pgced::ptted::entry == 0} {
		bell
	} else {
		foreach e [.ptteditor.f.lb curselection] {
			lset ::pgced::ptted::pgs $e $::pgced::ptted::entry
		}
		refresh_listbox
	}
}
}

proc cell_type_editor {parentbutton layerbreakvar} {

	set w ".celltypeeditor"
	if {![winfo exists $w]} {
		toplevel $w -class Dialog
		wm withdraw $w
		wm protocol $w WM_DELETE_WINDOW {set ::pgced::celltypedone 0}
		bind $w <Escape> "set ::pgced::celltypedone 0"
		wm transient $w .pgceditor
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm title $w "Cell Type Flags"

		frame $w.f1
		pack  $w.f1 -fill x -padx 4 -pady 4

		checkbutton $w.f1.c0 -text "Seamless angle (1)" -anchor w \
				-bd $::config(min_bd) -variable ::pgced::celltype_0 -offvalue 0 -onvalue 1
		pack $w.f1.c0 -fill x
		checkbutton $w.f1.c1 -text "STC (SCR) discontinuity (2)" -anchor w \
				-bd $::config(min_bd) -variable ::pgced::celltype_1 -offvalue 0 -onvalue 2
		pack $w.f1.c1 -fill x
		checkbutton $w.f1.c2 -text "Interleaved (4)" -anchor w \
				-bd $::config(min_bd) -variable ::pgced::celltype_2 -offvalue 0 -onvalue 4
		pack $w.f1.c2 -fill x
		checkbutton $w.f1.c3 -text "Seamless joint (8)" -anchor w \
				-bd $::config(min_bd) -variable ::pgced::celltype_3 -offvalue 0 -onvalue 8
		pack $w.f1.c3 -fill x
		checkbutton $w.f1.c4 -text "Angle block (16)" -anchor w \
				-bd $::config(min_bd) -variable ::pgced::celltype_4 -offvalue 0 -onvalue 16 \
				-command "::pgced::validate_cell_type_editor $w"
		pack $w.f1.c4 -fill x

		frame $w.ft -borderwidth 2 -relief groove
		pack  $w.ft -fill x -padx 4

		label $w.ft.lab -text "Angle type:" -anchor w
		pack  $w.ft.lab -fill x
		radiobutton $w.ft.r0 -variable ::pgced::celltype_67 -value 0 -anchor w \
				-text "normal (0)" -pady 0 -bd $::config(min_bd) -state disabled
		pack  $w.ft.r0 -fill x
		radiobutton $w.ft.r1 -variable ::pgced::celltype_67 -value 64 -anchor w \
				-text "first of angle block (64)" -pady 0 -bd $::config(min_bd)
		pack  $w.ft.r1 -fill x
		radiobutton $w.ft.r2 -variable ::pgced::celltype_67 -value 128 -anchor w \
				-text "middle of angle block (128)" -pady 0 -bd $::config(min_bd)
		pack  $w.ft.r2 -fill x
		radiobutton $w.ft.r3 -variable ::pgced::celltype_67 -value 192 -anchor w \
				-text "last of angle block (192)" -pady 0 -bd $::config(min_bd)
		pack  $w.ft.r3 -fill x

		frame  $w.fb
		pack   $w.fb -anchor e -padx 4 -pady 6
		button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::celltypedone 1}
		pack   $w.fb.ok -side left -padx 4
		button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::celltypedone 0}
		pack   $w.fb.cancel -side left

		update
		set wx [expr {[winfo rootx .pgceditor]+(([winfo width  .pgceditor]-[winfo reqwidth  $w])/2)}]
		set wy [expr {[winfo rooty .pgceditor]+(([winfo height .pgceditor]-[winfo reqheight $w])/2)-25}]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
		wm resizable $w 0 0
	}

	set ct [$parentbutton cget -text]
	set ::pgced::celltype_0  [expr {$ct & 1}]
	set ::pgced::celltype_1  [expr {$ct & 2}]
	set ::pgced::celltype_2  [expr {$ct & 4}]
	set ::pgced::celltype_3  [expr {$ct & 8}]
	set ::pgced::celltype_4  [expr {$ct & 16}]
	set ::pgced::celltype_67 [expr {$ct & 192}]

	validate_cell_type_editor $w

	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::celltypedone -1
	tkwait variable ::pgced::celltypedone
	grab release $w
	grab set .pgceditor
	wm withdraw $w
	if {$::pgced::celltypedone} {
		set  r $::pgced::celltype_0
		incr r $::pgced::celltype_1
		incr r $::pgced::celltype_2
		incr r $::pgced::celltype_3
		incr r $::pgced::celltype_4
		incr r $::pgced::celltype_67
		$parentbutton configure -text $r
		if {$::pgced::celltype_3 == 8} {
			set $layerbreakvar 0
		} else {
			set $layerbreakvar 1
		}

		return $r
	} else {
		return ""
	}
}
proc validate_cell_type_editor {w} {
	if {$::pgced::celltype_4} {
		set state normal
	} else {
		set state disabled
		set ::pgced::celltype_67 0
	}
	$w.ft.r1 configure -state $state
	$w.ft.r2 configure -state $state
	$w.ft.r3 configure -state $state
}

proc set_layer_break {breakflag typebutton} {
	set typeval [$typebutton cget -text]
	if {[subst $$breakflag]} {
		set typeval [expr {$typeval & 0XF7}]
	} else {
		set typeval [expr {$typeval | 8}]
	}
	$typebutton configure -text $typeval
}


proc reassign_VOB_ID {parent cur_cellnum cur_vobid cur_cellid} {
	set cur_type  [lindex $::editor_cmds_from 0]
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]

	# Search for all VOB IDs in the current VTS, and for all Cell IDs in each VOB IDs
	# and build the list of vobids, and the lists of cellids for all vobids.
	# Also, for each pair VOB/Cell ID, keep track of the belonging PGC's cell.
	set all_vobids ""
	if {$cur_type == "menu"} {
		for {set lu 1} {$lu <= $::menus($cur_vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($cur_vts,$lu,numpgcs)} {incr pgc} {
				set cellpostable $::menus($cur_vts,$lu,$pgc,cellpostable)
				for {set cell 1} {$cell <= $::menus($cur_vts,$lu,$pgc,numcells)} {incr cell} {
					set cellidx [expr {($cell - 1) * 4}]
					set vobid  [get_int cellpostable $cellidx 2]
					set cellid [get_int cellpostable [expr {$cellidx + 3}] 1]
					if {[lsearch $all_vobids $vobid] == -1} {
						lappend all_vobids $vobid
						set all_cellids($vobid) ""
					}
					if {[lsearch $all_cellids($vobid) $cellid] == -1} {
						lappend all_cellids($vobid) $cellid
					}
					set src_pgc_cell($vobid,$cellid) [list "menu" $lu $pgc $cell]
				}
			}
		}
	} else {
		for {set pgc 1} {$pgc <= $::pgcs($cur_vts,numpgcs)} {incr pgc} {
			set cellpostable $::pgcs($cur_vts,$pgc,cellpostable)
			for {set cell 1} {$cell <= $::pgcs($cur_vts,$pgc,numcells)} {incr cell} {
				set cellidx [expr {($cell - 1) * 4}]
				set vobid  [get_int cellpostable $cellidx 2]
				set cellid [get_int cellpostable [expr {$cellidx + 3}] 1]
				if {[lsearch $all_vobids $vobid] == -1} {
					lappend all_vobids $vobid
					set all_cellids($vobid) ""
				}
				if {[lsearch $all_cellids($vobid) $cellid] == -1} {
					lappend all_cellids($vobid) $cellid
				}
				set src_pgc_cell($vobid,$cellid) [list "pgc" 0 $pgc $cell]
			}
		}
	}
	set all_vobids [lsort -integer $all_vobids]
	foreach vobid $all_vobids {
		set all_cellids($vobid) [lsort -integer $all_cellids($vobid)]
	}

	# GUI
	set ::pgced::reassign_cell_mode 1	;# 1=use existing cell, 2=create new blank cell

	set w ".assignvobid"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::vobiddone 0}
	bind $w <Escape> "set ::pgced::vobiddone 0"
	wm transient $w $parent
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm title $w "Assign VOB/Cell ID for PGC Cell $cur_cellnum"

	set txt "Assing another VOB ID and/or Cell ID to Cell $cur_cellnum of the"
	set txt "$txt\ncurrent PGC.\n"
	set txt "$txt\nWARNING: The old VOB/Cell ID will be unavailable"
	set txt "$txt\nin the future, unless it is reused in another PGC !"
	if {$cur_type == "pgc"} {
		set txt "$txt\n(Hint: Use the \"New 'Play All' Title\" utility to 'backup'"
		set txt "$txt\nthe old VOB Cell references before using this function.)"
	}
	label $w.lab -text $txt -justify left -anchor w
	pack  $w.lab -fill x -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -fill x -padx 4
	radiobutton $w.f1.rb -text "Pick an existing VOB/Cell ID:" \
			-variable ::pgced::reassign_cell_mode -value 1 -anchor w \
			-command "::pgced::validate_reassign_gui $w"
	pack  $w.f1.rb -fill x
	frame $w.f1.l1
	pack  $w.f1.l1 -fill x -padx 5
	label $w.f1.l1.lab -text "New VOB ID:" -width 17 -anchor w
	pack  $w.f1.l1.lab -side left
	entry $w.f1.l1.e -width 4 -state readonly -justify right -bd 0 -relief flat \
			-textvariable ::pgced::new_vobid
	pack  $w.f1.l1.e -side left
	scale $w.f1.s1 -showvalue false -from 0 -to [expr {[llength $all_vobids]-1}] \
			-orient horizontal -variable ::pgced::vobid_scale \
			-command {::pgced::validate_vobid_scale .assignvobid $all_vobids all_cellids}
	pack  $w.f1.s1 -padx 1 -fill x

	frame $w.f1.l2
	pack  $w.f1.l2 -fill x -padx 5
	label $w.f1.l2.lab -text "New Cell ID:" -width 17 -anchor w
	pack  $w.f1.l2.lab -side left
	entry $w.f1.l2.e -width 3 -state readonly -justify right -bd 0 -relief flat \
			-textvariable ::pgced::new_cellid
	pack  $w.f1.l2.e -side left
	scale $w.f1.s2 -showvalue false -from 0 -to [expr {[llength $all_cellids($cur_vobid)]-1}] \
			-orient horizontal -variable ::pgced::cellid_scale \
			-command "::pgced::validate_cellid_scale \{$all_cellids($cur_vobid)\}"
	pack  $w.f1.s2 -padx 1 -fill x
	label $w.f1.lab -text "Note: The VOB file is not modified by this operation." -anchor w
	pack  $w.f1.lab -fill x -padx 2

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -fill x -padx 4 -pady 4
	radiobutton $w.f2.rb -text "Create a new blank VOB Cell:" \
			-variable ::pgced::reassign_cell_mode -value 2 -anchor w \
			-command "::pgced::validate_reassign_gui $w"
	pack  $w.f2.rb -fill x
	set txt "A new blank cell will be appended to the VOB file"
	set txt "$txt\nand its VOB ID and Cell ID will be assigned to"
	set txt "$txt\nCell $cur_cellnum of the current PGC."
	label $w.f2.lab -text $txt -justify left -anchor w
	pack  $w.f2.lab -fill x -padx 4
	checkbutton $w.f2.cb -text "Copy original VOB to Backup folder" \
			-variable ::pgced::copy_vob_to_backup -anchor w
	pack  $w.f2.cb -fill x -padx 2
	set ::pgced::copy_vob_to_backup 1
	set txt "WARNING: The VOB file will be modified immediately."
	set txt "$txt\nTHE DVD WILL BE SAVED after this operation,"
	set txt "$txt\nto keep track of the new VOB cell in the IFOs!"
	label $w.f2.lab2 -text $txt -justify left -anchor w
	pack  $w.f2.lab2 -fill x -padx 4

	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::vobiddone 1}
	pack   $w.fb.ok -side left -padx 4
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::vobiddone 0}
	pack   $w.fb.cancel -side left

	update
	set wx [expr {[winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm resizable $w 0 0

	validate_reassign_gui $w

	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::vobid_scale  [lsearch $all_vobids $cur_vobid]
	set ::pgced::cellid_scale [lsearch $all_cellids($cur_vobid) $cur_cellid]

	set ::pgced::vobiddone -1
	tkwait variable ::pgced::vobiddone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	wm withdraw $w
	if {!$::pgced::vobiddone} return
	set new_vobid  $::pgced::new_vobid
	set new_cellid $::pgced::new_cellid
	set new_lba [::blankcell::numSectorsInVOB $cur_type $cur_vts]
	if {$::pgced::reassign_cell_mode == 1 && $new_vobid == $cur_vobid && $new_cellid == $cur_cellid} return

	if {$::pgced::reassign_cell_mode == 1} {
		# Use the new VOB ID $new_vobid and the new Cell ID $new_cellid.
		set src $src_pgc_cell($new_vobid,$new_cellid)
		set src_type [lindex $src 0]
		set src_lu   [lindex $src 1]
		set src_pgc  [lindex $src 2]
		set src_cell [lindex $src 3]

		# get the relevant entry in the source cell playback table
		if {$src_type == "menu"} {
			set src_index "$cur_vts,$src_lu,$src_pgc"
		} else {
			set src_index "$cur_vts,$src_pgc"
		}
		set src_cellpbtable [subst \$::${src_type}s($src_index,cellpbtable)]
		set src_cellpbidx   [expr {($src_cell - 1) * 24}]
		set src_cellpbentry [string range $src_cellpbtable $src_cellpbidx [expr {$src_cellpbidx + 23}]]

		if {$cur_type == "menu"} {
			incr ::menus($cur_index,numbuttons) \
					[expr {$::buttons::butsinfo(menu,$cur_vts,$cur_vobid,$cur_cellid,numbuts) * -1}]
			incr ::menus($cur_index,numbuttons) \
					$::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,numbuts)
		} elseif {[info exists ::pgcs($cur_index,numbuttons)]} {
			if {[info exists ::buttons::butsinfo(pgc,$cur_vts,$cur_vobid,$cur_cellid,numbuts)]} {
				incr ::pgcs($cur_index,numbuttons) \
						[expr {$::buttons::butsinfo(pgc,$cur_vts,$cur_vobid,$cur_cellid,numbuts) * -1}]
			}
			if {[info exists ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,numbuts)]} {
				incr ::pgcs($cur_index,numbuttons) \
						$::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,numbuts)
			}
		}
	} else {
		# Append a blank cell to the VOB
		set resultlist [::blankcell::AppendBlankCell $cur_type $cur_vts $::pgced::copy_vob_to_backup]

		set new_vobid [lindex $resultlist 0]
		set src_cellpbentry [lindex $resultlist 1]
		if {$new_vobid == 0} return
		if {$new_vobid < 0} {
			set txt "Error appending new Cell to VOB file:\n"
			set txt "$txt\n$src_cellpbentry"
			tk_messageBox -title "Append Blank Cell to VOB" -icon error \
					-message $txt -parent .
			return
		}
		set new_cellid 1

		if {$cur_type == "menu"} {
			incr ::menus($cur_index,numbuttons) \
					[expr {$::buttons::butsinfo(menu,$cur_vts,$cur_vobid,$cur_cellid,numbuts) * -1}]
			set ::buttons::butsinfo(menu,$cur_vts,numvobids) $new_vobid
			set ::buttons::butsinfo(menu,$cur_vts,$::buttons::butsinfo(menu,$cur_vts,numvobids),numcellids) 1
			set ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,navpack) $new_lba
			set ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,numbuts) 0
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,numgroups)
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,forceselectbutn)
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,forceactionbutn)
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,1,grouptype)
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,2,grouptype)
			unset -nocomplain ::buttons::butsinfo(menu,$cur_vts,$new_vobid,$new_cellid,3,grouptype)
		} else {
			if {[info exists ::pgcs($cur_index,numbuttons)]} {
				incr ::pgcs($cur_index,numbuttons) \
						[expr {$::buttons::butsinfo(pgc,$cur_vts,$cur_vobid,$cur_cellid,numbuts) * -1}]
			}
			set ::buttons::butsinfo(pgc,$cur_vts,numvobids) $new_vobid
			set ::buttons::butsinfo(pgc,$cur_vts,$::buttons::butsinfo(pgc,$cur_vts,numvobids),numcellids) 1
			set ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,navpack) $new_lba
			set ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,numbuts) 0
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,numgroups)
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,forceselectbutn)
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,forceactionbutn)
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,1,grouptype)
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,2,grouptype)
			unset -nocomplain ::buttons::butsinfo(pgc,$cur_vts,$new_vobid,$new_cellid,3,grouptype)
		}
	}
	# copy the cell type and the cell command number from the original cell
	set cur_cellpbidx [expr {($cur_cellnum - 1) * 24}]
	set cur_celltype [expr {([get_int ::pgced::__cellpbtable $cur_cellpbidx 1] & 0xF7) | 2}] ;# add SRC Discontinuity flag, and remove Seamless Joint
	set cur_cmdnum [get_int ::pgced::__cellpbtable [expr {$cur_cellpbidx + 3}] 1]
	set_int src_cellpbentry 0 $cur_celltype 1
	set_int src_cellpbentry 3 $cur_cmdnum 1

	# replace the new entry in the current cell playback table
	set ::pgced::__cellpbtable "[string range $::pgced::__cellpbtable 0 [expr {$cur_cellpbidx-1}]]$src_cellpbentry[string range $::pgced::__cellpbtable [expr {$cur_cellpbidx+24}] end]"

	# set the VOB ID and Cell ID in the Cell Position table
	set cur_cellposidx [expr {($cur_cellnum - 1) * 4}]
	set_int ::pgced::__cellpostable $cur_cellposidx $new_vobid 2
	set_int ::pgced::__cellpostable [expr {$cur_cellposidx + 3}] $new_cellid 1

	# Set the STC Discontinuity flag in the modified Cell entry, and clear the Seamless flag
	set idx [expr {($cur_cellnum-1) * 24}]
	set_int ::pgced::__cellpbtable $idx [expr {([get_int ::pgced::__cellpbtable $idx 1] | 2) & 0xF7}] 1
	# Set the STC Discontinuity flag in the NEXT Cell entry, and clear the Seamless flag
	if {$::pgced::__numcells > $cur_cellnum} {
		set idx [expr {$cur_cellnum * 24}]
		set_int ::pgced::__cellpbtable $idx [expr {([get_int ::pgced::__cellpbtable $idx 1] | 2) & 0xF7}] 1
	}

	# reload the new tables in the editor
	set total_time [build_cell_table $parent]

	# correct the total playback time
	set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$total_time[string range $::pgced::__pgcheader 8 end]"
	set ::pgced::pbtimetxt [BCDtime $total_time true]
	set ::pgced::__pgcpbtime $total_time

	::main::dvd_modified true false

	# If cell replaced, then we must remember the current PGC tables,
	# the simplest way to do this is to close the PGC Editor window, with an OK signal.
	if {$::pgced::reassign_cell_mode == 2} {
		set ::pgced::done 1
		after 500 {
			tk_messageBox -title "Assign VOB/Cell ID" -icon warning \
					-message "The DVD will be saved now..." -parent .
			::main::Write_DVD false
		}
	}

	# refresh PGC selector to show the new number of menubuttons
	::main::init_tree_menu_no_move
	return
}
proc validate_reassign_gui {w} {
	if {$::pgced::reassign_cell_mode == 1} {
		::ced::scale_state $w.f1.s1 normal
		::ced::scale_state $w.f1.s2 normal
		$w.f2.cb configure -state disabled
	} else {
		::ced::scale_state $w.f1.s1 disabled
		::ced::scale_state $w.f1.s2 disabled
		$w.f2.cb configure -state normal
	}
}
proc validate_vobid_scale {w all_vobids all_cellids_upvar val} {
	upvar $all_cellids_upvar all_cellids
	set ::pgced::new_vobid [lindex $all_vobids $val]
	$w.f1.s2 configure -to [expr {[llength $all_cellids($::pgced::new_vobid)]-1}] \
			-command "::pgced::validate_cellid_scale \{$all_cellids($::pgced::new_vobid)\}"
	set pos [lsearch $all_cellids($::pgced::new_vobid) $::pgced::new_cellid]
	if {$pos == -1} {
		$w.f1.s2 set 0
		set ::pgced::new_cellid [lindex $all_cellids($::pgced::new_vobid) 0]
	} else {
		$w.f1.s2 set $pos
	}
}
proc validate_cellid_scale {all_cellids val} {
	set ::pgced::new_cellid [lindex $all_cellids $val]
}

proc create_playall_pgc {} {
	if {![info exists ::dvddir]} {bell; return}

	set cur_type  [lindex $::editor_cmds_from 0]
	if {$cur_type != "pgc"} {bell; return}
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]
	if {$cur_vts == 0} {bell; return}
	set cur_pgc   [lindex [split $cur_index ","] 1]

	if {! $::config(expert_mode)} {
		show_msg ""
		set txt "This utility will create a new Title (Title [expr {[get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]+1}]) with all the\n"
		set txt "${txt}VOB/Cell IDs of the current VTS (VTS $cur_vts).\n\n"
		set txt "${txt}The streams status and the color lookup table are\n"
		set txt "${txt}copied from the current PGC (PGC $cur_pgc).\n\n"
		set txt "${txt}A chapter will be created for each VOB/Cell ID.\n\n"
		set txt "${txt}This new Title will have the pre-commands of the\n"
		set txt "${txt}first PGC and the post-commands of the last PGC\n"
		set txt "${txt}of the current title domain.\n\n"
		set txt "${txt}Continue?"
		set r [tk_messageBox -title "New 'Play All' Title" -icon question \
				-message $txt -parent . -type okcancel -default ok]
		if {$r == "cancel"} return
	}

	#if {$::pgcs($cur_vts,numpgcs) < 2} {
	#	set txt "This utility needs two or more PGCs\n"
	#	set txt "${txt}in the current VTS to be useful!\n"
	#	set txt "${txt}(However, you may want to use this utility\n"
	#	set txt "${txt}to create a 'backup' of all VOB/Cell IDs\n"
	#	set txt "${txt}before remapping the IDs in the PGC Editor.)\n\n"
	#	set txt "${txt}There is only one PGC in this VTS.\nContinue anyway?"
	#	set r [tk_messageBox -title "New 'Play All' Title" -icon question \
	#			-message $txt -parent . -type okcancel -default ok]
	#	if {$r == "cancel"} return
	#}

	show_msg "Creating New PGC $::pgcs($cur_vts,numpgcs)..."

	# Search for all VOB IDs in the current VTS, and for all Cell IDs in each VOB IDs
	# and build the list of vobids, and the lists of cellids for all vobids.
	# Also, for each pair VOB/Cell ID, keep track of the belonging PGC's cell.
	set totcells 0
	set all_vobids ""
	for {set pgc 1} {$pgc <= $::pgcs($cur_vts,numpgcs)} {incr pgc} {
		set cellpostable $::pgcs($cur_vts,$pgc,cellpostable)
		for {set cell 1} {$cell <= $::pgcs($cur_vts,$pgc,numcells)} {incr cell} {
			incr totcells
			set cellidx [expr {($cell - 1) * 4}]
			set vobid  [get_int cellpostable $cellidx 2]
			set cellid [get_int cellpostable [expr {$cellidx + 3}] 1]
			if {[lsearch $all_vobids $vobid] == -1} {
				lappend all_vobids $vobid
				set all_cellids($vobid) ""
			}
			if {[lsearch $all_cellids($vobid) $cellid] == -1} {
				lappend all_cellids($vobid) $cellid
			}
			set src_pgc_cell($vobid,$cellid) [list $pgc $cell]
		}
	}
	if {$totcells > 255} {
		set txt "You have $totcells cells in this VTS.\n"
		set txt "${txt}This is too much for one PGC: the DVD\n"
		set txt "${txt}standard limit the number of cells per\n"
		set txt "${txt}PGC to 255.\n\n"
		set txt "${txt}Do you want ot create a title with only\n"
		set txt "${txt}the first 255 cells?"
		set r [tk_messageBox -title "New 'Play All' Title" -icon error \
				-message $txt -parent . -type okcancel -default cancel]
		if {$r == "cancel"} return
	}
	set all_vobids [lsort -integer $all_vobids]
	foreach vobid $all_vobids {
		set all_cellids($vobid) [lsort -integer $all_cellids($vobid)]
	}

	# Begin the job!
	#set ::pgced::__numprograms 0
	set numcells    0
	set header $::pgcs($cur_vts,[lindex [split $cur_index ","] 1],header)
	set programmap ""
	set cellpbtable ""
	set cellpostable ""
	set old_vobid 0
	set old_cellid 0
	set totalpbtime "\000\000\000\000"
	foreach vobid $all_vobids {
		if {$numcells == 256} break
		foreach cellid $all_cellids($vobid) {
			incr numcells
			if {$numcells == 256} break
			set thispgc  [lindex $src_pgc_cell($vobid,$cellid) 0]
			set thiscell [lindex $src_pgc_cell($vobid,$cellid) 1]
			set thiscellpbtable $::pgcs($cur_vts,$thispgc,cellpbtable)
			set idx [expr {($thiscell-1) * 0x18}]
			set thiscellpbentry [string range $thiscellpbtable $idx [expr {$idx+0x17}]]
			# fix playback mode
			if {$cellpostable == ""} {
				set mode 2
			} else {
				if {$old_vobid == $vobid && [expr {$old_cellid+1}] == $cellid} {
					set mode 8
				} else {
					set mode 10
				}
			}
			set oldmode [expr {[get_int thiscellpbentry 0 1] & 0xF5}]
			set_int thiscellpbentry 0 [expr {$oldmode | $mode}] 1
			# set cell still time to a max of 3 seconds
			if {[get_int thiscellpbentry 2 1] > 3} {
				set_int thiscellpbentry 2 3 1
			}
			# clear cell command
			set_int thiscellpbentry 3 0 1

			# Add cell time to total PGC time, but only if angle 1.
			set type [get_int thiscellpbtable [expr {($thiscell - 1) * 24}] 1]
			if {[expr {$type & 0x10}] == 0 || [expr {$type & 0xC0}] == 0x40} {
				set totalpbtime [AddBCDtime [string range $thiscellpbentry 4 7] $totalpbtime]
			}

			# add this cell infos to the new cell playback table
			set cellpbtable $cellpbtable$thiscellpbentry

			# add this cell position to the new cell position table
			set idx [string length $cellpostable]
			set cellpostable $cellpostable\000\000\000\000
			set_int cellpostable $idx $vobid 2
			incr idx 3
			set_int cellpostable $idx $cellid 1

			# create a new program for each cell
			set tmp \000
			set_int tmp 0 $numcells 1
			set programmap $programmap$tmp

			set old_cellid $cellid
			set old_vobid $vobid
		}
	}
	if {[expr {[string length $programmap] % 2}] == 1} {
		set programmap $programmap\000
	}
	if {$numcells > 255} {set numcells 255}

	set newpgcn [expr {$::pgcs($cur_vts,numpgcs)+1}]

	# fix PGC header attributes
	set_int header 2 $numcells 1	;# number of programs = total number of cells
	set_int header 3 $numcells 1	;# number of cells = total number of cells
	set_int header 4 [get_int totalpbtime 0 4] 4	;# total Playback time
	set_int header 8 0 4	;# No PUOs
	set_int header 0x9C $newpgcn 2	;# NextPGCN
	set_int header 0x9E $newpgcn 2	;# PrevPGCN
	set_int header 0xA0 0 2	;# GoUpPGCN
	set_int header 0xA3 0 1	;# PGC Still Time = 0
	set_int header 0xA2 0 1	;# PB Mode = sequential

	# take track of the new changes
	set precmds  [get_type_cmds pre  ::pgcs($cur_vts,1)]
	set postcmds [get_type_cmds post ::pgcs($cur_vts,$::pgcs($cur_vts,numpgcs))]
	incr ::pgcs($cur_vts,numpgcs)
	set ::pgcs($cur_vts,$newpgcn) \
			[concat {{* pre}} $precmds {{* post}} $postcmds {{* cell}}]
	set ::pgcs($cur_vts,$newpgcn,header) $header
	set ::pgcs($cur_vts,$newpgcn,numprograms) $numcells
	set ::pgcs($cur_vts,$newpgcn,numcells) $numcells
	set ::pgcs($cur_vts,$newpgcn,cellpbtable) $cellpbtable
	set ::pgcs($cur_vts,$newpgcn,cellpostable) $cellpostable
	set ::pgcs($cur_vts,$newpgcn,programmap) $programmap
	set ttn 0
	for {set i 1} {$i < $newpgcn} {incr i} {
		if {[info exists ::pgcs($cur_vts,$i,TTN)] && $::pgcs($cur_vts,$i,TTN) > $ttn} {
			set ttn $::pgcs($cur_vts,$i,TTN)
		}
	}
	incr ttn
	set ::pgcs($cur_vts,$newpgcn,TTN) $ttn

	# update VTS_PGCI table
	set idx [expr {($newpgcn * 8)}]
	set part1 [string range $::raw_sectors($cur_vts,VTS_PGCI) 0 [expr {$idx-1}]]
	set part2 [string range $::raw_sectors($cur_vts,VTS_PGCI) $idx end]
	set ::raw_sectors($cur_vts,VTS_PGCI) $part1[string repeat \000 8]$part2
	set_int ::raw_sectors($cur_vts,VTS_PGCI) $idx [expr {$ttn+128}] 1

	### build chapter table
	set ptttable $::raw_sectors($cur_vts,VTS_PTT_SRPT)
	set pttnumtitles [get_int ptttable 0 2]
	set newtitle [expr {$pttnumtitles+1}]
	set pttheader [string range $ptttable 0 [expr {($pttnumtitles*4) + 7}]]
	set pttrest [string range $ptttable [expr {($pttnumtitles*4) + 8}] end]
	# add entry to the PTT table header
	set idx [string length $pttheader]
	set pttheader $pttheader\000\000\000\000
	set_int pttheader $idx [expr {[string length $pttheader]+[string length $pttrest]}] 4
	for {set i 0} {$i < $pttnumtitles} {incr i} {
		set idx [expr {($i * 4)+8}]
		set_int pttheader $idx [expr {[get_int pttheader $idx 4] +4}] 4
	}
	# add a chapter for each cell
	set pttlist ""
	for {set ptt 1} {$ptt <= $numcells} {incr ptt} {
		set tmp \000\000\000\000
		set_int tmp 0 $newpgcn 2
		set_int tmp 2 $ptt 2
		set pttrest $pttrest$tmp
		lappend pttlist "$ptt $ptt $newtitle"
	}
	set ::pgcs($cur_vts,$newpgcn,pttlist) $pttlist
	set ::pgcs($cur_vts,numttus) $newtitle
	set ptttable $pttheader$pttrest
	set_int ptttable 0 [expr {$pttnumtitles+1}] 2
	set_int ptttable 4 [expr {[string length $ptttable]-1}] 4
	set ::raw_sectors($cur_vts,VTS_PTT_SRPT) $ptttable

	# Add the new title to the VMG global table of titles
	set data $::raw_sectors(0,VMG_TT_SRPT)
	set srptnumtitles [get_int data 0 2]
	set data [string range $data 0 [expr {($srptnumtitles*12)+7}]]
	incr srptnumtitles
	set_int data 0 $srptnumtitles 2
	set newentry \024\001\000\000\000\000\000\000\000\000\000\000
	set_int newentry 2 $numcells 2
	set_int newentry 6 $cur_vts 1
	set_int newentry 7 $::pgcs($cur_vts,$newpgcn,TTN) 1
	set data $data$newentry
	set_int data 4 [expr {[string length $data]-1}] 4
	set ::raw_sectors(0,VMG_TT_SRPT) $data
	set ::pgcs($cur_vts,$newpgcn,Title) $srptnumtitles

	# refresh GUI
	::main::dvd_modified true false
	::main::init_tree_menu
	::main::display_pgc pgc $cur_vts $newpgcn
	show_msg "New PGC $newpgcn created OK."
	return
}

proc remove_cell {parent} {
	if {$::pgced::__numcells <= 1} {
		show_msg "Cannot remove the last cell!"
		bell
		return
	}
	set cur_type  [lindex $::editor_cmds_from 0]
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]

	# GUI
	set w ".delcell"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::pgced::delcelldone 0}
	bind $w <Escape> "set ::pgced::delcelldone 0"
	wm transient $w $parent
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm title $w "Remove Cell"

	set txt "WARNING: The old VOB/Cell ID of the removed cell"
	set txt "$txt\nwill be unavailable in the future, unless it is reused"
	set txt "$txt\nin another PGC !"
	if {$cur_type == "pgc"} {
		set txt "$txt\n(Hint: Use the \"New 'Play All' Title\" utility to 'backup'"
		set txt "$txt\nthe old VOB Cell references before using this function.)"
	}
	set txt "$txt\n\nNote: The VOB file is not modified by this operation."

	label $w.lab -text $txt -justify left -anchor w
	pack  $w.lab -fill x -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -fill x -padx 4
	label $w.f1.lab -text "Remove cell" -width 11 -anchor w
	pack  $w.f1.lab -side left
	scale $w.f1.s1 -showvalue true -from 1 -to $::pgced::__numcells \
			-orient horizontal -variable ::pgced::remove_cell_num
	pack  $w.f1.s1 -padx 1 -fill x

	frame  $w.fb
	pack   $w.fb -anchor e -padx 4 -pady 4
	button $w.fb.ok -text "OK" -width 10 -command {set ::pgced::delcelldone 1}
	pack   $w.fb.ok -side left -padx 4
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::pgced::delcelldone 0}
	pack   $w.fb.cancel -side left

	update
	set wx [expr {[winfo rootx $parent]+(([winfo width  $parent]-[winfo reqwidth  $w])/2)}]
	set wy [expr {[winfo rooty $parent]+(([winfo height $parent]-[winfo reqheight $w])/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm resizable $w 0 0

	wm deiconify $w
	focus $w
	grab set $w

	set ::pgced::delcelldone -1
	tkwait variable ::pgced::delcelldone
	grab release $w
	if {$parent != "."} {
		grab set $parent
	}
	destroy $w
	if {!$::pgced::delcelldone} return

	set cellnum $::pgced::remove_cell_num

	incr ::pgced::__numcells -1

	# substract the buttons of the removed cells from the total number of buttons
	set cellposidx [expr {($cellnum - 1) * 4}]
	set vobid  [get_int ::pgced::__cellpostable $cellposidx 2]
	set cellid [get_int ::pgced::__cellpostable [expr {$cellposidx + 3}] 1]
	if {$cur_type == "menu"} {
		set numbuts $::buttons::butsinfo(menu,$cur_vts,$vobid,$cellid,numbuts)
		incr ::menus($cur_index,numbuttons) [expr {$numbuts * -1}]
	} else {
		if {[info exists ::pgcs($cur_index,numbuttons)]} {
			set numbuts $::buttons::butsinfo(pgc,$cur_vts,$vobid,$cellid,numbuts)
			incr ::pgcs($cur_index,numbuttons) [expr {$numbuts * -1}]
		}
	}

	# remove the new entry from the current cell playback table
	set cellpbidx [expr {($cellnum - 1) * 24}]
	set ::pgced::__cellpbtable "[string range $::pgced::__cellpbtable 0 [expr {$cellpbidx-1}]][string range $::pgced::__cellpbtable [expr {$cellpbidx+24}] end]"

	# remove the new entry from the current cell position table
	set cellposidx [expr {($cellnum - 1) * 4}]
	set ::pgced::__cellpostable "[string range $::pgced::__cellpostable 0 [expr {$cellposidx-1}]][string range $::pgced::__cellpostable [expr {$cellposidx+4}] end]"
	
	# remove unnecessary entries in the program map
	set newprogrammap {}
	for {set i 0} {$i < [string length $::pgced::__programmap]} {incr i} {
		set c [get_int ::pgced::__programmap $i 1]
		if {$c<=$::pgced::__numcells && $c!=0} {
			set newprogrammap $newprogrammap[d2c $c 1]
		}
	}
	set ::pgced::__numprograms [string length $newprogrammap]
	if {[expr {[string length $newprogrammap] % 2}]} {
		set newprogrammap $newprogrammap\000
	}
	set ::pgced::__programmap $newprogrammap

	# Set the STC Discontinuity flag in the NEXT Cell entry, and clear the Seamless flag
	if {$::pgced::__numcells >= $cellnum} {
		set_int ::pgced::__cellpbtable $cellpbidx [expr {([get_int ::pgced::__cellpbtable $cellpbidx 1] | 2) & 0xF7}] 1
	}
	# reload the new tables in the editor
	set total_time [build_cell_table $parent]

	# correct the total playback time
	set ::pgced::pbtimetxt [BCDtime $total_time true]
	set ::pgced::__pgcpbtime $total_time
	set ::pgced::__pgcheader "[string range $::pgced::__pgcheader 0 3]$::pgced::__pgcpbtime[string range $::pgced::__pgcheader 8 end]"
	::main::dvd_modified true false

	if {$cur_type == "pgc"} {
		.pgceditor.fb.ptt configure -foreground red
		if {! [info exists ::pgced::__warning_delete_chapters]} {
			set txt "You must edit the chapter table to remove or reassign\nchapters pointing to deleted programs!\n\nClick on the \"Edit Chapter (PTT) Table\" button.\n\n(This message will not be displayed anymore!)"
			show_warning -title "Remove Cell" -icon warning \
				-message $txt -parent .
			set ::pgced::__warning_delete_chapters true
		}
	}

	return
}

}	;# end of namespace ::pgced
