#############################################################################
# readdvd.tcl Copyright 2004, 2005 r0lZ
# read DVD IFOs in memory.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################


## Variables usage:
##
## pgcs(numvts)								;# number of VTS in the DVD
## pgcs($vtsnum,numpgcs)					;# number of PGCs in this VTS
## pgcs($vtsnum,$pgcnum)					;# list of commands in this PGC
## pgcs($vtsnum,$pgcnum,Title)				;# Title number for this PGC
## pgcs($vtsnum,$pgcnum,TTN)				;# Title number within titleset
## pgcs($vtsnum,$pgcnum,header)				;# Header part of the PGC
## pgcs($vtsnum,$pgcnum,numprograms)		;# Number of Programs in the PGC
## pgcs($vtsnum,$pgcnum,programmap)			;# PGC program map, in binary
## pgcs($vtsnum,$pgcnum,numcells)			;# Number of Cells in the PGC
## pgcs($vtsnum,$pgcnum,cellpbtable)		;# Cell Playback Table, in binary
## pgcs($vtsnum,$pgcnum,cellpostable)		;# Cell Position Table, in binary
## pgcs($vtsnum,$pgcnum,pttlist)			;# List of lists with {program chapter ttu}
##											;# formatted as {{pg# ptt#} {pg# ptt#} ...}
##
## ttns($vtsnum,$ttn)						;# Entry PGC for this TTN (title in VTS)
## titles($vtsnum,$title)					;# Entry PGC for this Title (global title)
##
##
## menus(numvts)							;# number of VTS in the DVD
## menus($vtsnum,numlus)					;# number of LUs in this VTS (0=no menu)
## menus($vtsnum,lus_are_identical)			;# Bool: There is the same number of PGCs in the
##											;# different LUs, and same commands in the PGCs.
## menus($vtsnum,$lunum,lang)				;# language code for this LU
## menus($vtsnum,$lunum,numpgcs)			;# number of PGCs in this LU
## menus($vtsnum,$lunum,$pgcnum,mask)		;# category mask value for this PGC
## menus($vtsnum,$lunum,$pgcnum,masktext)	;# mask in human readeable form, or ""
## menus($vtsnum,$lunum,$pgcnum)			;# list of commands in this PGC
## menus($vtsnum,$lunum,$pgcnum,header)		;# Header part of the PGC
## menus($vtsnum,$lunum,$pgcnum,numprograms);# Number of Programs in the PGC
## menus($vtsnum,$lunum,$pgcnum,programmap)	;# PGC program map, in binary
## menus($vtsnum,$lunum,$pgcnum,numcells)   ;# Number of Cells in the PGC
## menus($vtsnum,$lunum,$pgcnum,cellpbtable);# Cell Playback Table, in binary
## menus($vtsnum,$lunum,$pgcnum,cellpostable);# Cell Position Table, in binary
## menus($vtsnum,$lunum,$pgcnum,numbuttons)	;# total number of buttons for this PGC.
##
## raw_sectors($vtsnum,<tablename>)			;# raw IFO data, splitted by tables
## VMG tables names: VMGM_MAT VMG_TT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT
##					 VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP
## VTS tables names: VTSI_MAT VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI
##					 VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP
##
## dvd_modified								;# flag: the DVD has been modified.
##
## Notes:
## - Tcl arrays are one dimention only.  The indexes are __string__ values.
##   Therefore, you cannot use spaces around the comas in the index names!
## - the array indexes that are not prefixed with $ are constant text strings.
## - The vtsnum, pgcnum, lunum and commands indexes are numbered starting at 1
## - vtsnum 0 is the VMG (VIDEO_TS.IFO)
## - the lists of commands are Tcl lists with the commands in hex form,
##   with blank between bytes: "00 00 00 00 00 00 00 00".
##   Elements beginning with * are pre, post and cell separators
##
##
## Buttons informations.  Available ONLY for Windows and if the DVD is NOT encrypted!
## The Menu VOBs only are parsed.  Don't expect to find infos on BOV buttons!
## Note: vts = 0 for VMG.
##
## For each VTS, the following variables are set:
## ::buttons::butsinfo($type,$vts,errorcode)	;# 0=OK
## ::buttons::butsinfo($type,$vts,errormsg)
## ::buttons::butsinfo($type,$vts,numvobids)
## For each VOB ID
## ::buttons::butsinfo($type,$vts,$vobid,numcellids)
## For each Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,navpack)	;# first NAV pack LBA for current VOB/Cell ID
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numbuts)	;# number of buttons (in first group)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,numgroups)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceselectbutn)
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,forceactionbutn)
## For eacg group
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,grouptype)	;# 1=normal, 2=letterbox, 4=pan&scan (ored together)
## For each button
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,autoaction) ;# 0 or 1
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,colorsheme) ;# 0 to 3
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,coords)	;# Button position: {x_start y_start - x_end y_end}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,udlr)		;# Adjacent buttons: {up down left right}
## ::buttons::butsinfo($type,$vts,$vobid,$cellid,$group,$butnum,command)	;# Button command: {XX XX XX XX XX XX XX XX}

namespace eval ::readdvd {

namespace export readdvd readdvdfile parse_PGCI_UT

#############################################################################
## Procedure:  get_hex_s
## Get a hex string from inside the binary data.  Pad bytes with spaces.
proc {get_hex_s} {binarystring_upvar start_offset length} {

	upvar $binarystring_upvar binarystring

	set s [string range $binarystring $start_offset [expr {$start_offset + $length - 1}]]
	binary scan $s H* hex
	set len [string length $hex]
	set out ""
	for {set i 0} {$i<$len} {incr i 2} {
		set out "$out[string range $hex $i [expr {$i+1}]] "
	}
	return [string trim $out]
}

#############################################################################
## Procedure:  getCommandsFromPGC
## return a list with all commands of the provided PGC in hexadecimal form.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
## Elements beginning with * are pre, post and cell separators
proc {getCommandsFromPGC} {data_upvar pgcindex} {

	upvar $data_upvar data

	set cmdtable_start [get_int data [expr {$pgcindex+0XE4}] 2]
	if {$cmdtable_start == 0} {
		set num_pre_commands  0
		set num_post_commands 0
		set num_cell_commands 0
	} else {
		# remove PGC header, and keep only the command table and whatever follow...
		set cmds_table [string range $data [expr {$pgcindex+$cmdtable_start}] end]

		# get number of pre, post and cell commands
		set num_pre_commands  [get_int cmds_table 0 2]
		set num_post_commands [get_int cmds_table 2 2]
		set num_cell_commands [get_int cmds_table 4 2]
	}

	set idx 8
	lappend outlist "* pre"
	for {set i 0} {$i < $num_pre_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* post"
	for {set i 0} {$i < $num_post_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	lappend outlist "* cell"
	for {set i 0} {$i < $num_cell_commands} {incr i} {
		lappend outlist [get_hex_s cmds_table $idx 8]
		incr idx 8
	}
	return $outlist
}

#############################################################################
## Procedure:  getHeaderFromPGC
## return the header part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getHeaderFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	return [string range $data $pgcindex [expr {$pgcindex + 0XEB}]]
}

#############################################################################
## Procedure:  getProgramMapFromPGC
## return the program map part of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getProgramMapFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numprograms [get_int data [expr {$pgcindex + 2}] 1]
	if {$numprograms == 0} {return [list 0 ""]}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XE6}] 2]}]
	set end [expr {$start + $numprograms - 1}]
	if {[expr {$end % 2}] == 0} {incr end}	;# must be padded to word boundary
	return [list $numprograms [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPBTableFromPGC
## return the cells Playback table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPBTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr {$pgcindex + 3}] 1]
	if {$numcells == 0} {return [list 0 ""]}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XE8}] 2]}]
	set end [expr {$start + ($numcells * 24) - 1}]
	return [list $numcells [string range $data $start $end]]
}

#############################################################################
## Procedure:  getCellsPosTableFromPGC
## return the cells Position table out of the provided PGC.
## data must contain a valid PGC in binary, pointed to by the pgcindex.
proc {getCellsPosTableFromPGC} {data_upvar pgcindex} {
	upvar $data_upvar data
	set numcells [get_int data [expr {$pgcindex + 3}] 1]
	if {$numcells == 0} {return ""}
	set start [expr {$pgcindex + [get_int data [expr {$pgcindex + 0XEA}] 2]}]
	set end [expr {$start + ($numcells * 4) - 1}]
	return [string range $data $start $end]
}

#############################################################################
## Procedure:  parse_PGCI_UT
## process the data in the VMGM_PGCI_UT
proc {parse_PGCI_UT} {pgci_ut_raw_data_upvar vtsnum menusvar} {

	upvar $pgci_ut_raw_data_upvar pgci_ut_raw_data

	# get number of Language Units
	if {$pgci_ut_raw_data == ""} {
		set ${menusvar}($vtsnum,numlus) 0
		return
	} else {
		set numlus [get_int pgci_ut_raw_data 0 2]
		set ${menusvar}($vtsnum,numlus) $numlus
	}

	#dprint 2 "Number of Language Units: $numlus"

	set lu_pgcs(1) 0

	# for all Language units
	for {set lu 1} {$lu <= $numlus} {incr lu} {

		# get LU's language, number of PGCs and offset
		set lang [string range $pgci_ut_raw_data [expr {$lu * 8}] [expr {($lu * 8) + 1}]]
		if {$lang == "\000\000" || $lang == "\777\777" || $lang == "  "} {set lang "--"}
		set ${menusvar}($vtsnum,$lu,lang) $lang

		set offset [get_int pgci_ut_raw_data [expr {($lu * 8) + 4}] 4]

		set ludata [string range $pgci_ut_raw_data $offset end]

		set numpgcs [get_int ludata 0 2]
		set ${menusvar}($vtsnum,$lu,numpgcs) $numpgcs

		#dprint 2 "\nMenu language unit $lu: \"$lang\""
		#dprint 2 "offset to VMGM_LU = 0X[format %08X $offset] ($offset)"

		set lu_pgcs($lu) $numpgcs

		# for each PGC
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {

			# get and store category mask
			set category [get_int ludata [expr {$pgc * 8}] 1]
			set ${menusvar}($vtsnum,$lu,$pgc,mask) $category
			switch [expr {$category & 0X0F}] {
				0 {set masktext ""}
				2 {set masktext "TitleM"}
				3 {set masktext "RootM"}
				4 {set masktext "SubPicM"}
				5 {set masktext "AudioM"}
				6 {set masktext "AngleM"}
				7 {set masktext "ChapterM"}
				default {set masktext "UNKNOWN MENU"}
			}
			set ${menusvar}($vtsnum,$lu,$pgc,masktext) $masktext

			# get and store PGC offset
			set pgc_offset [get_int ludata [expr {($pgc * 8) + 4}] 4]

			# get and store PGC's commands
			set ${menusvar}($vtsnum,$lu,$pgc) [getCommandsFromPGC ludata $pgc_offset]

			# get and store the other PGC parts
			set ${menusvar}($vtsnum,$lu,$pgc,header) [getHeaderFromPGC ludata $pgc_offset]
			set tmp [getProgramMapFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numprograms) [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,programmap)  [lindex $tmp 1]

			set tmp [getCellsPBTableFromPGC ludata $pgc_offset]
			set ${menusvar}($vtsnum,$lu,$pgc,numcells)    [lindex $tmp 0]
			set ${menusvar}($vtsnum,$lu,$pgc,cellpbtable) [lindex $tmp 1]

			set ${menusvar}($vtsnum,$lu,$pgc,cellpostable) \
					[getCellsPosTableFromPGC ludata $pgc_offset]
		}
	}
	# verify that there is the same number of PGCs in each LUs,
	# and that PGC commands are identical!
	set numpgcs $lu_pgcs(1)
	set lus_are_identical true
	if {$numlus > 1} {
		if {$vtsnum == 0} {set domain "VMGM"} else {set domain "VTST $vtsnum"}
		for {set lu 2} {$lu <= $numlus} {incr lu} {
			if {$lu_pgcs($lu) != $numpgcs} {
				set txt "Warning: There is a different number of PGCs"
				set txt "$txt\nin the menu LUs of the $domain domain."
				set txt "$txt\n\nThis is somewhat unusual, and may cause"
				set txt "$txt\nproblems with the 'Set Menu Type' and"
				set txt "$txt\n'Copy LU Commands' utilities."
				tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
						-type ok -message $txt
				set lus_are_identical false
				break
			}
		}
		if {$lus_are_identical} {
			for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
				set lu1cmds [subst \$${menusvar}($vtsnum,1,$pgc)]
				for {set lu 2} {$lu <= $numlus} {incr lu} {
					if {[subst \$${menusvar}($vtsnum,$lu,$pgc)] != $lu1cmds} {
						set txt "Warning: There are different commands"
						set txt "$txt\nin the various PGCs of the menu LUs"
						set txt "$txt\nof the $domain domain."
						set txt "$txt\n\nThis is somewhat unusual, and may cause"
						set txt "$txt\nproblems with the 'Set Menu Type' and"
						set txt "$txt\n'Copy LU Commands' utilities."
						tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
								-type ok -message $txt
						set lus_are_identical false
						break
					}
				}
				if {!$lus_are_identical} break
			}
		}
	}
	set ${menusvar}($vtsnum,lus_are_identical) $lus_are_identical
}

#############################################################################
## Procedure:  parse_VTS_PGCI
## process the data in the VTS_PGCI of a VTS IFO
proc parse_VTS_PGCI {pgci_data_upvar vtsnum pgcsvar} {

	upvar $pgci_data_upvar pgci_data

	if {$pgci_data == ""} {
		set ${pgcsvar}($vtsnum,numpgcs) 0
		return
	}
	set numpgcs [get_int pgci_data 0 2]
	set ${pgcsvar}($vtsnum,numpgcs) $numpgcs

	# for each PGC
	for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
		# get and store PGC's TTN
		set ${pgcsvar}($vtsnum,$pgc,TTN) [expr {[get_int pgci_data [expr {$pgc * 8}] 1] & 0x7F}]	;# bug fixed

		# get PGC offset
		set pgc_offset [get_int pgci_data [expr {($pgc * 8) + 4}] 4]
		# get and store PGC's commands
		set ${pgcsvar}($vtsnum,$pgc) [getCommandsFromPGC pgci_data $pgc_offset]

		set ${pgcsvar}($vtsnum,$pgc,header) [getHeaderFromPGC pgci_data $pgc_offset]

		set tmp [getProgramMapFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numprograms) [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,programmap)  [lindex $tmp 1]

		set tmp [getCellsPBTableFromPGC pgci_data $pgc_offset]
		set ${pgcsvar}($vtsnum,$pgc,numcells)    [lindex $tmp 0]
		set ${pgcsvar}($vtsnum,$pgc,cellpbtable) [lindex $tmp 1]

		set ${pgcsvar}($vtsnum,$pgc,cellpostable) [getCellsPosTableFromPGC pgci_data $pgc_offset]
		
		set ${pgcsvar}($vtsnum,$pgc,pttlist) {}	;# bug fixed
	}
}


#############################################################################
## Procedure:  parse_VTS_PTT_SRPT
## process the data in the VTS_PTT_SRPT of a VTS IFO
## Build chapter list for each titles
proc parse_VTS_PTT_SRPT {vts_ptt_srpt_data_upvar vtsnum pgcsvar ttnsvar} {

	upvar $vts_ptt_srpt_data_upvar data

	set numttus 0
	if {$data != ""} {
		set numttus [get_int data 0 2]
		set maxpgcn -1
		for {set ttu 1} {$ttu <= $numttus} {incr ttu} {
			set ptt 1
			set idx [get_int data [expr {4 + ($ttu * 4)}] 4]
			if {$ttu < $numttus} {
				set endidx [expr {[get_int data [expr {4 + (($ttu+1) * 4)}] 4] -1}]
			} else {
				set endidx [get_int data 4 4]
			}

			# build table of entry PGCs for each TTU
			set pgcn [get_int data $idx 2]
			set ${ttnsvar}($vtsnum,$ttu) $pgcn

			while {$idx < $endidx} {
				set pgcn [get_int data $idx 2]
				if {$pgcn > $maxpgcn} {
					set maxpgcn $pgcn
				}
				set pg  [get_int data [expr {$idx+2}] 2]
				#lappend PTTlist($pgcn) "$pg $ptt"
				lappend PTTlist($pgcn) "$pg $ptt $ttu" ;# Added TTU in 0.5.0beta11

				#if {$ttu != ${pgcsvar}($vtsnum,$pgcn,TTN)} {
				#	WARNING!  TTU do not match TTN
				#}
				incr idx 4
				incr ptt
			}
		}
		for {set pgcn 1} {$pgcn <= $maxpgcn} {incr pgcn} {
			if {[info exists PTTlist($pgcn)]} {
				set ${pgcsvar}($vtsnum,$pgcn,pttlist) $PTTlist($pgcn)
			} else {
				set ${pgcsvar}($vtsnum,$pgcn,pttlist) {}
			}
		}
	}
	set ${pgcsvar}($vtsnum,numttus) $numttus
	return $numttus
}

#############################################################################
## Procedure:  read_VMG_IFO
## read the data of the VIDEO_TS.IFO of the DVD directory supplied
proc read_VMG_IFO {filename sectorsvar pgcsvar menusvar} {
	#dprint 2 "Input Directory: $dvddir"

	set data [readFile $filename binary]
	if {$data == ""} {return 1}

	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VMG"} {
		errmsg 1000 "checking IFO file type" \
				"File \"$filename\"\nis not of correct type DVDVIDEO-VMG"
		return 1000
	}

	#dprint 2 ""
	##### split the VIDEO_TS.IFO file data in main tables parts
	set start_address(0) 0

	set end_address(0) [get_int data 0X80 4]	;# should work, but...
	set fp_pgc_start [get_int data 0X84 4]


	### WARNING: BUG in IfoEdit: the end address of the VMGM_MAT is not exact!!!
	### IfoEdit never change this address, even when the command table is growing!
	### We must therefore calculate the real end address (what a pitty!).
	if {$fp_pgc_start != 0} {
		set cmdtable_start [get_int data [expr {$fp_pgc_start+0XE4}] 2]
		set numpre  [get_int data [expr {$fp_pgc_start+$cmdtable_start}] 2]
		set numpost [get_int data [expr {$fp_pgc_start+$cmdtable_start+2}] 2]
		set numcell [get_int data [expr {$fp_pgc_start+$cmdtable_start+4}] 2]
		set cmdtable_len [expr {8+($numpre*8)+($numpost*8)+($numcell*8)}]
		set cmdtable_end [expr {$fp_pgc_start+$cmdtable_start+$cmdtable_len-1}]
	} else {
		# Yes, there are DVDs WITHOUT FP-PGC!
		# A new FP-PGC is created by PgcEdit in parse_VMG_IFO
		set cmdtable_end $end_address(0)
	}
	if {$cmdtable_end != $end_address(0)} {
		set msg "Warning: The value for \"End byte of VMGI_MAT\" is incorrect!"
		set msg "$msg\n(This value is in the VIDEO_TS.IFO's VMGM_MAT at offset 0X80)"
		set msg "$msg\n\nValue in file: $end_address(0)"
		set msg "$msg\nCorrect value: $cmdtable_end"
		set msg "$msg\n\nIfoEdit doesn't upgrade this value when you add or remove"
		set msg "$msg\ncommands in the First-Play PGC."
		set msg "$msg\n\nThe correct value will be updated when you save the DVD."
		show_warning -title "PgcEdit: Open DVD" -icon error \
			-message $msg -parent .
		# finally, fix the error!
		set end_address(0) $cmdtable_end
		set ::dvd_modified true
	}

	for {set i 1} {$i < 8} {incr i} {
		set start_address($i) [expr {[get_int data [expr {(0XC0 + (4 * $i))}] 4] * 0X800}]
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VMGM_VOBU_ADMAP has the end_byte at position 0, and
			# VMG_TXTDT_MG at pos 16! All other tables at pos 4.
			switch $i {
				5 {set end_byte_pos 16}
				7 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr {$start_address($i) + \
					[get_int data [expr {$start_address($i)+$end_byte_pos}] 4]}]
		}
	}
	set i 0
	foreach p {VMGM_MAT VMG_TT_SRPT VMGM_PGCI_UT VMG_PTL_MAIT \
			VMG_VTS_ATRT VMG_TXTDT_MG VMGM_C_ADT VMGM_VOBU_ADMAP} {
		if {$end_address($i) > 0} {
			set ${sectorsvar}(0,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}(0,$p) ""
		}
		incr i
	}

	parse_VMG_IFO $sectorsvar $pgcsvar $menusvar

	return 0
}

#############################################################################
## Procedure:  parse_VMG_IFO
## process the data of the VIDEO_TS.IFO
proc parse_VMG_IFO {sectorsvar pgcsvar menusvar} {
	##### Number of title sets
	set num_vts [get_int ${sectorsvar}(0,VMGM_MAT) 0X3E 2]
	set ${pgcsvar}(numvts) $num_vts
	set ${menusvar}(numvts) $num_vts	;#!!!!!!

	##### VMGM_MAT/FP-PGC
	#dprint 2 "\nFirst Play PGC"
	set fp_pgc_start [get_int ${sectorsvar}(0,VMGM_MAT) 0X84 4]

	# Fix the No FP-PGC error
	if {$fp_pgc_start == 0} {
		set msg "Warning: There is no First-Play PGC in your VMG!"
		set msg "$msg\n(The First-Play PGC start byte value is 0)"
		set msg "$msg\n\nA new, blank First-Play PGC will be created."
		show_warning -title "PgcEdit: Open DVD" -icon error \
			-message $msg -parent .

		# build a new FP-PGC from scratch
		set ${pgcsvar}(0,0) {{* pre} {30 02 00 00 00 01 00 00} {* post} {* cell}}
		set tmp [string repeat "\000" 236]
		set_int tmp 4 64 4	;# fix Playback Time to legal value (assumes PAL)
		set ${pgcsvar}(0,0,header) $tmp
		# fix the FP-PGC Start Byte in VMGM_MAT
		set_int ${sectorsvar}(0,VMGM_MAT) 0X84 \
				[expr {[get_int ${sectorsvar}(0,VMGM_MAT) 0X80 4]+1}] 4

		set ::dvd_modified true
	} else {
		# Get the FP-PGC commands and header.
		set ${pgcsvar}(0,0) [getCommandsFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start]
		set ${pgcsvar}(0,0,header) [getHeaderFromPGC ${sectorsvar}(0,VMGM_MAT) $fp_pgc_start]
		set ${pgcsvar}(0,0,pttlist) ""
	}
	# Init the FP-PGC values.
	set ${pgcsvar}(0,numpgcs) 1
	set ${pgcsvar}(0,0,numprograms)  0
	set ${pgcsvar}(0,0,programmap)   ""
	set ${pgcsvar}(0,0,numcells)     0
	set ${pgcsvar}(0,0,cellpbtable)  ""
	set ${pgcsvar}(0,0,cellpostable) ""

	##### VMGM_PGCI_UT
	#dprint 2 "\nVMGM_PGCI_UT"
	parse_PGCI_UT ${sectorsvar}(0,VMGM_PGCI_UT) 0 ${menusvar}

	return 0
}

proc parse_VMG_TT_SRPT {sectorsvar pgcsvar ttnsvar} {
	##### get the table of titles from the VMG_TT_SRPT
	set data [subst $${sectorsvar}(0,VMG_TT_SRPT)]
	if {$data != ""} {
		set numtitles [get_int data 0 2]
		for {set t 1} {$t <= $numtitles} {incr t} {
			set t_idx [expr {14+(($t-1)*12)}]
			set vtsnum [get_int data $t_idx 1]
			set ttnnum [get_int data [expr {$t_idx+1}] 1]
			catch {set ${pgcsvar}($vtsnum,[subst $${ttnsvar}($vtsnum,$ttnnum)],Title) $t}
			#set ${titlesvar}($t) [list $vtsnum [subst $${ttnsvar}($vtsnum,$ttnnum)]]
		}
	}
}

#############################################################################
## Procedure:  read_VTS_IFO
## process the data in one VTS_xx_0.IFO of the DVD directory supplied
proc read_VTS_IFO {filename vtsnum sectorsvar pgcsvar menusvar ttnsvar} {

	# Open and read original IFO file in memory
	set data [readFile $filename binary]
	if {$data == ""} {return 1}

	# Verify file is of correct type
	if {[string range $data 0 11] != "DVDVIDEO-VTS"} {
		errmsg 1000 "checking IFO file type" \
				"File \"$filename\"\nis not of correct type DVDVIDEO-VTS"
		return 1000
	}

	##### split the VTS_xx_0.IFO file data in main tables parts
	set start_address(0) 0
	set end_address(0) [get_int data 0X80 4]
	for {set i 1} {$i < 9} {incr i} {
		set start_address($i) [expr {[get_int data [expr {(0XC4 + (4 * $i))}] 4] * 0X800}]
		if {$start_address($i) == 0} {
			set end_address($i) 0
		} else {
			# VTSM_VOBU_ADMAP and VTS_VOBU_ADMAP have the end_byte at position 0, not 4!
			switch $i {
				6 -
				8 {set end_byte_pos 0}
				default {set end_byte_pos 4}
			}
			set end_address($i) [expr {$start_address($i) + \
					[get_int data [expr {$start_address($i)+$end_byte_pos}] 4]}]
		}
	}

	set i 0
	foreach p {VTSI_MAT VTS_PTT_SRPT VTS_PGCI VTSM_PGCI_UT VTS_TMAPTI \
			VTSM_C_ADT VTSM_VOBU_ADMAP VTS_C_ADT VTS_VOBU_ADMAP} {
		if {$end_address($i) > 0} {
			set ${sectorsvar}($vtsnum,$p) [string range $data $start_address($i) \
						$end_address($i)]
		} else {
			set ${sectorsvar}($vtsnum,$p) ""
		}
		incr i
	}
	# Fix the number of VOB IDs in DVDShrink reauthored movies
	set data [subst $${sectorsvar}($vtsnum,VTS_C_ADT)]
	set endbyte [get_int data 4 4]
	set numvodids [get_int data 0 2]
	set realnvids 0
	set lastvid 0
	for {set idx 8} {$idx<$endbyte} {incr idx 12} {
		if {$lastvid != [get_int data $idx 2]} {
			set lastvid [get_int data $idx 2]
			incr realnvids
		}
	}
	if {$numvodids != $realnvids} {
		set txt "WARNING:\nThe number of VOB IDs stored in the VTS_C_ADT table of VTS $vtsnum ($numvodids)"
		set txt "$txt\ndon't match the real number of VOB IDs in the table ($realnvids)."
		set txt "$txt\n\n(This is a known bug of DVDShrink in reauthor mode)"
		set txt "$txt\n\nThe table will be fixed with the correct number of VOB IDs ($realnvids)."
		if {$sectorsvar == "::raw_sectors"} {
			set txt "$txt\nPlease save the DVD!"
		}
		show_warning -title "PgcEdit: Open DVD" -icon error -message $txt -parent .
		set_int data 0 $realnvids 2
		set ${sectorsvar}($vtsnum,VTS_C_ADT) $data
		set ::dvd_modified true
	}

	unset data

	parse_VTS_IFO $vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar
	return 0
}

#############################################################################
## Procedure:  parse_VTS_IFO
## process the data in one VTS_xx_0.IFO
proc parse_VTS_IFO {vtsnum sectorsvar pgcsvar menusvar ttnsvar} {

	##### VTS_PGCI
	#dprint 2 "\nVTS_PGCI"
	parse_VTS_PGCI ${sectorsvar}($vtsnum,VTS_PGCI) $vtsnum $pgcsvar

	##### VTS_PTT_SRPT
	parse_VTS_PTT_SRPT ${sectorsvar}($vtsnum,VTS_PTT_SRPT) $vtsnum $pgcsvar $ttnsvar

	##### VTSM_PGCI_UT
	#dprint 2 "\nVTSM_PGCI_UT"
	parse_PGCI_UT ${sectorsvar}($vtsnum,VTSM_PGCI_UT) $vtsnum $menusvar

	return 0
}

#############################################################################
## Procedure:  readdvd
## read the DVD in memory, and refresh the interface
## args: directoryname:  Path to the VIDEO_TS directory.
##       sectorsvar, pgcsvar, menusvar: name of variables to store the IFOs in.
proc readdvd {directoryname sectorsvar pgcsvar menusvar ttnsvar} {

	set zerovobfiles {}

	# unset old global variables
	set ${sectorsvar}(x) ""
	set ${pgcsvar}(x) ""
	set ${menusvar}(x) ""
	set ${ttnsvar}(x) ""
	unset $sectorsvar $pgcsvar $menusvar $ttnsvar

	set ::dvd_modified false

	# Open and read VMG IFO file
	show_msg "Reading VMG IFO."
	set errcode [read_VMG_IFO [file join $directoryname "VIDEO_TS.IFO"] \
			$sectorsvar $pgcsvar $menusvar]
	if {$errcode == 0} {
		set filename [file join $directoryname "VIDEO_TS.VOB"]
		if {[file exists $filename] && [file size $filename]==0} {
			set zerovobfiles 0
		}

		set numvts [subst $${pgcsvar}(numvts)]

		# compare stored number of VTS and real number of VTS_XX_0.IFO files.
		for {set vtsnum2 1} {$vtsnum2 <= 99} {incr vtsnum2} {
			if {! [file exists [file join $directoryname "VTS_[format %02i $vtsnum2]_0.IFO"]]} {
				incr vtsnum2 -1
				break
			}
		}
		if {$vtsnum2 != $numvts} {
			set txt "WARNING:\nThe number of VTS stored in VIDEO_TS.IFO ($numvts)"
			set txt "$txt\ndon't match the real number of Title Sets stored on disk ($vtsnum2)."
			if {$numvts < $vtsnum2} {
				set txt "$txt\n\nYou should manually remove all files related to VTS [expr $numvts+1]"
				if {[expr $numvts+1] < $vtsnum2} {set txt "${txt}-$vtsnum2"}
				set txt "${txt}.\n\nIf you really need this Title Set, try to import it with one"
				set txt "$txt\nof the \"Import title clip\" macros or the \"Import VTS Titles\""
				set txt "$txt\nand \"Import Menu\" utilities."
			} else {
				set txt "$txt\n\nYou should manually copy all files related to VTS [expr $vtsnum2+1]"
				if {[expr $vtsnum2+1] < $numvts} {set txt "${txt}-$numvts"}
				set txt "${txt},\nor restore the backup."
			}
			set txt "${txt}.\n\nWarning: If you do so, there is a risk of navigation errors!"
			set txt "${txt}.\nCheck your DVD carefully! (Use Info -> Calls Cross References)"
			tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
					-type ok -message $txt
		}

		# read all VTS
		for {set vtsnum 1} {$vtsnum <= $numvts} {incr vtsnum} {
			show_msg "Reading VTS $vtsnum IFO."
			set errcode [read_VTS_IFO \
					[file join $directoryname "VTS_[format %02i $vtsnum]_0.IFO"] \
					$vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar]
			set filename [file join $directoryname "VTS_[format %02i $vtsnum]_0.VOB"]
			if {[file exists $filename] && [file size $filename]==0} {
				lappend zerovobfiles $vtsnum
			}
			# missing VTS
			if {$errcode != 0} {
				set txt "ERROR: Could not open the VTS $vtsnum IFO file."
				set txt "$txt\n\nTry to fix the problem with IfoEdit, or restore the backup."
				tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
						-type ok -message $txt
				break
			}
		}
	}

	parse_VMG_TT_SRPT $sectorsvar $pgcsvar $ttnsvar

	if {$errcode != 0} {
		set ${sectorsvar}(x) ""
		set ${pgcsvar}(x) ""
		set ${menusvar}(x) ""
		unset $sectorsvar $pgcsvar $menusvar
	}

	if {$errcode != 0} {
		show_msg "Error $errcode reading DVD trom $directoryname"
		return $errcode
	}

	# Read menus buttons informations
	if {[info exists ::buttons::dvd_reopened]} {
		set txt "DVD reopened OK.\n\nDo you want to parse the menu buttons again?\n"
		set txt "$txt\nYou MUST answer yes if a VOB file has been modified!"
		if {[tk_messageBox -icon question -title "PgcEdit: Reopen DVD" -parent . \
				-type yesno -default yes -message $txt] == "yes"} {
			set parsebuttons true
			set ::buttons::dvd_reopened 1
		} else {
			set parsebuttons false
			for {set vts 0} {$vts <= $numvts} {incr vts} {
				::menued::add_menu_buttons_marker $vts
			}
			set ::buttons::dvd_reopened -1
		}
	} else {
		set parsebuttons true
	}
	if {$parsebuttons} {
		namespace eval ::buttons {}
		namespace delete ::buttons
		namespace eval ::buttons {}
		unset -nocomplain ::buttons::butinfo
		::menued::get_buttons_all_menu_domains $directoryname
	}

	if {[llength $zerovobfiles] > 0 && $errcode == 0} {
		set txt "WARNING: There are empty (0 byte) VOB files in the menu domains:\n"
		foreach i $zerovobfiles {
			if {$i == 0} {
				set txt "${txt}VMGM, "
			} else {
				set txt "${txt}VTSM $i, "
			}
		}
		set txt "[string trim $txt ", "]\nAlthough it is legal, Nero will complain.\n"
		set txt "$txt\nDelete the empty VOB file"
		if {[llength $zerovobfiles] > 1} {set txt "${txt}s"}
		set txt "${txt}, and fix the references in the IFOs now?\n"
		set txt "${txt}(You will need to save the DVD!)"
		if {[tk_messageBox -icon warning -title "PgcEdit: Read DVD" -parent . \
				-type yesno -default yes -message $txt] == "yes"} {
			foreach vtsnum $zerovobfiles {
				if {$vtsnum == 0} {
					set name "VIDEO_TS.VOB"
					set tablesubname "VMGM"
				} else {
					set name "VTS_[format %02i $vtsnum]_0.VOB"
					set tablesubname "VTSM"
				}
			if {[deletefile [file join $directoryname $name] "PgcEdit: Read DVD" .]} {
					break
				} else {
					# remove the VTS_C_ADT and VTS_VOBU_ADMAP tables
					set ${sectorsvar}($vtsnum,${tablesubname}_C_ADT) ""
					set ${sectorsvar}($vtsnum,${tablesubname}_VOBU_ADMAP) ""
				}
				set ::dvd_modified true
			}
		}
	}

	#. configure -cursor $oldcursor
	show_msg ""
	return $errcode
}

proc reload_buttons_infos {vts} {
	update idletasks
	if {$vts == 0} {set vtsname "VMGM"} else {set vtsname "VTSM $vts"}
	set txt "You have modified the $vtsname VOB file.\n"
	set txt "${txt}The menu buttons informations will be updated\n"
	set txt "${txt}only when you load the DVD again.\n\n"
	set txt "${txt}Do you want to save the DVD and reload it now?"
	set r [tk_messageBox -title "PgcEdit: Button Informations" -icon question \
		-message $txt -parent . \
		-type yesno -default yes]
	if {$r == "yes"} {
		::main::Write_DVD false
		::main::ReRead_DVD false
	}
}

## read a specific IFO file.
## return 0 if file is a VMG VIDEO_TS.IFO
## return the number of the VTS if file is a VTS_XX_0.IFO file
## return -1 upon error
proc readdvdfile {filename sectorsvar pgcsvar menusvar ttnsvar} {
	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return -1
	}
	fconfigure $in -translation binary
	set errnum [catch {set data [read $in 12]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return -1
	}
	close $in

	switch $data {
		"DVDVIDEO-VMG" {
			if {[read_VMG_IFO $filename $sectorsvar $pgcsvar $menusvar] != 0} {
				return -1
			} else {
				return 0
			}
		}
		"DVDVIDEO-VTS" {
			if {[catch {set vtsnum [expr {[string trimleft [string range [file tail $filename] 4 5] "0"] * 1}]}] != 0} {
				errmsg 1000 "getting VTS number from filename" "Invalid filename \"[file tail $filename]\""
				return -1
			}
			if {[read_VTS_IFO $filename $vtsnum $sectorsvar $pgcsvar $menusvar $ttnsvar] != 0} {
				return -1
			} else {
				return $vtsnum
			}
		}
		default {
			errmsg 1000 "checking IFO file type" \
					"File \"$filename\"\nis not of correct type DVDVIDEO-VMG or DVDVIDEO-VTS"
			return -1
		}
	}
}

} ;# end namespace ::readdvd
