#############################################################################
# streams.tcl Copyright 2004, 2005 r0lZ
# Compile informations about all Video/Audio/Sub-picture streams in the DVD.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::streams {
namespace export get_streams_info
namespace import ::readdvd::*

set ::streams::audio_code_ext_list {
	{unspecified}
	{normal}
	{for visual impaired}
	{director comment}
	{alt director comment}
}
set ::streams::subpic_code_ext_list {
	{unspecified}
	{normal}
	{large}
	{children}
	-
	{normal captions}
	{large captions}
	{children captions}
	-
	{forced}
	-
	-
	-
	{director comments}
	{large director comments}
	{comments for children}
}
set ::streams::illegalextlabel "ILLEGAL CODE EXTENSION"

####### main proc
proc get_streams_info {} {
	if {! [info exists ::dvddir]} {
		bell
		return ""
	}

	set txt "DVD: \"$::dvddir/\""
	set txt "$txt\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set txt "$txt\nProvider ID: \"[::utils::get_provider_ID]\""
	set txt "$txt\nNumber of VTS: $::pgcs(numvts)"
	set txt "$txt\n"

	set numaudios [get_int ::raw_sectors(0,VMGM_MAT) 0X102 2]
	set numsubpics [get_int ::raw_sectors(0,VMGM_MAT) 0X154 2]
	set txt "$txt\nVMG Menu streams attributes (VIDEO_TS.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
	set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors(0,VMGM_MAT) 0X100 2] true]"
	for {set s 0} {$s < $numaudios} {incr s} {
		set offset [expr {(0X104 + ( $s * 8 ))}]
		set attr [string range $::raw_sectors(0,VMGM_MAT) $offset [expr {$offset + 7}]]
		set txt "$txt\n- Audio $s:  [audio_description $attr true $s]"
	}
	if {$numsubpics > 0} {
		set attr [string range $::raw_sectors(0,VMGM_MAT) 0X156 0X15B]
		#set txt "$txt\n- Subpic:   [subpic_description $attr], ID: 0x20"
		set txt "$txt\n- Subpic:   [subpic_description $attr]"
	}

	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0X102 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0X154 2]

		set txt "$txt\n\n"
		set txt "$txt\nVTS $vts Menu streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors($vts,VTSI_MAT) 0X100 2] true]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr {(0X104 + ( $s * 8 ))}]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr {$offset + 7}]]
			set txt "$txt\n- Audio $s:  [audio_description $attr true $s]"
		}
		if {$numsubpics > 0} {
			set attr [string range $::raw_sectors($vts,VTSI_MAT) 0X156 0X15B]
			#set txt "$txt\n- Subpic:   [subpic_description $attr], ID: 0x20"
			set txt "$txt\n- Subpic:   [subpic_description $attr]"
		}

		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0X202 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0X254 2]
		set txt "$txt\n"
		set txt "$txt\nVTS $vts VOB streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors($vts,VTSI_MAT) 0X200 2] true]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr {(0X204 + ( $s * 8 ))}]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr {$offset + 7}]]
			set txt "$txt\n- Audio $s:  [audio_description $attr true $s]"
		}
		for {set s 0} {$s < $numsubpics} {incr s} {
			set offset [expr {(0X256 + ( $s * 6 ))}]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr {$offset + 5}]]
			#set txt "$txt\n- Subpic $s: [subpic_description $attr], ID: 0x[format %02X [expr {(0x20+$s)}]]"
			set txt "$txt\n- Subpic $s: [subpic_description $attr]"
		}
	}
	::main::showtext .streamsinfo "PgcEdit: Streams information" $txt
	return $txt
}


proc set_streams_attribs {parentwin} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set pgctype [lindex $::editor_cmds_from 0]
	set vtsnum  [lindex [split [lindex $::editor_cmds_from 1] ","] 0]

	verify_num_streams_domain $pgctype $vtsnum

	if {$vtsnum == 0} {
		if {$pgctype == "pgc"} {
			bell
			return
		}
		set vtsname "VMGM"
		set mat_table $::raw_sectors(0,VMGM_MAT)
	} else {
		if {$pgctype == "menu"} {
			set vtsname "VTSM $vtsnum"
		} else {
			set vtsname "VTST $vtsnum"
		}
		set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
	}
	if {$pgctype == "menu"} {
		set astart [expr {(0X100)}]
		set numaudios  [get_int mat_table [expr {$astart+2}] 2]
		set numsubpics [get_int mat_table [expr {$astart+0X54}] 2]
	} else {
		set astart [expr {(0X200)}]
		if {$vtsnum > 0} {
			set numaudios  [get_int mat_table [expr {$astart+2}] 2]
			set numsubpics [get_int mat_table [expr {$astart+0X54}] 2]
		} else {
			set numaudios  0
			set numsubpics 0
		}
	}

	set w ".streams"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::streams::done 0}
	bind $w <Escape> "set ::streams::done 0"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm title $w "Streams attributes"

	if {$pgctype == "pgc"} {set domain "Title"} else {set domain "Menu"}
	label  $w.lab -text "$vtsname $domain domain streams:" -anchor w
	pack   $w.lab -side top -fill x -padx 4

	# VIDEO

	set vfmt [get_int mat_table $astart 2]
	set ::streams::screenformat  [expr {$vfmt & 0X0C00}]
	set ::streams::autoletterbox [expr {$vfmt & 0x0100}]
	set ::streams::autopanscan   [expr {$vfmt & 0x0200}]
	set ::streams::lbcropped     [expr {$vfmt & 0x0002}]

	frame  $w.f1 -borderwidth 2 -relief groove
	pack   $w.f1 -padx 4 -pady 2 -fill x
	label  $w.f1.lab -text "Video:" -width 9 -anchor w
	pack   $w.f1.lab -side left
	frame  $w.f1.af
	pack   $w.f1.af -side left
	radiobutton $w.f1.af.r1 -text "4:3"  -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::screenformat -value 0
	pack   $w.f1.af.r1 -fill x
	radiobutton $w.f1.af.r2 -text "16:9" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::screenformat -value [expr {(0X0C00)}]
	pack   $w.f1.af.r2 -fill x

	frame  $w.f1.at
	pack   $w.f1.at -side left
	checkbutton $w.f1.at.r2 -text "Automatic Pan&Scan"  -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::autopanscan   -offvalue [expr {(0X0200)}] -onvalue 0
	pack   $w.f1.at.r2 -fill x
	checkbutton $w.f1.at.r1 -text "Automatic Letterbox" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::autoletterbox -offvalue [expr {(0X0100)}] -onvalue 0
	pack   $w.f1.at.r1 -fill x
	checkbutton $w.f1.at.r3 -text "Letterbox: Top & bottom cropped" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::lbcropped -offvalue 0 -onvalue [expr {(0X0002)}]
	pack   $w.f1.at.r3 -fill x

	label  $w.f1.d -text "[video_description $vfmt false]" -justify left -anchor w
	pack   $w.f1.d -side left

	# AUDIO

	frame  $w.f2 -borderwidth 2 -relief groove
	pack   $w.f2 -padx 4 -pady 2 -ipady 1 -fill x
	if {$numaudios == 0} {
		label $w.f2.lab -text "No Audio streams"
		pack  $w.f2.lab
	} else {
		frame  $w.f2.pad -borderwidth 1
		pack   $w.f2.pad
	}
	for {set s 0} {$s < $numaudios} {incr s} {
		set offset [expr {$astart+4+($s*8)}]
		set aattr [string range $mat_table $offset [expr {$offset + 7}]]
		set ::streams::tmp_aattrs($s) $aattr
		set attrlist [audio_attributes $aattr $s]
		set ::streams::audio_lang_code_$s [lindex $attrlist 0]
		set tmp ::streams::audio_lang_code_$s
		if {[subst $$tmp] == "--"} {set ::streams::audio_lang_code_$s ""}
		set ::streams::audio_code_ext_$s [lindex $attrlist 1]

		frame  $w.f2.f$s
		pack   $w.f2.f$s -fill x
		label  $w.f2.f$s.lab -text "Audio $s:" -width 9 -anchor w
		pack   $w.f2.f$s.lab -side left
		entry  $w.f2.f$s.lc -textvariable ::streams::audio_lang_code_$s -width 2 \
				 -font $::config(fixed_width_font) -bd 1 \
				 -validate all -validatecommand "::streams::validate_langcode %P"
		pack   $w.f2.f$s.lc -side left -padx 2
		menubutton $w.f2.f$s.ce -textvariable ::streams::audio_code_ext_$s \
				-pady 0 -width 22 -menu $w.f2.f$s.ce.audiocodeext
		pack   $w.f2.f$s.ce -side left -padx 2
		menu $w.f2.f$s.ce.audiocodeext -tearoff 0
		foreach i $::streams::audio_code_ext_list {
			$w.f2.f$s.ce.audiocodeext add command -label $i \
					-command "set ::streams::audio_code_ext_$s \"$i\""
		}
		#label  $w.f2.f$s.d -text [audio_description $aattr false $s]
		#pack   $w.f2.f$s.d -side left
		button  $w.f2.f$s.d -text [audio_description $aattr false $s] -pady 0 \
				-highlightthickness 1 -anchor w -command \
				"set ::streams::tmp_aattrs($s) \[::streams::get_audio_modes \$::streams::tmp_aattrs($s) $pgctype $vtsnum $s $w\]; \
				$w.f2.f$s.d configure -text \[::streams::audio_description \$::streams::tmp_aattrs($s) false $s\]"
		pack   $w.f2.f$s.d -side left -padx 2 -pady 0 -fill x -expand 1
	}

	# SUBPIC

	frame  $w.f3 -borderwidth 2 -relief groove
	pack   $w.f3 -padx 4 -pady 2 -ipady 1 -fill x
	set break 16
	if {$numsubpics == 0} {
		label $w.f3.lab -text "No Sub-Picture streams"
		pack  $w.f3.lab
	} else {
		frame  $w.f3.pad -borderwidth 1
		pack   $w.f3.pad
		frame  $w.f3.f1
		pack   $w.f3.f1 -side left
		frame  $w.f3.f2
		pack   $w.f3.f2 -side left -fill y
		if {$numsubpics > 16} {
			set break [expr {($numsubpics+1)/2}]
		}
	}
	for {set s 0} {$s < $numsubpics} {incr s} {
		set offset [expr {$astart+0X56+($s*6)}]
		set sattr [string range $mat_table $offset [expr {$offset + 5}]]
		set attrlist [subpic_attributes $sattr]
		set ::streams::subpic_lang_code_$s [lindex $attrlist 0]
		if {[lindex $attrlist 0] == "--"} {
			set ::streams::subpic_lang_code_$s ""
		}
		set ::streams::subpic_code_ext_$s [lindex $attrlist 1]

		set fr $w.f3.f1
		if {$s >= $break} {set fr $w.f3.f2}
		frame  $fr.f$s
		pack   $fr.f$s -fill x
		label  $fr.f$s.lab -text "SubPic $s:" -width 9 -anchor w
		pack   $fr.f$s.lab -side left
		entry  $fr.f$s.lc -textvariable ::streams::subpic_lang_code_$s -width 2 \
				 -font $::config(fixed_width_font) -bd 1 \
				 -validate all -validatecommand "::streams::validate_langcode %P"
		pack   $fr.f$s.lc -side left -padx 2
		menubutton $fr.f$s.ce -textvariable ::streams::subpic_code_ext_$s \
				-pady 0 -width 22 -menu $fr.f$s.ce.subpiccodeext
		pack   $fr.f$s.ce -side left -padx 2
		menu $fr.f$s.ce.subpiccodeext -tearoff 0
		#label  $fr.f$s.id -text "ID: 0x[format "%02X" [expr {(0x20+$s)}]]" -width 9 -anchor w
		#pack   $fr.f$s.id -side left
		foreach i $::streams::subpic_code_ext_list {
			if {$i != "-"} {
				$fr.f$s.ce.subpiccodeext add command -label $i \
						-command "set ::streams::subpic_code_ext_$s \"$i\""
			}
		}
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	button $w.fb.lc -text " ISO639 language codes " -command "::main::showtextfile .langcodes \
			{ISO639 language codes} {$::config(apppath)/lib/data/lang_codes.txt} grab"
	pack   $w.fb.lc -side left -padx 3
	frame  $w.fb.pad
	pack   $w.fb.pad -expand 1 -fill x -side left
	button $w.fb.ok -text "OK" -width 10 -command {set ::streams::done 1}
	pack   $w.fb.ok -side left -padx 3
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::streams::done 0}
	pack   $w.fb.cancel -side left -padx 2

	# window geometry
	update
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-$ww)/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-$wh)/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 1 1

	wm deiconify $w
	focus $w
	grab set $w

	set cont true
	set ::streams::done -1
	while {$cont==true} {
		set cont false
		tkwait variable ::streams::done
		if {$::streams::done} {
			for {set s 0} {$s < $numaudios} {incr s} {
				set tmp ::streams::audio_lang_code_$s
				if {[string length [subst $$tmp]]==1} {
					focus $w.f2.f$s.lc
					bell
					set cont true
					break
				}
			}
			if {! $cont} {
				for {set s 0} {$s < $numsubpics} {incr s} {
					set tmp ::streams::subpic_lang_code_$s
					if {[string length [subst $$tmp]]==1} {
						set ent $w.f3.f1.f$s.lc
						if {$s >= $break} {set ent $w.f3.f2.f$s.lc}
						focus $ent
						bell
						set cont true
						break
					}
				}
			}
		}
	}
	grab release $w
	if {$parentwin != "."} {
		grab set $parentwin
		focus $parentwin
	}
	destroy $w
	if {$::streams::done} {
		::main::dvd_modified true false

		# VIDEO
		set oldval [expr {[get_int mat_table $astart 2] & 0XF0FD}]
		set_int mat_table $astart [expr {$oldval | $::streams::screenformat | \
				$::streams::autoletterbox | $::streams::autopanscan | \
				$::streams::lbcropped}] 2

		# AUDIO
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr {$astart+4+($s*8)}]
			set aattr $::streams::tmp_aattrs($s)
			set byte0 [get_int aattr 0 1]
			set byte5 [get_int aattr 5 1]

			set oldval [expr {$byte0 & 0XF3}]
			set tmp ::streams::audio_lang_code_$s
			set lc [subst $$tmp]
			if {[string length $lc]==2} {
				set_int aattr 0 [expr {$oldval+4}] 1
				set aattr "[string range $aattr 0 1]$lc[string range $aattr 4 7]"
			} else {
				if {[string length $lc]!=0} bell
				set_int aattr 0 $oldval 1
			}

			set tmp ::streams::audio_code_ext_$s
			set ce [subst $$tmp]
			if {$ce == $::streams::illegalextlabel} {
				set_int aattr 5 0 1
			} else {
				set_int aattr 5 [lsearch $::streams::audio_code_ext_list $ce] 1
			}

			set mat_table "[string range $mat_table 0 [expr {$offset-1}]]$aattr[string range $mat_table [expr {$offset+8}] end]"
		}

		# SUBPIC
		for {set s 0} {$s < $numsubpics} {incr s} {
			set offset [expr {$astart+0X56+($s*6)}]
			set sattr [string range $mat_table $offset [expr {$offset + 5}]]
			set byte0 [get_int sattr 0 1]
			set oldval [expr {$byte0 & 0XFC}]
			set tmp ::streams::subpic_lang_code_$s
			set lc [subst $$tmp]
			if {[string length $lc]==2 && $lc != "\000\000" && $lc != "  " && $lc != "--"} {
				set_int sattr 0 [expr {$oldval+1}] 1
				set sattr "[string range $sattr 0 1]$lc[string range $sattr 4 5]"
			} else {
				if {[string length $lc]!=0} bell
				set_int sattr 0 $oldval 1
			}

			set tmp ::streams::subpic_code_ext_$s
			set ce [subst $$tmp]
			if {$ce == $::streams::illegalextlabel || $ce == "-"} {
				set_int sattr 5 0 1
			} else {
				set_int sattr 5 [lsearch $::streams::subpic_code_ext_list $ce] 1
			}

			set mat_table "[string range $mat_table 0 [expr {$offset-1}]]$sattr[string range $mat_table [expr {$offset+6}] end]"
		}
		unset -nocomplain ::streams::tmp_aattrs

		if {$vtsnum == 0} {
			set ::raw_sectors(0,VMGM_MAT) $mat_table
		} else {
			set ::raw_sectors($vtsnum,VTSI_MAT) $mat_table
		}
	}
}

proc video_attributes {vattr} {

	if {[expr {$vattr & 0X4000}] == 0} {
		set coding_mode "MPEG-1"
	} else {
		set coding_mode "MPEG-2"
	}

	if {[expr {$vattr & 0X1000}] == 0} {
		set standard "NTSC"
	} else {
		set standard "PAL"
	}

	if {[expr {$vattr & 0X0C00}] == 0} {
		set aspect "4:3"
	} else {
		set aspect "16:9"
	}

	if {[expr {$vattr & 0X0020}] == 0} {
		set bitrate "VBR"
	} else {
		set bitrate "CBR"
	}

	set m [expr {($vattr & 0X0300) / 0X100}]
	switch $m {
		3 {set auto_mode ""}
		2 {set auto_mode "Automatic Letterbox"}
		1 {set auto_mode "Automatic Pan&Scan"}
		0 {set auto_mode "Automatic Letterbox & Pan&Scan"}
		default {set auto_mode "Unknow Automatic mode #$m"}
	}

	if {$standard == "NTSC"} {
		set h 480
	} else {
		set h 576
	}
	set r [expr {($vattr & 0X0018) / 8}]
	switch $r {
		0 {set w 720}
		1 {set w 704}
		2 {set w 352}
		3 {set w 352; set h [expr {$h / 2}]}
		default {set w 720}
	}
	set resolution ${w}x${h}

	if {[expr {$vattr & 0X0004}] == 0} {
		set letterboxed "not cropped"
	} else {
		set letterboxed "top & bottom cropped"
	}

	return [list $aspect $standard $resolution $letterboxed $coding_mode $bitrate $auto_mode]
}
proc video_description {vattr showall} {
	set lst [video_attributes $vattr]
	if {$showall} {
		set sep ", "
	} else {
		set sep "\n"
	}
	if {$showall} {
		set txt "[lindex $lst 0] "
	} else {
		set txt ""
	}
	set txt "${txt}[lindex $lst 1] ([lindex $lst 2])${sep}[lindex $lst 4] ([lindex $lst 5])"

	if {$showall} {
		set auto_mode [lindex $lst 6]
		if {$auto_mode != ""} {
			set txt "$txt, $auto_mode"
			set txt "$txt (letterbox: [lindex $lst 3])"
		}
	}
	return $txt
}


proc audio_attributes {aattr streamnum} {

	set byte0 [get_int aattr 0 1]

	set c [expr {($byte0 & 0XE0) / 0X20}]
	switch $c {
		0 {	set coding_mode "AC3"
			#set id [format "0x%02X" [expr {(0x80) | $streamnum}]]
		}
		2 {	set coding_mode "MPEG"
			#set id [format "0x%02X" [expr {(0xC0) | $streamnum}]]
		}
		3 {set coding_mode "MPEG2-ext"
			#set id [format "0x%02X" [expr {(0xC8) | $streamnum}]]
		}
		4 {set coding_mode "LPCM"
			#set id [format "0x%02X" [expr {(0xA0) | $streamnum}]]
		}
		6 {set coding_mode "DTS"
			#set id [format "0x%02X" [expr {(0x88) | $streamnum}]]
		}
		7 {set coding_mode "SDDS"
			#set id [format "0x%02X" [expr {(0x90) | $streamnum}]]
		}
		default {set coding_mode "Unknow coding mode #$c"
			#set id "0x??"
		}
	}

	if {[expr {$byte0 & 0X0C}] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $aattr 2 3]
	}

	set a [expr {$byte0 & 0X03}]
	switch $a {
		0 {set applic_mode "Unspecified application mode"}
		1 {set applic_mode "Karaoke"}
		2 {
			set byte7 [get_int aattr 7 1]
			if {[expr {$byte7 & 8}] == 8} {
				set applic_mode "Dolby-Surround"
			} else {
				set applic_mode "Surround"
			}
		}
		3 {set applic_mode "Unknow applic. mode #$a"}
	}

	set byte1 [get_int aattr 1 1]

	set num_channels [expr {($byte1 & 7) + 1}]

	set byte5 [get_int aattr 5 1]
	if {$byte5 < 5} {
		set code_ext [lindex $::streams::audio_code_ext_list $byte5]
	} else {
		set code_ext $::streams::illegalextlabel
	}

	#return [list $lang_code $code_ext $coding_mode $num_channels $applic_mode $id]
	return [list $lang_code $code_ext $coding_mode $num_channels $applic_mode]
}
proc audio_description {vattr showall streamnum} {
	set lst [audio_attributes $vattr $streamnum]
	if {$showall} {
		set txt "[lindex $lst 0], [lindex $lst 1], "
	} else {
		set txt ""
	}
	set txt "${txt}[lindex $lst 2] ([lindex $lst 3] channel"
	if {[lindex $lst 3] > 1} {
		set txt "${txt}s)"
	} else {
		set txt "${txt})"
	}
	if {$showall} {
		set txt "$txt, [lindex $lst 4]"
	}
	#set txt "$txt, ID: [lindex $lst 5]"
	return $txt
}

proc subpic_attributes {spattr} {
	set byte0 [get_int spattr 0 1]
	if {[expr {$byte0 & 0x03}] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $spattr 2 3]
	}

	set byte5 [get_int spattr 5 1]
	if {$byte5 < 16} {
		set code_ext [lindex $::streams::subpic_code_ext_list $byte5]
		if {$code_ext == "-"} {
			set code_ext $::streams::illegalextlabel
		}
	} else {
		set code_ext $::streams::illegalextlabel
	}

	return [list $lang_code $code_ext]
}
proc subpic_description {vattr} {
	set lst [subpic_attributes $vattr]
	return "[lindex $lst 0], [lindex $lst 1]"
}

set valid_lang_codes {
	aa ab af am ar as ay az ba be bg bh bi bn bo br ca co cs cy da de dz el en eo es et
	eu fa fi fj fo fr fy ga gd gl gn gu ha hi hr hu hy ia ie ik in is it iw ja ji jw
	ka kk kl km kn ko ks ku ky la ln lo lt lv mg mi mk ml mn mo mr ms mt my na ne nl
	no oc om or pa pl ps pt qu rm rn ro ru rw sa sd sg sh si sk sl sm sn so sq sr ss
	st su sv sw ta te tg th ti tk tl tn to tr ts tt tw uk ur uz vi vo wo xh yo zh zu
}

proc validate_langcode {newcode} {
	if {$newcode==""} {
		return true
	}
	set r [valid_langcode $newcode]
	if {! $r} bell
	return $r
}

proc valid_langcode {newcode} {
	if {$newcode == ""} {return false}
	if {[string length $newcode]>2} {return false}
	if {! [string is alpha $newcode]} {return false}
	if {! [string is lower $newcode]} {return false}
	if {[string length $newcode]==2 && [lsearch $::streams::valid_lang_codes $newcode] < 0} {
		return false
	}
	return true
}


####### main proc
proc get_pgc_num_audio {pgctype pgcindex} {
	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return 0
		} else {
			return [get_int ::raw_sectors(0,VMGM_MAT) 0X102 2]
		}
	} else {
		if {$pgctype == "menu"} {
			return [get_int ::raw_sectors($vts,VTSI_MAT) 0X102 2]
		} else {
			return [get_int ::raw_sectors($vts,VTSI_MAT) 0X202 2]
		}
	}
}
proc get_pgc_audio_descr {pgctype pgcindex streamnum} {
	set streamnum [expr {$streamnum & 0x0F}]
	if {$streamnum == 15} {return "15: Not set"}
	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return ""
		} else {
			set data $::raw_sectors(0,VMGM_MAT)
			set numaudios [get_int data 0X102 2]
			set offset [expr {(0X104) + ( $streamnum * 8 )}]
		}
	} else {
		if {$pgctype == "menu"} {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numaudios [get_int data 0X102 2]
			set offset [expr {(0X104 + ( $streamnum * 8 ))}]
		} else {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numaudios [get_int data 0X202 2]
			set offset [expr {(0X204 + ( $streamnum * 8 ))}]
		}
	}
	set attr [string range $data $offset [expr {$offset + 7}]]

	set lst [audio_attributes $attr $streamnum]
	return "[expr {$streamnum & 0x07}]: [lindex $lst 0], [lindex $lst 1] ([lindex $lst 2])"
}
proc get_pgc_num_subpic {pgctype pgcindex} {
	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return 0
		} else {
			return [get_int ::raw_sectors(0,VMGM_MAT) 0X154 2]
		}
	} else {
		if {$pgctype == "menu"} {
			return [get_int ::raw_sectors($vts,VTSI_MAT) 0X154 2]
		} else {
			return [get_int ::raw_sectors($vts,VTSI_MAT) 0X254 2]
		}
	}
}
proc get_pgc_subpic_descr {pgctype pgcindex streamnum} {
	set streamnum [expr {$streamnum & 0x7F}]
	if {$streamnum == 62} {return "62: Not set"}
	if {$streamnum == 63} {return "63: Forced"}
	if {[expr {$streamnum & 0x40}] == 0} {
		set mode "Off"
	} else {
		set mode "On"
	}
	set streamnum [expr {$streamnum & 0x1F}]

	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return ""
		} else {
			set data $::raw_sectors(0,VMGM_MAT)
			set numsubpics [get_int data 0X154 2]
			set offset [expr {(0X156 + ( $streamnum * 6 ))}]
		}
	} else {
		if {$pgctype == "menu"} {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numsubpics [get_int data 0X154 2]
			set offset [expr {(0X156 + ( $streamnum * 6 ))}]
		} else {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numsubpics [get_int data 0X254 2]
			set offset [expr {(0X256 + ( $streamnum * 6 ))}]
		}
	}
	set attr [string range $data $offset [expr {$offset + 5}]]
	return "$streamnum $mode: [subpic_description $attr]"
}

# ::streams::verify_num_streams
proc verify_num_streams {} {
	if {! [info exists ::dvddir]} {
		bell
		return ""
	}
	set numerrors 0
	set numfixed  0
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		set rc [verify_num_streams_domain "menu" $vts]
		incr numerrors [lindex $rc 0]
		incr numfixed  [lindex $rc 1]

		if {$vts > 0} {
			set rc [verify_num_streams_domain "pgc" $vts]
			incr numerrors [lindex $rc 0]
			incr numfixed  [lindex $rc 1]
		}
	}
	show_msg "Fix Streams:  Fixed $numfixed / $numerrors discrepancies."
}

# ::streams::verify_num_streams_domain
proc verify_num_streams_domain {type vts} {
	set numerrors 0
	set numfixed 0
	if {$type == "menu"} { 
		if {$::menus($vts,numlus) > 0} {
			set numaudios  [get_pgc_num_audio  menu $vts,1,1]
			set maxaudios  0
			set numsubpics [get_pgc_num_subpic menu $vts,1,1]
			set maxsubpics 0
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					if {$::menus($vts,$lu,$pgc,numcells) > 0} {
						set idxa 12
						for {set a 1} {$a <= 8} {incr a} {
							if {[expr {[get_int ::menus($vts,$lu,$pgc,header) $idxa 1] & 0x80}] > 0 \
									&& $a > $maxaudios} {
								set maxaudios $a
							}
							incr idxa 2
						}
						if {[expr {[get_int ::menus($vts,$lu,$pgc,header) 0x1C 1] & 0x80}] > 0} {
							set maxsubpics 1
						}
					}
				}
			}
			if {$vts == 0} {
				set vtsname "VMGM"
				set vtstable "VMGM_MAT"
			} else {
				set vtsname "VTST $vts"
				set vtstable "VTSI_MAT"
			}
			if {$numaudios != $maxaudios} {
				incr numerrors
				set txt "There are $numaudios audio streams defined for the Menu domain"
				set txt "$txt\nin the $vtstable table of the $vtsname."
				set txt "$txt\n$maxaudios audio streams are really used by the PGCs."
				set txt "$txt\n\nFix the number of audio streams in $vtstable to $maxaudios?"
				set r [tk_messageBox -title "Verify number of streams" -icon warning \
					-message $txt -parent . -type yesno -default yes]
				if {$r == "yes"} {
					incr numfixed
					set_int ::raw_sectors($vts,$vtstable) 0X102 $maxaudios 2
					::main::dvd_modified true true
				}
			}
			if {$numsubpics != $maxsubpics} {
				incr numerrors
				if {$maxsubpics == 0} {
					set txt "There is one subpic stream defined for the Menu domain"
					set txt "$txt\nin the $vtstable table of the $vtsname."
					set txt "$txt\nThis subpic stream is not really used by the PGCs."
				} else {
					set txt "There is no subpic stream defined for the Menu domain"
					set txt "$txt\nin the $vtstable table of the $vtsname."
					set txt "$txt\nHowever, there is one subpic used by the PGCs."
				}
				set txt "$txt\n\nFix the number of subpic streams in $vtstable to $maxsubpics?"
				set r [tk_messageBox -title "Verify number of streams" -icon warning \
					-message $txt -parent . -type yesno -default yes]
				if {$r == "yes"} {
					incr numfixed
					set_int ::raw_sectors($vts,$vtstable) 0X154 $maxsubpics 2
					::main::dvd_modified true true
				}
			}
		}
	} elseif {$vts > 0} {
		set numaudios  [get_pgc_num_audio  pgc $vts,1]
		set maxaudios  0
		set numsubpics [get_pgc_num_subpic pgc $vts,1]
		set maxsubpics 0
		for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
			if {$::pgcs($vts,$pgc,numcells) > 0} {
				set idxa 12
				for {set a 1} {$a <= 8} {incr a} {
					if {[expr {[get_int ::pgcs($vts,$pgc,header) $idxa 1] & 0x80}] > 0 \
							&& $a > $maxaudios} {
						set maxaudios $a
					}
					incr idxa 2
				}
				set idxs 28
				for {set s 1} {$s <= 32} {incr s} {
					if {[expr {[get_int ::pgcs($vts,$pgc,header) $idxs 1] & 0x80}] > 0 \
							&& $s > $maxsubpics} {
						set maxsubpics $s
					}
					incr idxs 4
				}
			}
		}
		if {$numaudios != $maxaudios} {
			incr numerrors
			set txt "There are $numaudios audio streams defined for the Title domain"
			set txt "$txt\nin the VTSI_MAT table of the VTST $vts."
			set txt "$txt\n$maxaudios audio streams are really used by the PGCs."
			set txt "$txt\n\nFix the number of audio streams in VTSI_MAT to $maxaudios?"
			set r [tk_messageBox -title "Verify number of streams" -icon warning \
				-message $txt -parent . -type yesno -default yes]
			if {$r == "yes"} {
				incr numfixed
				set_int ::raw_sectors($vts,VTSI_MAT) 0X202 $maxaudios 2
				::main::dvd_modified true true
			}
		}
		if {$numsubpics != $maxsubpics} {
			incr numerrors
			set txt "There are $numsubpics subpic streams defined for the Title domain"
			set txt "$txt\nin the VTSI_MAT table of the VTST $vts."
			set txt "$txt\n$maxsubpics subpic streams are really used by the PGCs."
			set txt "$txt\n\nFix the number of subpic streams in VTSI_MAT to $maxsubpics?"
			set r [tk_messageBox -title "Verify number of streams" -icon warning \
				-message $txt -parent . -type yesno -default yes]
			if {$r == "yes"} {
				incr numfixed
				set_int ::raw_sectors($vts,VTSI_MAT) 0X254 $maxsubpics 2
				::main::dvd_modified true true
			}
		}
	}
	return [list $numerrors $numfixed]
}

# set aattr [string range $::raw_sectors(1,VTSI_MAT) 0x204 0x20b]
# ::streams::get_audio_modes $aattr pgc 1 0 0 false .
proc get_audio_modes {aattr type vts streamnum parentwin} {
	set byte0 [get_int aattr 0 1]
	set ::streams::audio(codingmode) [expr {$byte0 & 0xE0}]
	#set ::streams::audio(multichext) [expr {$byte0 & 0x10}]
	#set ::streams::audio(langtype)   [expr {$byte0 & 0x0C}]
	set ::streams::audio(appmode)    [expr {$byte0 & 0x03}]

	set byte1 [get_int aattr 1 1]
	set ::streams::audio(quantization) [expr {$byte1 & 0xC0}]
	set ::streams::audio(samplerate)   [expr {$byte1 & 0x30}]
	set ::streams::audio(nchannels)    [expr {($byte1 & 0x07)+1}]

	set ::streams::audio(surroundmode) [get_int aattr 7 1]

	set w ".audiosetup"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm protocol $w WM_DELETE_WINDOW {set ::streams::audiodone 0}
	bind $w <Escape> "set ::streams::audiodone 0"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm title $w "Creation of Audio Stream"

	# AUDIO

	if {$vts == 0} {
		set domain "VMGM menu domain"
	} else {
		if {$type == "menu"} {
			set domain "VTSM $vts menu domain"
		} else {
			set domain "VTST $vts title domain"
		}
	}
	label  $w.title -text "Audio stream # $streamnum, $domain"
	pack   $w.title -pady 4

	if {$::streams::audio(appmode) == 1} {	
		label  $w.ko -text "WARNING: Current mode is Karaoke.\nThere is no support to edit the karaoke settings.\nPress Cancel to exit without changes." -foreground red
		pack   $w.ko -pady 2
	}

	frame  $w.f2 -borderwidth 2 -relief groove
	pack   $w.f2 -padx 4 -pady 2

	frame  $w.f2.cm
	pack   $w.f2.cm -anchor w
	label  $w.f2.cm.l -text "Coding mode:" -width 16 -anchor w
	pack   $w.f2.cm.l -side left
	radiobutton $w.f2.cm.ac3 -text AC3 -variable ::streams::audio(codingmode) -value 0
	pack   $w.f2.cm.ac3 -side left
	radiobutton $w.f2.cm.mpeg -text MPEG -variable ::streams::audio(codingmode) -value 64
	pack   $w.f2.cm.mpeg -side left
	radiobutton $w.f2.cm.m2e -text MPEG2-ext -variable ::streams::audio(codingmode) -value 96
	pack   $w.f2.cm.m2e -side left
	radiobutton $w.f2.cm.lpcm -text LPCM -variable ::streams::audio(codingmode) -value 128
	pack   $w.f2.cm.lpcm -side left
	radiobutton $w.f2.cm.dts -text DTS -variable ::streams::audio(codingmode) -value 192
	pack   $w.f2.cm.dts -side left
	radiobutton $w.f2.cm.sdds -text SDDS -variable ::streams::audio(codingmode) -value 224
	pack   $w.f2.cm.sdds -side left

	frame  $w.f2.am
	pack   $w.f2.am -anchor w
	label  $w.f2.am.l -text "Application mode:" -width 16 -anchor w
	pack   $w.f2.am.l -side left
	radiobutton $w.f2.am.ns -text "Not specified" -variable ::streams::audio(appmode) -value 0
	pack   $w.f2.am.ns -side left
	radiobutton $w.f2.am.ka -text "Karaoke" -variable ::streams::audio(appmode) -value 1 -state disabled
	pack   $w.f2.am.ka -side left
	radiobutton $w.f2.am.as -text "Surround" -variable ::streams::audio(appmode) -value 2
	pack   $w.f2.am.as -side left
	checkbutton $w.f2.am.ds -text "Dolby surround" -variable ::streams::audio(surroundmode) -onvalue 8 -offvalue 0
	pack   $w.f2.am.ds -side left -padx 8

	frame  $w.f2.qu
	pack   $w.f2.qu -anchor w
	label  $w.f2.qu.l -text "Quantization/DRC:" -width 16 -anchor w
	pack   $w.f2.qu.l -side left
	radiobutton $w.f2.qu.b16 -text "16bps" -variable ::streams::audio(quantization) -value 0
	pack   $w.f2.qu.b16 -side left
	radiobutton $w.f2.qu.b20 -text "20bps" -variable ::streams::audio(quantization) -value 64
	pack   $w.f2.qu.b20 -side left
	radiobutton $w.f2.qu.b24 -text "24bps" -variable ::streams::audio(quantization) -value 128
	pack   $w.f2.qu.b24 -side left
	radiobutton $w.f2.qu.drc -text "DRC (Dynamic Range Control)" -variable ::streams::audio(quantization) -value 192
	pack   $w.f2.qu.drc -side left

	frame  $w.f2.sr
	pack   $w.f2.sr -anchor w
	label  $w.f2.sr.l -text "Sample rate:" -width 16 -anchor w
	pack   $w.f2.sr.l -side left
	radiobutton $w.f2.sr.b48 -text "48Ksps" -variable ::streams::audio(samplerate) -value 0
	pack   $w.f2.sr.b48 -side left
	radiobutton $w.f2.sr.b96 -text "96Ksps" -variable ::streams::audio(samplerate) -value 16
	pack   $w.f2.sr.b96 -side left

	frame  $w.f2.ch
	pack   $w.f2.ch -anchor w ;#-fill x -expand 1
	label  $w.f2.ch.l -text "Channels:" -width 16 -anchor w
	pack   $w.f2.ch.l -side left
	label  $w.f2.ch.ls -textvariable ::streams::audio(nchannels) -width 1 -padx 6
	pack   $w.f2.ch.ls -side left
	scale  $w.f2.ch.s -from 1 -to 6 -variable ::streams::audio(nchannels) -showvalue false
	pack   $w.f2.ch.s -side left ;#-fill x -expand 1

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::streams::audiodone 0}
	pack   $w.fb.cancel -side right -padx 2
	button $w.fb.ok -text "OK" -width 10 -command {set ::streams::audiodone 1}
	pack   $w.fb.ok -side right -padx 3

	# window geometry
	update
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr {[winfo rootx .]+(([winfo width  .]-$ww)/2)}]
	set wy [expr {[winfo rooty .]+(([winfo height .]-$wh)/2)-25}]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 0 1

	wm deiconify $w
	focus $w
	grab set $w

	set ::streams::audiodone -1
	tkwait variable ::streams::audiodone
	if {$::streams::audiodone} {
		set byte0 [expr {$::streams::audio(codingmode) + 4 + $::streams::audio(appmode)}]
		set_int aattr 0 $byte0 1
	
		set byte1 [expr {$::streams::audio(quantization) + $::streams::audio(samplerate) + $::streams::audio(nchannels)-1}]
		set_int aattr 1 $byte1 1

		if {$::streams::audio(appmode) != 1} {	
			set_int aattr 7 $::streams::audio(surroundmode) 1
		}
	}

	grab release $w
	destroy $w
	if {$parentwin != "."} {
		grab set $parentwin
	}
	focus $parentwin
	return $aattr
}

}	;# end namespace streams
 