#############################################################################
# cmdedit.tcl Copyright 2004, 2005 r0lZ
# The command editor window stuff.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::ced {

namespace export cmdeditor

#############################################################################
#############################################################################
## UTILITY PROCEDURES
#############################################################################
#############################################################################

#############################################################################
## Ajuste la hauteur de la window en fonction de la commande courrante

proc adjust_win_size {} {
	if {$::config(adjust_win_size)} {
		update
		set width  [winfo width .cmdeditor]
		set height [winfo reqheight .cmdeditor]
		wm geometry .cmdeditor ${width}x$height
	}
}

#############################################################################
## Recherche infos sur la commande courrante sur Internet

proc launch_web_info {} {
	set cmd [string tolower [$::ced::widget(CommandName) cget -text]]
	if {$::config(web_browser) == ""} return
	if {$cmd == ""} return
	set cmd [string map {"_" "" "cell" "c"} [lindex [split $cmd "-"] 0]]

	switch $cmd {
		"set" {set cmd "setgprm"}
		"setclnk" {set cmd "type31"}
		"csetclnk" {set cmd "type32"}
		"csetlnk" {set cmd "type33"}
	}

	catch {
		exec -- $::config(web_browser) "http://www.dvd-replica.com/DVD/${cmd}.php" &
	}
}

#############################################################################
## return the number provided forced to the valid range

proc force_range {num min max} {
	#dprint 3 "> force_range {num min max}"
	if {$num < $min} {set num $min}
	if {$num > $max} {set num $max}
	return $num
}

#############################################################################
## set a scale widget in disabled or normal state
proc scale_state {w state} {
	switch $state {
		enabled -
		normal {
			$w configure -state normal -troughcolor white \
				-sliderlength 15 -foreground black
		}
		disabled {
			$w configure -state disabled -troughcolor $::config(systembuttonface) \
				-sliderlength 1 -foreground gray
		}
	}
}

#############################################################################
## Parse DVD hex command and put the disassembled string in the disasm label.

proc hexentry_disassemble {args} {
	set hex [c_hex_entry_get_ascii $::ced::widget(hex_entry) false]
	if {$hex != $::ced::old_hex_for_disassemble} {
		set ::ced::old_hex_for_disassemble $hex
		set txt [::disasm::disasm $hex]
		if {[string first "ILLEGAL" $txt] >= 0} {
			set color pink
		} else {
			set color $::config(systembuttonface)
		}
		$::ced::widget(disasm) configure -text $txt -background $color
	}
}

#############################################################################
## Parse DVD hex command and put the disassembled string in the disasm label.

proc hexcmd_disasm {args} {
	if {$::hexcommand != $::ced::old_hex_for_disassemble} {
		set ::ced::old_hex_for_disassemble $::hexcommand
		set txt [::disasm::disasm $::hexcommand]
		if {[string first "ILLEGAL" $txt] >= 0} {
			set color pink
		} else {
			set color $::config(systembuttonface)
		}
		$::ced::widget(disasm) configure -text $txt -background $color
	}
}

#############################################################################
## Parse DVD hex command, put the params in the global variables,
## and call the specific command interface (or cmd_illegal_command).
## The 'args' argument is present to allow this proc to be called directly by
## a widget, but is not used by the proc.

proc hexcmd_to_gui {args} {
	#dprint 3 "> hexcmd_to_gui {}"

	destroy $::ced::widget(cmdFrame)

	set ::ced::old_hexcommand $::hexcommand

	set linkerr 0

	set cmdtype [peek_hex_nibble 0 0]
	set cmdname ""
	switch $cmdtype {
		0 {
			# special/flow control commands
			set subtype [peek_hex_nibble 1 1]
			switch $subtype {
				0 {
					# NOP
					h2gui:if1
					cmd_NOP
				}
				1 {
					# Goto
					h2gui:if1
					set ::ced::c_goto(linenum) [peek_hex_byte 7]
					cmd_GoTo
				}
				2 {
					# Break
					h2gui:if1
					cmd_Break
				}
				3 {
					# SetTmpPML
					h2gui:if1
					set level [peek_hex_nibble 6 1]
					if {$level == 15} {
						set ::ced::c_settmppml(none) true
					} else {
						set ::ced::c_settmppml(none) false
						set ::ced::c_TmpPNL(level) $level
					}
					set ::ced::c_goto(linenum) [peek_hex_byte 7]
					cmd_SetTmpPML
				}
				default {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "flow control" \
							"byte 1 lower nibble must be 0, 1, 2 or 3"
					return
				}
			}
		}
		2 {
			# link commands
			h2gui:if1
			set tmp [expr {[peek_hex_byte 6] & 0XFC}]
			if {$tmp==0} {
				set ::ced::c_link(highlightmode) false
			} else {
				set ::ced::c_link(highlightmode) true
				set ::ced::c_link(highlightbutton) [expr {$tmp/4}]
			}

			set ::ced::c_link(k)  [peek_hex_nibble 1 1]
			set ::ced::c_link(ll) [peek_hex_byte 7]
			switch $::ced::c_link(k) {
				1 {
					switch $::ced::c_link(ll) {
						0 {
							cmd_LinkNoLink
						}
						1 {
							cmd_LinkTopCell
						}
						2 {
							cmd_LinkNextCell
						}
						3 {
							cmd_LinkPrevCell
						}
						5 {
							cmd_LinkTopPG
						}
						6 {
							cmd_LinkNextPG
						}
						7 {
							cmd_LinkPrevPG
						}
						9 {
							cmd_LinkTopPGC
						}
						10 {
							cmd_LinkNextPGC
						}
						11 {
							cmd_LinkPrevPGC
						}
						12 {
							cmd_LinkGoUpPGC
						}
						13 {
							cmd_LinkTailPGC
						}
						16 {
							cmd_RSM
						}
						default {
							focus $::ced::widget(hex_entry).d_7_1
							cmd_illegal_command link \
								"when byte 1 lower nibble is 1,\nbyte 7 must be 00, 01, 02, 03, 05, 06, 07, 09, 0A, 0B, 0C, 0D or 0F"
							return
						}
					}
				}
				4 {
					# LinkPGCN
					set ::ced::c_link(pgcn) [peek_hex_word 6]
					cmd_LinkPGCN
				}
				5 {
					# LinkPTT
					set ::ced::c_link(pttn) [expr {[peek_hex_word 6] & 0X03FF}]
					cmd_LinkPTT
				}
				6 {
					# LinkPGN
					set ::ced::c_link(pgn) [peek_hex_byte 7]
					cmd_LinkPGN
				}
				7 {
					# LinkCN
					set ::ced::c_link(cell) [peek_hex_byte 7]
					cmd_LinkCN
				}
				default {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command link "byte 1 lower nibble must be 1, 4, 5, 6 or 7"
					return
				}
			}
		}
		3 {
			# jump/call commands
			h2gui:if2
			set cell [peek_hex_byte 4]
			if {$cell == 0} {
				set ::ced::c_call(resume_mode) false
			} else {
				set ::ced::c_call(resume_cell) $cell
				set ::ced::c_call(resume_mode) true
			}
			set subtype [peek_hex_nibble 1 1]
			switch $subtype {
				1 {
					# Exit
					cmd_Exit
				}
				2 {
					# JumpTT
					set ::ced::c_jump(title) [peek_hex_byte 5]
					cmd_JumpTT
				}
				3 {
					# JumpVTS_TT
					set ::ced::c_jump(title) [peek_hex_byte 5]
					cmd_JumpVTS_TT
				}
				5 {
					# JumpVTS_PTT
					set ::ced::c_jump(ptt)   [peek_hex_word 2]
					set ::ced::c_jump(title) [peek_hex_byte 5]
					cmd_JumpVTS_PTT
				}
				6 {
					# JumpSS *
					set subtype [peek_hex_nibble 5 0]
					switch $subtype {
						0 {
							# JumpSS-FP
							cmd_JumpSS-FP
						}
						4 {
							# JumpSS-VMGM-menu
							cmd_JumpSS-VMGM-menu
						}
						8 {
							# JumpSS-VTSM
							set ::ced::c_jump(menunumber) [peek_hex_nibble 5 1]
							set ::ced::c_jump(title) [peek_hex_byte 3]
							set ::ced::c_jump(vts)   [peek_hex_byte 4]
							cmd_JumpSS-VTSM
						}
						12 {
							# JumpSS-VMGM-pgc
							set ::ced::c_jump(pgc) [peek_hex_word 2]
							cmd_JumpSS-VMGM-pgc
						}
						default {
							focus $::ced::widget(hex_entry).d_5_1
							cmd_illegal_command JumpSS \
									"byte 5 upper nibble must be 0, 4, 8 or C"
							return
						}
					}
				}
				8 {
					# CallSS *
					set subtype [peek_hex_nibble 5 0]
					switch $subtype {
						0 {
							# CallSS-FP
							cmd_CallSS-FP
						}
						4 {
							# CallSS-VMGM-menu
							cmd_CallSS-VMGM-menu
						}
						8 {
							# CallSS-VTSM
							set ::ced::c_jump(menunumber) [peek_hex_nibble 5 1]
							cmd_CallSS-VTSM
						}
						12 {
							# CallSS-VMGM-pgc
							set ::ced::c_jump(pgc) [peek_hex_word 2]
							cmd_CallSS-VMGM-pgc
						}
						default {
							focus $::ced::widget(hex_entry).d_5_1
							cmd_illegal_command CallSS \
									"byte 5 upper nibble must be 0, 4, 8 or C"
							return
						}
					}
				}
				default {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "call/jump" \
							"byte 1 lower nibble must be 1, 2, 3, 5, 6 or 8"
					return
				}
			}
		}
		4 -
		5 {
			# Set*(SPRM) commands
			if {$cmdtype == 4} {
				set ::ced::c_set(sourcemode) 2
			} else {
				set ::ced::c_set(sourcemode) 1
			}

			set subtype [peek_hex_nibble 0 1]
			switch $subtype {
				1 {set cmdname SetSTN}
				2 {set cmdname SetNVTMR}
				3 {set cmdname SetGPRMMD}
				4 {set cmdname SetAMXMD}
				6 {set cmdname SetHL_BTN}
				default {
					focus $::ced::widget(hex_entry).d_0_2
					cmd_illegal_command "set (SPRM)" \
							"byte 0 lower nibble must be 1, 2, 3, 4 or 6"
					return
				}
			}

			switch [h2gui:link 1] {
				1 {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "$cmdname sub-link" \
						"byte 1 lower nibble must be 1, 4, 5, 6 or 7 (or 0 to disable the link)"
					return
				}
				2 {
					focus $::ced::widget(hex_entry).d_7_1
					cmd_illegal_command "$cmdname sub-link" \
						"when byte 1 lower nibble is 1,\nbyte 7 must be 00, 01, 02, 03, 05, 06, 07, 09, 0A, 0B, 0C, 0D or 0F"
					return
				}
			}

			h2gui:if2
			if {$::ced::c_if(check) && $::ced::c_link(check)} {
				focus $::ced::widget(hex_entry).d_1_1
				cmd_illegal_command "sub-link or if" \
					"The $cmdname command could not have both the 'If' and the 'Link' parts enabled together:\nthe upper or lower nibble of Byte 1 must be set to 0 to disable either the If or the Link."
				return
			}

			switch $subtype {
				1 {
					# SetSTN
					set tmp [peek_hex_byte 3]
					set audiomode [expr {$tmp & 0X80}]
					set audio [expr {$tmp - $audiomode}]
					if {$audiomode == 0} {
						set ::ced::c_setstn(audio_mode) 0
					} else {
						if {$::ced::c_set(sourcemode) == 1} {
							if {$audio == 15} {
								set ::ced::c_setstn(audio_mode) 2
							} else {
								set ::ced::c_setstn(audio_mode) 1
								set ::ced::c_setstn(audio_value) $audio
							}
						} else {
							set ::ced::c_setstn(audio_mode) 1
							set ::ced::c_setstn(audio_gprm) $audio
						}
					}
					set tmp [peek_hex_byte 4]
					set subpicmode  [expr {$tmp & 0X80}]
					set subpiconoff [expr {$tmp & 0X40}]
					set subpic [expr {$tmp - $subpicmode - $subpiconoff}]
					if {$subpicmode == 0} {
						set ::ced::c_setstn(subpic_mode) 0
					} else {
						if {$::ced::c_set(sourcemode) == 1} {
							if {$subpiconoff == 0} {
								set ::ced::c_setstn(subpic_onoff) false
							} else {
								set ::ced::c_setstn(subpic_onoff) true
							}
							switch $subpic {
								62 {
									set ::ced::c_setstn(subpic_mode) 2
								}
								63 {
									set ::ced::c_setstn(subpic_mode) 3
								}
								default {
									set ::ced::c_setstn(subpic_mode) 1
									set ::ced::c_setstn(subpic_value) $subpic
								}
							}
						} else {
							set ::ced::c_setstn(subpic_gprm) $subpic
							set ::ced::c_setstn(subpic_mode) 1
						}
					}
					set tmp [peek_hex_byte 5]
					set anglemode [expr {$tmp & 0X80}]
					set angle [expr {$tmp - $anglemode}]
					if {$anglemode == 0} {
						set ::ced::c_setstn(angle_mode) 0
					} else {
						set ::ced::c_setstn(angle_mode) 1
						if {$::ced::c_set(sourcemode) == 1} {
							set ::ced::c_setstn(angle_value) $angle
						} else {
							set ::ced::c_setstn(angle_gprm) $angle
						}
					}
					cmd_SetSTN
				}
				2 {
					# SetNVTMR
					if {$::ced::c_set(sourcemode) == 1} {
						# direct mode
						set ::ced::c_set(value) [peek_hex_word 2]
					} else {
						# indirect mode
						set r [peek_hex_byte 3]
						if {[expr {$r & 0X80}] == 0} {
							set ::ced::c_set(sourcemode) 2
							set ::ced::c_set(gprm) $r
						} else {
							set ::ced::c_set(sourcemode) 3
							set ::ced::c_set(sprm) [expr {$r & 0X7F}]
							if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
						}
					}
					set ::ced::c_setnvtmr(pgc) [peek_hex_word 4]
					if {$::ced::c_setnvtmr(pgc) > 0} {
						set ::ced::c_setnvtmr(pgcmode) true
					} else {
						set ::ced::c_setnvtmr(pgcmode) false
					}
					cmd_SetNVTMR
				}
				3 {
					# SetGPRMMD
					if {$::ced::c_set(sourcemode) == 1} {
						# direct mode
						set ::ced::c_set(value) [peek_hex_word 2]
					} else {
						# indirect mode
						set r [peek_hex_byte 3]
						if {[expr {$r & 0X80}] == 0} {
							set ::ced::c_set(sourcemode) 2
							set ::ced::c_set(gprm) $r
						} else {
							# operand 2 may not be a SPRM for SetGPRMMD!
							set ::ced::c_set(sourcemode) 2
							set ::ced::c_set(gprm) [expr {$r & 0X7F}]
							if {$::ced::c_set(gprm) > 15} {set ::ced::c_set(gprm) 15}
						}
					}
					set ::ced::c_setgprmmd(destination) [peek_hex_nibble 5 1]
					set ::ced::c_setgprmmd(mode) [expr {[peek_hex_nibble 5 0] & 0X8}]
					cmd_SetGPRMMD
				}
				4 {
					# SetAMXMD
					if {$::ced::c_set(sourcemode) == 1} {
						# direct mode
						set ::ced::c_set(value) [expr {[peek_hex_word 4] & 0x1C1C}]
						set ::ced::c_setamxmd_mix(20) [expr {$::ced::c_set(value) & 0x0400}]
						set ::ced::c_setamxmd_mix(30) [expr {$::ced::c_set(value) & 0x0800}]
						set ::ced::c_setamxmd_mix(40) [expr {$::ced::c_set(value) & 0x1000}]
						set ::ced::c_setamxmd_mix(21) [expr {$::ced::c_set(value) & 0x0004}]
						set ::ced::c_setamxmd_mix(31) [expr {$::ced::c_set(value) & 0x0008}]
						set ::ced::c_setamxmd_mix(41) [expr {$::ced::c_set(value) & 0x0010}]

					} else {
						# indirect mode
						set r [peek_hex_byte 5]
						if {[expr {$r & 0X80}] == 0} {
							set ::ced::c_set(sourcemode) 2
							set ::ced::c_set(gprm) $r
						} else {
							set ::ced::c_set(sourcemode) 3
							set ::ced::c_set(sprm) [expr {$r & 0X7F}]
							if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
						}
					}
					cmd_SetAMXMD
				}
				6 {
					# SetHL_BTN
					if {$::ced::c_set(sourcemode) == 1} {
						# direct mode
						set ::ced::c_set(value) [expr {[peek_hex_word 4] / 0X400}]
					} else {
						# indirect mode
						set r [peek_hex_byte 5]
						if {[expr {$r & 0X80}] == 0} {
							set ::ced::c_set(sourcemode) 2
							set ::ced::c_set(gprm) $r
						} else {
							set ::ced::c_set(sourcemode) 3
							set ::ced::c_set(sprm) [expr {$r & 0X7F}]
							if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
						}
					}
					cmd_SetHL_BTN
				}
				default {
					focus $::ced::widget(hex_entry).d_0_2
					cmd_illegal_command "set (SPRM)" \
							"byte 0 lower nibble must be 1, 2, 3, 4 or 6"
					return
				}
			}
		}
		6 -
		7 {
			# Set
			if {$cmdtype == 6} {
				set ::ced::c_set(sourcemode) 2
			} else {
				set ::ced::c_set(sourcemode) 1
			}
			switch [h2gui:link 1] {
				1 {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "$cmdname sub-link" \
						"byte 1 lower nibble must be 1, 4, 5, 6 or 7 (or 0 to disable the link)"
					return
				}
				2 {
					focus $::ced::widget(hex_entry).d_7_1
					cmd_illegal_command "$cmdname sub-link" \
						"when byte 1 lower nibble is 1,\nbyte 7 must be 00, 01, 02, 03, 05, 06, 07, 09, 0A, 0B, 0C, 0D or 0F"
					return
				}
			}
			set iftype [peek_hex_nibble 1 0]
			if {[expr {$iftype & 7}] == 0} {
				set ::ced::c_if(check) false
			} else {
				if {$::ced::c_link(check)} {
					focus $::ced::widget(hex_entry).d_1_1
					cmd_illegal_command "sub-link or if" \
						"The Set command could not have both the 'If' and the 'Link' parts enabled together:\nthe upper or lower nibble of Byte 1 must be set to 0 to disable either the If or the Link."
					return
				}
				set ::ced::c_if(check) true
				set ::ced::c_if(operatornum) [expr {$iftype & 7}]
				set ::ced::c_if(operator) [h2gui:if_operator_text $iftype]
				set op [peek_hex_byte 2]
				set ::ced::c_if(gprm1) [expr {$op & 0xF}]
				if {[expr {$iftype & 8}] == 0} {
					set op [peek_hex_byte 7]
					if {[expr {$op & 0X80}] == 0} {
						set ::ced::c_if(radioOp2) 2
						set ::ced::c_if(gprm2) $op
					} else {
						set ::ced::c_if(radioOp2) 3
						set ::ced::c_if(sprm2) [expr {$op & 0X7F}]
						if {$::ced::c_if(sprm2) > 23} {set ::ced::c_if(sprm2) 23}
					}
				} else {
					set ::ced::c_if(radioOp2) 1
					set ::ced::c_if(value2) [peek_hex_word 6]
				}
			}

			if {$cmdtype == 7} {
				# direct mode
				set ::ced::c_set(sourcemode) 1
				set ::ced::c_set(value) [peek_hex_word 4]
			} else {
				# indirect mode
				set r [peek_hex_byte 5]
				if {[expr {$r & 0X80}] == 0} {
					set ::ced::c_set(sourcemode) 2
					set ::ced::c_set(gprm) $r
				} else {
					set ::ced::c_set(sourcemode) 3
					set ::ced::c_set(sprm) [expr {$r & 0X7F}]
					if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
				}
			}
			set ::ced::c_set(destination) [peek_hex_byte 3]
			set ::ced::c_set(operator_value) [peek_hex_nibble 0 1]
			set found false
			foreach i $::ced::c_Setoperators {
				if {[lindex $i 0] == $::ced::c_set(operator_value)} {
					set ::ced::c_set(operator_text) [lindex $i 1]
					set found true
					break
				}
			}
			if {! $found} {
				set ::ced::c_set(operator_value) 1
				set ::ced::c_set(operator_text) {MOV    =}
			}
			cmd_Set
		}
		8 -
		9 {
			# SetCLnk
			if {$cmdtype == 8} {
				# indirect mode
				set r [peek_hex_byte 3]
				if {[expr {$r & 0X80}] == 0} {
					set ::ced::c_set(sourcemode) 2
					set ::ced::c_set(gprm) $r
				} else {
					set ::ced::c_set(sourcemode) 3
					set ::ced::c_set(sprm) [expr {$r & 0X7F}]
					if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
				}
			} else {
				# direct mode
				set ::ced::c_set(sourcemode) 1
				set ::ced::c_set(value) [peek_hex_word 2]
			}
			switch [h2gui:link 3] {
				1 {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "SetCLnk sub-link" \
						"byte 1 lower nibble must be 1, 4, 5, 6 or 7 (or 0 to disable the link)"
					return
				}
				2 {
					focus $::ced::widget(hex_entry).d_7_1
					cmd_illegal_command "SetCLnk sub-link" \
						"when byte 1 lower nibble is 1,\nbyte 7 must be 00, 01, 02, 03, 05, 06, 07, 09, 0A, 0B, 0C, 0D or 0F"
					return
				}
				3 {
					focus $::ced::widget(hex_entry).d_7_2
					cmd_illegal_command "sub-link" \
						"With SetCLnk, you can't use this link command:\nbyte 1 lower nibble must be 1  (or 0 to disable the link)"
					return
				}
			}
			h2gui:if1

			set ::ced::c_set(destination) [peek_hex_nibble 1 1]
			set ::ced::c_set(operator_value) [peek_hex_nibble 0 1]
			set found false
			foreach i $::ced::c_Setoperators {
				if {[lindex $i 0] == $::ced::c_set(operator_value)} {
					set ::ced::c_set(operator_text) [lindex $i 1]
					set found true
					break
				}
			}
			if {! $found} {
				set ::ced::c_set(operator_value) 1
				set ::ced::c_set(operator_text) {MOV    =}
			}

			cmd_SetCLnk
		}
		10 -
		11 -
		12 -
		13 {
			# CSetCLnk and CSetLnk
			if {$cmdtype <= 11} {
				set cmdname "CSetCLnk"
			} else {
				set cmdname "CSetLnk"
			}
			if {$cmdtype == 10 || $cmdtype == 12} {
				# indirect mode
				set ::ced::c_set(sourcemode) 2
			} else {
				# direct mode
				set ::ced::c_set(sourcemode) 1
			}

			switch [h2gui:link 3] {
				1 {
					focus $::ced::widget(hex_entry).d_1_2
					cmd_illegal_command "$cmdname sub-link" \
						"byte 1 lower nibble must be 1, 4, 5, 6 or 7 (or 0 to disable the link)"
					return
				}
				2 {
					focus $::ced::widget(hex_entry).d_7_1
					cmd_illegal_command "$cmdname sub-link" \
						"when byte 1 lower nibble is 1,\nbyte 7 must be 00, 01, 02, 03, 05, 06, 07, 09, 0A, 0B, 0C, 0D or 0F"
					return
				}
				3 {
					focus $::ced::widget(hex_entry).d_7_2
					cmd_illegal_command "sub-link" \
						"With $cmdname, you can't use this link command:\nbyte 1 lower nibble must be 1  (or 0 to disable the link)"
					return
				}
			}
			set iftype [peek_hex_nibble 1 0]
			set ::ced::c_if(check) true

			if {$::ced::c_set(sourcemode) == 1} {
				set ::ced::c_set(value) [peek_hex_word 2]

				# direct set => indirect IF only
				if {[expr {$iftype & 8}] == 0} {
					set ::ced::c_if(operatornum) $iftype
					set ::ced::c_if(operator) [h2gui:if_operator_text $iftype]

					set op [peek_hex_byte 4]
					set ::ced::c_if(gprm1) [expr {$op & 0xF}]
					set op [peek_hex_byte 5]
					if {[expr {$op & 0X80}] == 0} {
						set ::ced::c_if(radioOp2) 2
						set ::ced::c_if(gprm2) $op
					} else {
						set ::ced::c_if(radioOp2) 3
						set ::ced::c_if(sprm2) [expr {$op & 0X7F}]
						if {$::ced::c_if(sprm2) > 23} {set ::ced::c_if(sprm2) 23}
					}
				} else {
					# illegal direct IF
					set ::ced::c_if(operatornum) [expr {$iftype & 7}]
				}
			} else {
				set r [peek_hex_byte 2]
				if {[expr {$r & 0X80}] == 0} {
					set ::ced::c_set(sourcemode) 2
					set ::ced::c_set(gprm) $r
				} else {
					set ::ced::c_set(sourcemode) 3
					set ::ced::c_set(sprm) [expr {$r & 0X7F}]
					if {$::ced::c_set(sprm) > 23} {set ::ced::c_set(sprm) 23}
				}

				# indirect set => direct or indirect if
				# direct set => indirect IF only
				set ::ced::c_if(operatornum) [expr {$iftype & 7}]
				set ::ced::c_if(operator) [h2gui:if_operator_text $iftype]
				set op [peek_hex_byte 3]
				set ::ced::c_if(gprm1) [expr {$op & 0xF}]
				if {[expr {$iftype & 8}] == 0} {
					set op [peek_hex_byte 5]
					if {[expr {$op & 0X80}] == 0} {
						set ::ced::c_if(radioOp2) 2
						set ::ced::c_if(gprm2) $op
					} else {
						set ::ced::c_if(radioOp2) 3
						set ::ced::c_if(sprm2) [expr {$op & 0X7F}]
						if {$::ced::c_if(sprm2) > 23} {set ::ced::c_if(sprm2) 23}
					}
				} else {
					set ::ced::c_if(radioOp2) 1
					set ::ced::c_if(value2) [peek_hex_word 4]
				}
			}
			set ::ced::c_set(destination) [peek_hex_nibble 1 1]
			set ::ced::c_set(operator_value) [peek_hex_nibble 0 1]
			set found false
			foreach i $::ced::c_Setoperators {
				if {[lindex $i 0] == $::ced::c_set(operator_value)} {
					set ::ced::c_set(operator_text) [lindex $i 1]
					set found true
					break
				}
			}
			if {! $found} {
				set ::ced::c_set(operator_value) 1
				set ::ced::c_set(operator_text) {MOV    =}
			}

			cmd_$cmdname
		}
		default {
			# illegal commands
			focus $::ced::widget(hex_entry).d_0_1
			cmd_illegal_command main "byte 0 upper nibble must be 0 or 2 to D"
			return
		}
	}

	# highlights changes in the hex entry after cmd validation
	c_hex_entry_highlight_clear
	for {set b 0} {$b < 8} {incr b} {
		set pos [expr {$b * 3}]
		if {[string range $::hexcommand $pos $pos] \
				!= [string range $::ced::old_hexcommand $pos $pos]} {
			c_hex_entry_highlight_set $b 1 pink
		}
		incr pos
		if {[string range $::hexcommand $pos $pos] \
				!= [string range $::ced::old_hexcommand $pos $pos]} {
			c_hex_entry_highlight_set $b 2 pink
		}
	}
}

proc h2gui:if_operator_text {if_operator_num} {
	#dprint 3 "> h2guiif_operator_text {if_operator_num}"
	switch [expr {$if_operator_num & 7}] {
		0 {bell; return "????"}
		1 {return "&"}
		2 {return "=="}
		3 {return "!="}
		4 {return ">="}
		5 {return ">"}
		6 {return "<="}
		7 {return "<"}
	}
}

proc h2gui:if1 {} {
	#dprint 3 "> hexcmd_to_gui:if1 {}"
	set iftype [peek_hex_nibble 1 0]
	if {[expr {$iftype & 7}] == 0} {
		set ::ced::c_if(check) false
	} else {
		set ::ced::c_if(check) true
		set ::ced::c_if(operatornum) [expr {$iftype & 7}]
		set ::ced::c_if(operator) [h2gui:if_operator_text $iftype]
		set op [peek_hex_byte 3]
		set ::ced::c_if(gprm1) [expr {$op & 0xF}]
		if {[expr {$iftype & 8}] == 0} {
			set op [peek_hex_byte 5]
			if {[expr {$op & 0X80}] == 0} {
				set ::ced::c_if(radioOp2) 2
				set ::ced::c_if(gprm2) $op
			} else {
				set ::ced::c_if(radioOp2) 3
				set ::ced::c_if(sprm2) [expr {$op & 0X7F}]
				if {$::ced::c_if(sprm2) > 23} {set ::ced::c_if(sprm2) 23}
			}
		} else {
			set ::ced::c_if(radioOp2) 1
			set ::ced::c_if(value2) [peek_hex_word 4]
		}
	}
}

proc h2gui:if2 {} {
	#dprint 3 "> hexcmd_to_gui:if2 {}"
	set iftype [peek_hex_nibble 1 0]
	if {[expr {$iftype & 7}] == 0} {
		set ::ced::c_if(check) false
	} else {
		set ::ced::c_if(check) true
		set ::ced::c_if(operatornum) [expr {$iftype & 7}]
		set ::ced::c_if(operator) [h2gui:if_operator_text $iftype]
		set op [peek_hex_byte 6]
		set ::ced::c_if(gprm1) [expr {$op & 0xF}]
		set ::ced::c_if(value2) 0
		set op [peek_hex_byte 7]
		if {[expr {$op & 0X80}] == 0} {
			set ::ced::c_if(radioOp2) 2
			set ::ced::c_if(gprm2) $op
		} else {
			set ::ced::c_if(radioOp2) 3
			set ::ced::c_if(sprm2) [expr {$op & 0X7F}]
			if {$::ced::c_if(sprm2) > 23} {set ::ced::c_if(sprm2) 23}
		}
	}
}

proc h2gui:link {type} {
	#dprint 3 "> hexcmd_to_gui:link {type}"
	set ::ced::c_link(check) false
	if {$type == 3} {
		set k 1
	} else {
		set k  [peek_hex_nibble 1 1]
	}
	if {$k > 0} {
		if {$type > 1 && $k > 1} {
			return 3
		}
		set highlightbutton [expr {([peek_hex_byte 6] & 0XFC) / 4}]
		if {$highlightbutton > 0} {
			set highlightmode true
		} else {
			set highlightmode false
		}
		switch $k {
			1 {
				set ll [peek_hex_byte 7]
				if {$ll==14 || $ll==15 || $ll>16} {
					return 2
				}
				set ::ced::c_link(highlightmode) $highlightmode
				set ::ced::c_link(ll) $ll
				set ::ced::c_link(highlightbutton) $highlightbutton
			}
			4 {
				set ::ced::c_link(pgcn) [peek_hex_word 6]
			}
			5 {
				set ::ced::c_link(pttn) [expr {[peek_hex_word 6] & 0X03FF}]
				set ::ced::c_link(highlightbutton) $highlightbutton
				set ::ced::c_link(highlightmode)   $highlightmode
			}
			6 {
				set ::ced::c_link(pgn) [peek_hex_byte 7]
				set ::ced::c_link(highlightbutton) $highlightbutton
				set ::ced::c_link(highlightmode)   $highlightmode
			}
			7 {
				set ::ced::c_link(cell) [peek_hex_byte 7]
				set ::ced::c_link(highlightbutton) $highlightbutton
				set ::ced::c_link(highlightmode)   $highlightmode
			}
			default {
				return 1
			}
		}
		set ::ced::c_link(check) true
	}
	set ::ced::c_link(k) $k
	return 0
}

#############################################################################
## copies the contents of the entry compound widget holding the DVD command
## in the global variable ::hexcommand, in ASCII with spaces form

proc c_hex_entry_poke {} {
	#dprint 3 "> c_hex_entry_poke {}"
	set str ""
	for {set b 0} {$b < 8} {incr b} {
		set str "$str[$::ced::widget(hex_entry).d_${b}_1 get][$::ced::widget(hex_entry).d_${b}_2 get] "
	}
	set ::hexcommand [string trim $str]
}

#############################################################################
## copies the contents of the global variable ::hexcommand
## to the entry compound widget holding the DVD command.

proc c_hex_entry_refresh {} {
	#dprint 4 "> c_hex_entry_refresh {}"
	set oldfocus [focus]
	set byteslist [split $::hexcommand]
	for {set i 0} {$i < 8} {incr i} {
		set b [lindex $byteslist $i]
		$::ced::widget(hex_entry).d_${i}_1 configure -validate none
		$::ced::widget(hex_entry).d_${i}_1 delete 0 10
		$::ced::widget(hex_entry).d_${i}_1 insert 0 [string range $b 0 0]
		$::ced::widget(hex_entry).d_${i}_1 configure -validate key
		$::ced::widget(hex_entry).d_${i}_2 configure -validate none
		$::ced::widget(hex_entry).d_${i}_2 delete 0 10
		$::ced::widget(hex_entry).d_${i}_2 insert 0 [string range $b 1 1]
		$::ced::widget(hex_entry).d_${i}_2 configure -validate key
	}
	catch {focus $oldfocus}
}

#############################################################################
## compound widget that allow the editing of the hex command.
## The callback routine is called whenever the user press Return in the hex string
## and must have one argument: the widget path

proc c_hex_entry {w callback} {
	#dprint 3 "> c_hex_entry {w callback}"

	frame  $w -borderwidth 2 -relief sunken -background white
	for {set i 0} {$i < 8} {incr i} {
		entry $w.d_${i}_1 -width 1 -borderwidth 0 -selectbackground white -insertwidth 20 \
				-insertbackground gray -takefocus 0 -font $::config(fixed_width_font)
		pack  $w.d_${i}_1 -side left
		entry $w.d_${i}_2 -width 1 -borderwidth 0 -selectbackground white -insertwidth 20\
				-insertbackground gray -takefocus 0 -font $::config(fixed_width_font)
		pack  $w.d_${i}_2 -side left
		bind  $w.d_${i}_1 <Button-1> "::ced::focus_hex_digit $w.d_${i}_1"
		bind  $w.d_${i}_2 <Button-1> "::ced::focus_hex_digit $w.d_${i}_2"
		bind  $w.d_${i}_1 <KeyPress-Right> "::ced::focus_hex_digit $w.d_${i}_2"
		bind  $w.d_${i}_2 <KeyPress-Left>  "::ced::focus_hex_digit $w.d_${i}_1"
		bind  $w.d_${i}_2 <KeyPress-BackSpace> "::ced::focus_hex_digit $w.d_${i}_1"
		bind  $w.d_${i}_1 <KeyPress-Home>  "::ced::focus_hex_digit $w.d_0_1"
		bind  $w.d_${i}_2 <KeyPress-Home>  "::ced::focus_hex_digit $w.d_0_1"
		bind  $w.d_${i}_1 <KeyPress-End>   "::ced::focus_hex_digit $w.d_7_1"
		bind  $w.d_${i}_2 <KeyPress-End>   "::ced::focus_hex_digit $w.d_7_1"
		bind  $w.d_${i}_1 <Control-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) false]}
		bind  $w.d_${i}_2 <Control-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) false]}
		bind  $w.d_${i}_1 <Control-Shift-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) true]}
		bind  $w.d_${i}_2 <Control-Shift-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) true]}
		bind  $w.d_${i}_1 <Alt-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_binary $::ced::widget(hex_entry)]}
		bind  $w.d_${i}_2 <Alt-c> {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_binary $::ced::widget(hex_entry)]}
		bind  $w.d_${i}_1 <Control-v> "::ced::c_hex_entry_paste $w"
		bind  $w.d_${i}_2 <Control-v> "::ced::c_hex_entry_paste $w"
		bind  $w.d_${i}_1 <Control-z> "::ced::c_hex_entry_undo_validation $w"
		bind  $w.d_${i}_2 <Control-z> "::ced::c_hex_entry_undo_validation $w"
		bind  $w.d_${i}_1 <Control-d> "::ced::hexentry_disassemble"
		bind  $w.d_${i}_2 <Control-d> "::ced::hexentry_disassemble"
		if {$callback != ""} {
			bind  $w.d_${i}_1 <KeyPress-Return> "$callback $w"
			bind  $w.d_${i}_2 <KeyPress-Return> "$callback $w"
		}
		if {$i < 7} {
			bind  $w.d_${i}_2 <KeyPress-Right> "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"
			bind  $w.d_${i}_1 <KeyPress-Next>  "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"
			bind  $w.d_${i}_2 <KeyPress-Next>  "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"

			bind  $w.d_${i}_1 <KeyPress-Down>  "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"
			bind  $w.d_${i}_2 <KeyPress-Down>  "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"

			frame $w.pad_$i -width 6 -background white -cursor xterm
			pack  $w.pad_$i -side left -fill y
			bind  $w.pad_$i <Button-1> "::ced::focus_hex_digit $w.d_[expr {$i + 1}]_1"
		}
		if {$i > 0} {
			bind  $w.d_${i}_1 <KeyPress-Left>  "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_2"
			bind  $w.d_${i}_1 <KeyPress-BackSpace> "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_2"
			bind  $w.d_${i}_1 <KeyPress-Prior> "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_1"
			bind  $w.d_${i}_2 <KeyPress-Prior> "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_1"

			bind  $w.d_${i}_1 <KeyPress-Up> "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_1"
			bind  $w.d_${i}_2 <KeyPress-Up> "::ced::focus_hex_digit $w.d_[expr {$i - 1}]_1"
		}
	}
	$w.d_0_1 configure -takefocus 1
	bind  $w.d_0_1 <KeyRelease-Left>  "::ced::focus_hex_digit $w.d_0_1"
	bind  $w.d_7_2 <KeyRelease-Right> "::ced::focus_hex_digit $w.d_7_2"
	bind  $w.d_0_2 <KeyPress-Up>      "::ced::focus_hex_digit $w.d_0_1"

	for {set i 0} {$i < 8} {incr i} {
		$w.d_${i}_1 insert 0 "0"
		$w.d_${i}_2 insert 0 "0"
		$w.d_${i}_1 configure -validate key \
				-validatecommand "::ced::validate_hexcommand $w.d $i 1 %d %S"
		$w.d_${i}_2 configure -validate key \
				-validatecommand "::ced::validate_hexcommand $w.d $i 2 %d %S"
	}
	return $w
}
proc validate_hexcommand {path numbyte numdigit action text} {
	#dprint 4 "> c_hex_entry:validate_hexcommand {path numbyte numdigit action text}"
	if {[string length $text] != 1} {return false}
	if {![string is xdigit $text]}  {return false}
	${path}_${numbyte}_$numdigit configure -validate none
	${path}_${numbyte}_$numdigit delete 0 1000
	${path}_${numbyte}_$numdigit insert 0 [string toupper $text]
	${path}_${numbyte}_$numdigit configure -validate key
	if {$action == 1} {
		if {$numdigit == 1} {
			focus_hex_digit ${path}_${numbyte}_2
		} else {
			if {$numbyte < 7} {
				focus_hex_digit ${path}_[expr {$numbyte + 1}]_1
			} else {
				focus_hex_digit ${path}_7_2
			}
		}
	}
	return false
}
proc focus_hex_digit {w} {
	#dprint 4 "> c_hex_entry:focus_hex_digit {w}"
	focus $w
	$w icursor 0
	$w selection from 0
	$w selection to 1000
}

#############################################################################
## return the string stored in the hex entry in ASCII form, with or without
## spaces (16 or 23 characters)
## strip_spaces: boolean: removes spaces from the hex string

proc c_hex_entry_get_ascii {w strip_spaces} {
	#dprint 3 "> c_hex_entry_get_ascii {w strip_spaces}"
	if {$strip_spaces} {
		set pad ""
	} else {
		set pad " "
	}
	set str ""
	for {set i 0} {$i < 8} {incr i} {
		set str "$str[$w.d_${i}_1 get][$w.d_${i}_2 get]$pad"
	}
	return [string trim $str]
}

#############################################################################
## return the string stored in the hex entry in binary form (8 characters)

proc c_hex_entry_get_binary {w} {
	#dprint 3 "> c_hex_entry_get_binary {w}"
	set str ""
	for {set i 0} {$i < 8} {incr i} {
		set str [format "%s%c" $str "0X[$w.d_${i}_1 get][$w.d_${i}_2 get]"]
	}
	return $str
}

#############################################################################
## paste the clipboard in the hex entry.
## the clipboard must contain a valid string: 8 binary chars or an hexadecimal
## string (with or without embedded spaces) with 16 hex digits, or an hex list
## element like the ones used in the main GUI.

proc c_hex_entry_paste {w} {
	#dprint 3 "> c_hex_entry_paste {w}"
	if {[catch {set str [clipboard get]}]} {
		bell
		return ""
	}
	set def [extract_hexcommand $str]
	if {$def == ""} {
		bell
	} else {
		for {set i 0} {$i < 8} {incr i} {
			${w}.d_${i}_1 configure -validate none
			${w}.d_${i}_1 delete 0 1000
			${w}.d_${i}_1 insert 0 [string toupper [string range [lindex $def $i] 0 0]]
			${w}.d_${i}_1 configure -validate key
			${w}.d_${i}_2 configure -validate none
			${w}.d_${i}_2 delete 0 1000
			${w}.d_${i}_2 insert 0 [string toupper [string range [lindex $def $i] 1 1]]
			${w}.d_${i}_2 configure -validate key
		}
	}
	::ced::c_hex_entry_poke
	::ced::hexcmd_to_gui

	return $def
}


#############################################################################
## Extract the hexadecimal command from a command string which may be a binary
## string, a formatted string, or a standard hex command surrounded or not by
## accolades or square brackets
proc extract_hexcommand {str} {
	set def ""
	# if string is 8 bytes long, assumes it's the input string in binary format
	if {[string length $str]==8} {
		set def $str
	} else {
		set str [string toupper [string trim $str]]
		if {[string range $str 0 0] == "\[" && [string length $str] >= 25} {
			set str [string range $str 1 23]
		}
		if {[string range $str 0 0] == "\{" && [string length $str] >= 25} {
			set str [string range $str 1 23]
		}
		# if string is not 16 bytes long or is not in hexadecimal,
		# assumes there are white spaces in the string, and strip them
		if {[string length $str]!=16 || ![string is xdigit $str]} {
			set tmp ""
			foreach c [split $str] {
				set tmp "$tmp$c"
			}
			set str $tmp
		}
		# if string is 16 bytes long and is in hexadecimal, convert it to binary
		if {[string length $str]==16 && [string is xdigit $str]} {
			set def [binary format H* $str]
		}
	}
	if {$def == ""} {
		return ""
	} else {
		binary scan $def "H2H2H2H2H2H2H2H2" v(0) v(1) v(2) v(3) v(4) v(5) v(6) v(7)
		set hex [string toupper "$v(0) $v(1) $v(2) $v(3) $v(4) $v(5) $v(6) $v(7)"]
		return $hex
	}
}

#############################################################################
## Revert the content of the hex widget to his content before the last validation
proc c_hex_entry_undo_validation {w} {
	#dprint 3 "> c_hex_entry_undo_validation {w}"

	if {$::ced::old_hexcommand == ""} {
		bell
	} else {
		set i 0
		foreach c [split $::ced::old_hexcommand] {
			${w}.d_${i}_1 configure -validate none
			${w}.d_${i}_1 delete 0 1000
			${w}.d_${i}_1 insert 0 [string range $c 0 0]
			${w}.d_${i}_1 configure -validate key
			${w}.d_${i}_2 configure -validate none
			${w}.d_${i}_2 delete 0 1000
			${w}.d_${i}_2 insert 0 [string range $c 1 1]
			${w}.d_${i}_2 configure -validate key
			incr i
		}
	}
}

#############################################################################
## highlight an hexadecimal character in the hex entry widget

proc c_hex_entry_highlight_set {byte nibble color} {
	#dprint 3 "> c_hex_entry_highlight_nibble {byte nibble color}"
	$::ced::widget(hex_entry).d_${byte}_$nibble configure -background $color
}

#############################################################################
## clear all the highlighted characters in the hex entry widget

proc c_hex_entry_highlight_clear {} {
	#dprint 3 "> c_hex_entry_highlight_clear {}"
	for {set b 0} {$b < 8} {incr b} {
		$::ced::widget(hex_entry).d_${b}_1 configure -background white
		$::ced::widget(hex_entry).d_${b}_2 configure -background white
	}
}

#############################################################################
## Return the value of the most or less significant nibble of a byte of the
## current hex command in decimal form.
## nibble: 0 for most significant nibble, 1 for less

proc peek_hex_nibble {bytenum nibble} {
	#dprint 4 "> peek_hex_nibble {bytenum nibble}"
	return [x2d [string range [lindex $::hexcommand $bytenum] $nibble $nibble]]
}

#############################################################################
## Return the value of a byte of the current hex command in decimal form.

proc peek_hex_byte {bytenum} {
	#dprint 4 "> peek_hex_byte {bytenum}"
	return [x2d [lindex $::hexcommand $bytenum]]
}

#############################################################################
## Return the value of a word (2 bytes) of the current hex command in decimal form.
## The argument 'bytenum' is the index of the first byte of the word (0-6)

proc peek_hex_word {bytenum} {
	#dprint 4 "> peek_hex_word {bytenum}"
	set byte2num [expr {$bytenum + 1}]
	return [x2d [lindex $::hexcommand $bytenum][lindex $::hexcommand $byte2num]]
}

#############################################################################
## Replace the most or less significant nibble of a byte of the current hex command,
## and refresh the hex entry.
## nibble: 0 for most significant nibble, 1 for less

proc poke_hex_nibble {bytenum nibble value} {
	#dprint 4 "> poke_hex_nibble {bytenum nibble value}"
	if {$value==""||$value=="0x"||$value=="0X"} {set value 0}
	if {$value < 0} {set value 0}
	if {$value > 15} {set value 15}

	set newnybble [format "%01X" $value]
	set origbyte [lindex $::hexcommand $bytenum]
	if {$nibble == 0} {
		set byte $newnybble[string range $origbyte 1 1]
	} else {
		set byte [string range $origbyte 0 0]$newnybble
	}
	lset ::hexcommand $bytenum $byte

	c_hex_entry_refresh
}

#############################################################################
## Replace a byte of the current hex command, and refresh the hex entry.

proc poke_hex_byte {bytenum value} {
	#dprint 4 "> poke_hex_byte {bytenum value}"
	if {[string range $value 0 0] == "'"} {
		set asc [string range $value 1 end]
		if {[string length $asc] > 1 || $asc == ""} {
			set value 0
		} else {
			set value [b2i $asc]
		}
	}
	if {$value==""||$value=="0x"||$value=="0X"} {set value 0}
	if {$value < 0} {set value 0}
	if {$value > 255} {set value 255}
	lset ::hexcommand $bytenum [format "%02X" $value]
	c_hex_entry_refresh
}

#############################################################################
## Replace a word (2 bytes) of the current hex command, and refresh the hex entry.
## The argument 'bytenum' is the index of the first byte of the word (0-6)

proc poke_hex_word {bytenum value} {
	#dprint 4 "> poke_hex_word {bytenum value}"
	if {[string range $value 0 0] == "'"} {
		set asc [string range $value 1 end]
		if {[string length $asc] > 2 || $asc == ""} {
			set value 0
		} else {
			set value [b2i $asc]
		}
	}
	if {$value==""||$value=="0x"||$value=="0X"} {set value 0}
	if {$value > 65535} {set value 65535}
	if {$value < 0} {set value 0}

	set byte [format "%02X" [expr {$value / 256}]]
	lset ::hexcommand $bytenum $byte
	incr bytenum
	set byte [format "%02X" [expr {$value % 256}]]
	lset ::hexcommand $bytenum $byte

	c_hex_entry_refresh
}

#############################################################################
## Replace the last word (2 bytes) of the current hex command by the value
## of the PTTN $::ced::c_link(pttn) (10 less significant bits) and the value of the
## Highlighted Button $::ced::c_link(highlightbutton) (6 most significant bits).
## Do NOT refresh the hex entry widget.

proc poke_LinkPTT {args} {
	#dprint 3 "> poke_LinkPTT {args}"
	set val 0
	if {[string range $::ced::c_link(pttn) 0 0] == "'"} {
		set asc [string range $::ced::c_link(pttn) 1 end]
		if {[string length $asc] > 2 || $asc == ""} {
			set val 0
		} else {
			set val [b2i $asc]
		}
	}
	if {$val == 0} {
		if {[string is integer $::ced::c_link(pttn)] && $::ced::c_link(pttn) != ""} {
			set val $::ced::c_link(pttn)
		}
	}
	if {$::ced::c_link(highlightmode)} {
		set byte [expr {($::ced::c_link(highlightbutton) * 4) + ($val / 256)}]
	} else {
		set byte [expr {$val / 256}]
	}
	poke_hex_byte 6 $byte
	poke_hex_byte 7 [expr {$val % 256}]
}

#############################################################################
## procedure called by the entry widgets to validate the number typed.
## Do NOT refresh the hex entry widget.
## If $args is 1, then allow to enter ascii values in the form 'xx
proc value_entry_validate {min max action val args} {
	if {$action != 0} {
		if {$args == 1 && [string range $val 0 0] == "'"} {
			set asc [string range $val 1 end]
			if {[string length $asc] > 2} {bell; return false}
			if {$asc == ""} {
				set val 0
			} else {
				set val [b2i $asc]
			}
		}
		if {$min < 0 && ($val=="-"||$val=="")} {
			return true
		}
		if {$val=="0"||$val=="0x"||$val=="0X"} {
			if {$min>0 && $val=="0"} bell
			return true
		}
		if {![string is integer $val]} {bell; return false}
		if {$val < $min} {bell; return false}
		if {$val > $max} {bell; return false}
	}
	return true
}


#############################################################################
## validate the resume cell value, and modify the hexcommand accordingly.
## Do NOT refresh the hex entry widget.

proc c_resumecell_validate_and_poke {w} {
	#dprint 3 "> c_resumecell_validate_and_poke {in}"
	if {$::ced::c_call(resume_mode)} {
		poke_hex_byte 4 $::ced::c_call(resume_cell)
		$w.f.f1.f2.ent configure -state normal
	} else {
		poke_hex_byte 4 0
		$w.f.f1.f2.ent configure -state disabled
	}
}

#############################################################################
## create a resume cell compound widget

proc c_resumecell_create {w} {
	#dprint 3 "> c_resumecell_create {w}"
	destroy $w

	set ::ced::c_call(resume_cell) [force_range $::ced::c_call(resume_cell) 1 255]

	# main resume cell frame
	frame $w

	frame $w.f
	pack $w.f -side top -fill x

	label $w.f.lab -text "If a resume is issued (with RSM or remote), then"
	pack $w.f.lab -side left

	frame $w.f.f1 -borderwidth 2 -relief groove
	pack $w.f.f1 -side left -fill x
	radiobutton $w.f.f1.but1 -bd $::config(min_bd) -height 2 \
			-text "return to the point in playback\nwhere the CallSS was issued." \
			-value false -variable ::ced::c_call(resume_mode) \
			-command "::ced::c_resumecell_validate_and_poke $w"
	pack $w.f.f1.but1 -side top -fill x -anchor w

	frame $w.f.f1.f2
	pack $w.f.f1.f2 -side top -fill x -anchor w
	radiobutton $w.f.f1.f2.but2 -bd $::config(min_bd) \
			-text "return to cell" \
			-value true -variable ::ced::c_call(resume_mode) \
			-command "::ced::c_resumecell_validate_and_poke $w"
	pack $w.f.f1.f2.but2 -side left
	entry $w.f.f1.f2.ent -textvariable ::ced::c_call(resume_cell) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 255 {%d} {%P} 1"
	pack $w.f.f1.f2.ent -side left -padx 2
	bind $w.f.f1.f2.ent <KeyRelease> "::ced::c_resumecell_validate_and_poke $w"
	label $w.f.f1.f2.lab -text "(1-255)."
	pack  $w.f.f1.f2.lab -side left

	c_resumecell_validate_and_poke $w

	return $w
}

#############################################################################
## create a compound widget with the value, GPRM and SPRM options.
## Arguments:
##	widget - window name
##	callbackcmd - command to be called when the interface is modified.  Must have 1 dummy arg
##	modevar - name of the variable holding the mode of operation (1=value, 2=GPRM, 3=SPRM)
##	gprmvar - Name of the variable holding the GPRM value
##	sprmvar - name of the variable holding the SPRM value
##	valuevar - name of the variable holding the numeric value
##	valuemin - minimum value allowed for the numeric value
##	valuemax - maximum value allowed for the numeric value
## Return:
##	The name of the created compound widget
## Notes:
##	All variable must exists and must be properly initialized before calling this proc.
##	If the gprmvar, sprmvar or valuevar is an empty string, the corresponding option is not displayed.
## Example usage:
##	proc callback {args} {puts "mode=$mode gprm=$gprm sprm=$sprm value=$value"}
##	set mode 1; set gprm 0; set sprm 0; set value 0
##	pack [c_regval_create $::ced::widget(cmdFrame) callback ::ced::mode ::ced::gprm ::ced::sprm ::ced::value 0 0XFFFF]

proc c_regval_create {widget callbackcmd modevar gprmvar sprmvar valuevar valuemin valuemax} {
	#dprint 3 "> c_regval_create {widget callbackcmd modevar gprmvar sprmvar valuevar valuemin valuemax}"

	frame $widget -relief groove -borderwidth 2
	if {$valuevar != ""} {
		set $valuevar [force_range [subst $$valuevar] $valuemin $valuemax]
		frame $widget.val
		pack  $widget.val  -side top -fill x
		radiobutton $widget.val.r  -text "value" -value 1 -width 6 -anchor w -variable $modevar \
				-command "::ced::c_regval_refresh $widget \"$callbackcmd\""
		pack  $widget.val.r  -side left
		if {$valuemax >= 63} {
			entry $widget.val.e  -textvariable $valuevar -validate all \
					-width [expr {[string length [expr {$valuemax}]]+2}] \
					-validatecommand "::ced::value_entry_validate $valuemin $valuemax {%d} {%P} 1"
			pack  $widget.val.e  -side left -padx 1
			bind  $widget.val.e  <KeyRelease> "$callbackcmd dummy"
			label $widget.val.l  -text "([expr {$valuemin}]-[expr {$valuemax}])"
			pack  $widget.val.l  -side left -padx 1
			if {$valuemax >= 36864} {
				button $widget.val.bb -text "B" -width 1 -pady 0 -bd 1 \
						-command "::ced::enter_button_num $valuevar $callbackcmd $widget.val.bb"
				pack   $widget.val.bb -side left -padx 2 -pady 1
			}
		} else {
			scale $widget.val.s -from $valuemin -to $valuemax -variable "$valuevar" -command "$callbackcmd"
			pack  $widget.val.s -side left
		}
	}
	if {$gprmvar != ""} {
		frame $widget.gprm
		pack  $widget.gprm -side top -fill x -pady 2 -expand 1
		radiobutton $widget.gprm.r -text "GPRM" -value 2 -width 6 -anchor w -variable $modevar \
				-command "::ced::c_regval_refresh $widget \"$callbackcmd\""
		pack  $widget.gprm.r -side left
		scale $widget.gprm.s -from 0 -to 15 -variable "$gprmvar" -command "$callbackcmd"
		pack  $widget.gprm.s -side left -fill x -expand 1
	}
	if {$sprmvar != ""} {
		frame $widget.sprm
		pack  $widget.sprm -side top -fill x -expand 1
		radiobutton $widget.sprm.r -text "SPRM" -value 3 -width 6 -anchor w -variable $modevar \
				-command "::ced::c_regval_refresh $widget \"$callbackcmd\""
		pack  $widget.sprm.r -side left
		menubutton $widget.sprm.m -textvariable $sprmvar -pady 1 -width 10 \
				-menu $widget.sprm.m.mnu
		pack  $widget.sprm.m -side left -fill x -expand 1
		menu  $widget.sprm.m.mnu -tearoff 0
		foreach i $::ced::c_SPRMlist {
			$widget.sprm.m.mnu add command -label "$i" \
					-command "set $sprmvar [lindex $i 0]; $callbackcmd dummy"
		}
	}
	c_regval_refresh $widget $callbackcmd
	return $widget
}
proc enter_button_num {valuevar callbackcmd srcentry} {
	set val [subst \$$valuevar]
	if {$val == ""} {set $valuevar 0; set val 0}
	if {![string is integer $val]} {set $valuevar 0; set val 0}
	set ::ced::butnum [expr {$val / 1024}]

	set w .butnum
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Button number"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::ced::butnum_done 0}
	bind $w <Escape> "set ::ced::butnum_done 0"

	scale $w.s -from 1 -to 36 -variable ::ced::butnum
	pack  $w.s -fill x -padx 4

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::ced::butnum_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::ced::butnum_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr {[winfo rootx $srcentry] + [winfo width $srcentry] - [winfo reqwidth $w] -8}]
	set wy [expr {[winfo rooty $srcentry] + [winfo height $srcentry]}]
	wm geometry $w +$wx+$wy
	wm deiconify $w
	set oldgrab [grab current .]
	grab set $w

	tkwait variable ::ced::butnum_done
	grab release $w
	destroy $w
	if {$oldgrab != ""} {
		grab set $oldgrab
		wm deiconify $oldgrab
	}
	if {$::ced::butnum_done} {
		set $valuevar [expr {$::ced::butnum * 1024}]
		$callbackcmd dummy
		return $::ced::butnum
	}
	return 0
}
#############################################################################
## refresh a regval widget, and optionally call the callback command
## The callback proc must exists and have 1 dummy arg, or the callbackcommand
## parameter must be an empty string

proc c_regval_refresh {w callbackcommand} {
	#dprint 3 "> c_regval_refresh {w callbackcommand}"
	if {[catch {set modevar [$w.gprm.r cget -variable]}]} {
		if {[catch {set modevar [$w.val.r cget -variable]}]} {
			set modevar [$w.sprm.r cget -variable]
		}
	}
	set mode [subst $$modevar]
	switch $mode {
		1 {
			catch {$w.val.e  configure -state normal}
			catch {$w.val.l  configure -foreground black}
			catch {$w.val.bb configure -state normal}
			catch {scale_state $w.val.s normal}
			catch {scale_state $w.gprm.s disabled}
			catch {$w.sprm.m configure -state disabled}
		}
		2 {
			catch {$w.val.e  configure -state disabled}
			catch {$w.val.l  configure -foreground gray}
			catch {$w.val.bb configure -state disabled}
			catch {scale_state $w.val.s disabled}
			catch {scale_state $w.gprm.s normal}
			catch {$w.sprm.m configure -state disabled}
		}
		3 {
			catch {$w.val.e  configure -state disabled}
			catch {$w.val.l  configure -foreground gray}
			catch {$w.val.bb configure -state disabled}
			catch {scale_state $w.val.s disabled}
			catch {scale_state $w.gprm.s disabled}
			catch {$w.sprm.m configure -state normal}
		}
	}
	if {"$callbackcommand" != ""} {
		$callbackcommand dummy
	}
}

#############################################################################
## change the state of all the widgets in a compound regval widget.
## state must be normal or disabled

proc c_regval_refresh_state {w state} {
	#dprint 3 "> c_regval_refresh_state {w state}"
	switch $state {
		normal {
			catch {$w.val.r  configure -state normal}
			catch {$w.gprm.r configure -state normal}
			catch {$w.sprm.r configure -state normal}
			c_regval_refresh $w {}
		}
		disabled {
			catch {$w.val.r  configure -state disabled}
			catch {$w.val.e  configure -state disabled}
			catch {$w.val.bb configure -state disabled}
			catch {$w.val.l  configure -foreground gray}
			catch {$w.gprm.r configure -state disabled}
			catch {scale_state $w.gprm.s disabled}
			catch {$w.sprm.r configure -state disabled}
			catch {$w.sprm.m configure -state disabled}
		}
	}
}

#############################################################################
## refresh a compound highlight_button widget

proc c_highlight_button_refresh {w} {
	#dprint 3 "> c_highlight_button_refresh {w}"

	# there is no Link's Highlight Button for SetHL_BTN
	set cmdtype [peek_hex_nibble 0 0]
	if {($cmdtype == 4 || $cmdtype == 5) && [peek_hex_nibble 0 1]==6} return

	if {$::ced::c_link(highlightmode)} {
		scale_state $w.f1.scale normal
	} else {
		scale_state $w.f1.scale disabled
	}
}

#############################################################################
## poke the value of a compound highlight_button widget in the hexcommand,
## and refresh the hex entry widget.

proc c_highlight_button_poke {args} {
	#dprint 3 "> c_highlight_button_poke {args}"
	set oldvalue [expr {[x2d [lindex $::hexcommand 6]] & 3}]
	if {$::ced::c_link(highlightmode)} {
		poke_hex_byte 6 [expr {($::ced::c_link(highlightbutton) * 4) + $oldvalue}]
	} else {
		poke_hex_byte 6 $oldvalue
	}
	c_hex_entry_refresh
}

#############################################################################
## create a compound highlight_button widget

proc c_highlight_button_create {w} {
	#dprint 3 "> c_highlight_button_create {w}"

	frame $w

	# do not display the Link's Highlight Button for SetHL_BTN
	set cmdtype [peek_hex_nibble 0 0]
	if {($cmdtype == 4 || $cmdtype == 5) && [peek_hex_nibble 0 1]==6} return

	label $w.lab -text "and "
	pack  $w.lab -side left

	frame $w.f1
	pack  $w.f1 -side top -fill x
	checkbutton $w.f1.r -variable ::ced::c_link(highlightmode) -onvalue true -offvalue false \
			-text "highlight button" \
			-command "::ced::c_highlight_button_refresh $w; ::ced::c_highlight_button_poke"
	pack  $w.f1.r -side left
	scale $w.f1.scale -from 1 -to 36 -variable ::ced::c_link(highlightbutton) \
			-command ::ced::c_highlight_button_poke
	pack  $w.f1.scale -side left

	c_highlight_button_refresh $w
	c_highlight_button_poke
}

#############################################################################
## proc used by most Link commands to create the common command interface.

proc c_cmds_link_create {cmd_description} {
	#dprint 3 "> c_cmds_link_create {cmd_description}"
	#set ::ced::c_link(highlightbutton) 0

	create_commandframe

	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text $cmd_description -anchor w -justify left
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	c_highlight_button_create $::ced::widget(cmdFrame).cmd.hbcmd
	pack  $::ced::widget(cmdFrame).cmd.hbcmd -side left -fill x

	adjust_win_size
}

#############################################################################
## create the Link part of a Set* command, with a checkbutton and a menubutton
## that allow the user to select a specific Link sub-command.
## This proc is used by most Set* commands.

proc c_link_create {checktext} {
	#dprint 3 "> c_link_create {checktext}"

	set ::ced::c_iflink(link_exists) true
	set ::ced::c_link(active) true

	frame $::ced::widget(cmdFrame).lnk
	frame $::ced::widget(cmdFrame).lnk.f1
	pack  $::ced::widget(cmdFrame).lnk.f1 -side top -fill x
	checkbutton $::ced::widget(cmdFrame).lnk.f1.c -text $checktext -variable ::ced::c_link(check) \
			-command "::ced::c_link_checked" -onvalue true -offvalue false
	pack  $::ced::widget(cmdFrame).lnk.f1.c -side left

	menubutton $::ced::widget(cmdFrame).lnk.f1.m -textvariable ::ced::c_link(text) -pady 1 \
			-menu $::ced::widget(cmdFrame).lnk.f1.m.mnu -width 11
	pack  $::ced::widget(cmdFrame).lnk.f1.m -side left -pady 2
	menu  $::ced::widget(cmdFrame).lnk.f1.m.mnu -tearoff 0
	foreach i $::ced::link_commands_list {
		if {[lindex $i 0] == "-"} {
			$::ced::widget(cmdFrame).lnk.f1.m.mnu add separator
		} else {
			$::ced::widget(cmdFrame).lnk.f1.m.mnu add command -label "[lindex $i 2]" \
				-command "set ::ced::c_link(k) [lindex $i 0]; set ::ced::c_link(ll) [lindex $i 1]; ::ced::c_link_configure"
		}
	}
	c_link_configure

	return $::ced::widget(cmdFrame).lnk
}
proc c_link_configure {} {
	#dprint 3 "> c_link_create:c_link_configure {}"
	switch $::ced::c_link(k) {
		1 {
			foreach i $::ced::link_commands_list {
				if {$::ced::c_link(k) == [lindex $i 0] && $::ced::c_link(ll) == [lindex $i 1]} {
					set ::ced::c_link(text) "[lindex $i 2]"
					break
				}
			}
		}
		4 {set ::ced::c_link(text) LinkPGCN}
		5 {set ::ced::c_link(text) LinkPTT}
		6 {set ::ced::c_link(text) LinkPGN}
		7 {set ::ced::c_link(text) LinkCN}
	}
	c_link_refresh
	c_link_poke
}

#############################################################################
## Create the IF part of most of the DVD commands.
## Args: in: containing frame.
##       type: must be 1 for operand 2 allowing value, gprm and sprm,
##             or 2 for operand 2 allowing only gprm and sprm.

proc c_if_create {in type} {
	#dprint 3 "> c_if_create {in type}"
	set ::ced::c_iflink(link_exists) false
	set ::ced::c_if(active) true
	set ::ced::c_if(type) $type

	# if type is 2, operand 2 may not be a value
	if {$type == 2 && $::ced::c_if(radioOp2) == 1} {set ::ced::c_if(radioOp2) 2}

	# main IF frame
	destroy $in.if
	frame $in.if ;# -borderwidth 2 -relief groove

	# if checkbutton
	checkbutton $in.if.ifcheck -command {::ced::c_if_checked; ::ced::c_if_poke} \
			-text {If} -variable ::ced::c_if(check) -onvalue true -offvalue false
	pack  $in.if.ifcheck -side left
	label $in.if.gprmlabel -text {GPRM}
	pack  $in.if.gprmlabel -side left

	# GPRM or SPRM operand 1
	scale $in.if.gprm -from 0 -to 15 -variable ::ced::c_if(gprm1) -command ::ced::c_if_poke
	pack  $in.if.gprm -side left

	# Operation
	menubutton $in.if.operator -textvariable ::ced::c_if(operator) -pady 1 -width 2 \
			-menu $in.if.operator.mnu
	pack  $in.if.operator -side left -padx 4
	menu  $in.if.operator.mnu -tearoff 0
	foreach i $::ced::c_compare_operators_list {
		$in.if.operator.mnu add command -label "[lindex $i 1]   [lindex $i 2]" \
				-command "set ::ced::c_if(operator) [lindex $i 2]; set ::ced::c_if(operatornum) [lindex $i 0]; ::ced::c_if_poke"
	}

	# GPRM or SPRM (or value if type is 1) operand 2
	if {$type == 1} {
		c_regval_create $in.if.regval2 ::ced::c_if_poke ::ced::c_if(radioOp2) ::ced::c_if(gprm2) ::ced::c_if(sprm2) ::ced::c_if(value2) 0 0XFFFF
	} else {
		c_regval_create $in.if.regval2 ::ced::c_if_poke ::ced::c_if(radioOp2) ::ced::c_if(gprm2) ::ced::c_if(sprm2) {} 0 0XFFFF
	}
	pack  $in.if.regval2 -side left

	# then
	label $in.if.then -text "then"
	pack  $in.if.then -side left

	c_if_refresh
	c_if_poke
	return $in.if
}

#############################################################################
## proc called when the Link checkbox value change, to enable or disable the IF
## part of the interface.
## Refresh both IF and Link parts of the interface, and update the hex string.

proc c_link_checked {} {
	#dprint 3 "> c_link_checked {}"
	if {$::ced::c_link(check)} {
		set ::ced::c_if(check) false
		set ::ced::c_if(active) false
	} else {
		set ::ced::c_if(active) true
	}
	c_if_refresh
	c_link_refresh
	c_link_poke
}

#############################################################################
## proc called when the IF checkbox value change, to enable or disable the Link
## part of the interface.
## Refresh both IF and Link parts of the interface, and update the hex string.

proc c_if_checked {} {
	#dprint 3 "> c_if_checked {}"
	if {$::ced::c_if(check)} {
		if {$::ced::c_iflink(link_exists)} {
			set ::ced::c_link(check) false
			set ::ced::c_link(active) false
			c_link_refresh
		}
	} else {
		if {$::ced::c_iflink(link_exists)} {
			set ::ced::c_link(active) true
			c_link_refresh
		}
	}
	c_if_refresh
}

#############################################################################
## refresh the (sub-) Link part of the interface

proc c_link_refresh {} {
	#dprint 3 "> c_link_refresh {}"
	if {$::ced::c_link(active)} {
		$::ced::widget(cmdFrame).lnk.f1.c configure -state normal
	} else {
		set ::ced::c_link(check) false
		$::ced::widget(cmdFrame).lnk.f1.c configure -state disabled
	}
	destroy $::ced::widget(cmdFrame).lnk.f2
	destroy $::ced::widget(cmdFrame).lnk.f3

	if {! $::ced::c_link(check)} {
		$::ced::widget(cmdFrame).lnk.f1.m configure -state disabled
	} else {
		if {$::ced::c_link(k) == 0} {set ::ced::c_link(k) 1}
		switch $::ced::c_link(k) {
			1 {set text [lindex $::ced::c_link_1_descriptions $::ced::c_link(ll)]}
			4 {set text "Jump to PGC"}
			5 {set text "Jump to Chapter (PTT)"}
			6 {set text "Jump to Program"}
			7 {set text "Jump to Cell"}
			default {set text "???"}
		}
		frame $::ced::widget(cmdFrame).lnk.f2
		pack  $::ced::widget(cmdFrame).lnk.f2 -side top -fill x -pady 2
		label $::ced::widget(cmdFrame).lnk.f2.l -text $text -justify left
		pack  $::ced::widget(cmdFrame).lnk.f2.l -side left
		frame $::ced::widget(cmdFrame).lnk.f3
		pack  $::ced::widget(cmdFrame).lnk.f3 -side top -fill x -pady 2
		switch $::ced::c_link(k) {
			0 - 1 {
				c_highlight_button_create $::ced::widget(cmdFrame).lnk.f3.hb
				pack $::ced::widget(cmdFrame).lnk.f3.hb -side left
			}
			4 {
				set ::ced::c_link(pgcn) [force_range $::ced::c_link(pgcn) 1 32767]
				entry $::ced::widget(cmdFrame).lnk.f2.ent -textvariable ::ced::c_link(pgcn) -width 6 \
						-validate all -validatecommand "::ced::value_entry_validate 1 32767 {%d} {%P} 1"
				pack $::ced::widget(cmdFrame).lnk.f2.ent -side left
				bind $::ced::widget(cmdFrame).lnk.f2.ent <KeyRelease> {::ced::poke_hex_word 6 $::ced::c_link(pgcn)}
				label $::ced::widget(cmdFrame).lnk.f2.lab2 -text "(1-32767)  in the same domain."
				pack $::ced::widget(cmdFrame).lnk.f2.lab2 -side left
				destroy $::ced::widget(cmdFrame).lnk.f3
			}
			5 {
				set ::ced::c_link(pttn) [force_range $::ced::c_link(pttn) 1 999]
				entry $::ced::widget(cmdFrame).lnk.f2.ent -textvariable ::ced::c_link(pttn) -width 5 \
						-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
				pack $::ced::widget(cmdFrame).lnk.f2.ent -side left -padx 2
				bind $::ced::widget(cmdFrame).lnk.f2.ent <KeyRelease> {::ced::poke_LinkPTT}
				label $::ced::widget(cmdFrame).lnk.f2.lab2 -text "(1-999) in the current title"
				pack $::ced::widget(cmdFrame).lnk.f2.lab2 -side left
				c_highlight_button_create $::ced::widget(cmdFrame).lnk.f3.hb
				pack  $::ced::widget(cmdFrame).lnk.f3.hb -side left -fill x -pady 4
			}
			6 {
				set ::ced::c_link(pgn) [force_range $::ced::c_link(pgn) 1 99]
				entry $::ced::widget(cmdFrame).lnk.f2.ent -textvariable ::ced::c_link(pgn) -width 4 \
						-validate all -validatecommand "::ced::value_entry_validate 1 99 {%d} {%P} 1"
				pack $::ced::widget(cmdFrame).lnk.f2.ent -side left -padx 2
				bind $::ced::widget(cmdFrame).lnk.f2.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_link(pgn)}
				label $::ced::widget(cmdFrame).lnk.f2.lab2 -text "(1-99)  in the current PGC"
				pack $::ced::widget(cmdFrame).lnk.f2.lab2 -side left
				c_highlight_button_create $::ced::widget(cmdFrame).lnk.f3.hb
				pack  $::ced::widget(cmdFrame).lnk.f3.hb -side left -fill x -pady 4
			}
			7 {
				set ::ced::c_link(cell) [force_range $::ced::c_link(cell) 1 255]
				entry $::ced::widget(cmdFrame).lnk.f2.ent -textvariable ::ced::c_link(cell) -width 4 \
						-validate all -validatecommand "::ced::value_entry_validate 1 255 {%d} {%P} 1"
				pack $::ced::widget(cmdFrame).lnk.f2.ent -side left
				bind $::ced::widget(cmdFrame).lnk.f2.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_link(cell)}
				label $::ced::widget(cmdFrame).lnk.f2.lab2 -text "(1-255)  in the current PGC"
				pack $::ced::widget(cmdFrame).lnk.f2.lab2 -side left
				c_highlight_button_create $::ced::widget(cmdFrame).lnk.f3.hb
				pack  $::ced::widget(cmdFrame).lnk.f3.hb -side left -fill x -pady 4
			}
		}
		$::ced::widget(cmdFrame).lnk.f1.m configure -state normal
		if {[winfo exists $::ced::widget(cmdFrame).lnk.f3.hb]} {
			c_highlight_button_refresh $::ced::widget(cmdFrame).lnk.f3.hb
		}
	}
	adjust_win_size
}

#############################################################################
## refresh the IF part of the interface

proc c_if_refresh {} {
	#dprint 3 "> c_if_refresh {}"
	if {$::ced::c_if(active)} {
		$::ced::widget(cmdFrame).if.ifcheck configure -state normal
	} else {
		set ::ced::c_if(check) false
		$::ced::widget(cmdFrame).if.ifcheck configure -state disabled
	}
	if {$::ced::c_if(check)} {
		# if command
		$::ced::widget(cmdFrame).if.gprmlabel configure -foreground black
		scale_state $::ced::widget(cmdFrame).if.gprm normal
		$::ced::widget(cmdFrame).if.operator configure -state normal
		c_regval_refresh_state $::ced::widget(cmdFrame).if.regval2 normal
		$::ced::widget(cmdFrame).if.then configure -foreground black
	} else {
		# no IF command
		$::ced::widget(cmdFrame).if.gprmlabel configure -foreground gray
		scale_state $::ced::widget(cmdFrame).if.gprm disabled
		$::ced::widget(cmdFrame).if.operator configure -state disabled
		c_regval_refresh_state $::ced::widget(cmdFrame).if.regval2 disabled
		$::ced::widget(cmdFrame).if.then configure -foreground gray
	}
}

#############################################################################
## Update the hexcommand to reflect the Link part of the interface.

proc c_link_poke {args} {
	#dprint 3 "> c_link_poke {args}"
	# if "if" is selected, then no link is permitted
	if {!$::ced::c_if(check)} {
		if {$::ced::c_link(check)} {
			poke_hex_nibble 1 0 0
			#poke_hex_byte 2 0	;# WARNING: removed- possible bug
			poke_hex_nibble 1 1 $::ced::c_link(k)
			switch $::ced::c_link(k) {
				1 {
					poke_hex_byte 6 0
					poke_hex_byte 7 $::ced::c_link(ll)
				}
				4 {
					poke_hex_word 6 $::ced::c_link(pgcn)
				}
				5 {
					poke_hex_word 6 $::ced::c_link(pttn)
				}
				6 {
					poke_hex_byte 6 0
					poke_hex_byte 7 $::ced::c_link(pgn)
				}
				7 {
					poke_hex_byte 6 0
					poke_hex_byte 7 $::ced::c_link(cell)
				}
			}
			if {$::ced::c_link(k) != 4} {
				c_highlight_button_poke
			}
		} else {
			# remove the link(ll) and the highlight button values
			poke_hex_byte 7 0
			poke_hex_byte 6 0
			# remove the link(k) value (byte 1, lower nibble)
			# but ONLY if the command is NOT SetCLik, CSetCLnk or CSetLnk
			# because they use this nibble to store the destination register
			if {[peek_hex_nibble 0 0] < 8} {
				poke_hex_nibble 1 1 0
			}
		}
	}
}

#############################################################################
## Update the hexcommand to reflect the IF part of the interface (generic).

proc c_if_poke {args} {
	#dprint 3 "> c_if_poke {args}"
	if {$::ced::c_if(type) == 1} {
		c_if_poke1
	} else {
		c_if_poke2
	}
}

#############################################################################
## Update the hexcommand to reflect the IF part of the interface, when the IF
## operand 2 allow value, gprm and sprm.

proc c_if_poke1 {} {
	#dprint 3 "> c_if_poke1 {}"
	if {! $::ced::c_if(check)} {
		# no IF command: remove IF values in hex string
		poke_hex_nibble 1 0 0
		poke_hex_byte 3 0
		poke_hex_byte 4 0
		poke_hex_byte 5 0
	} else {
		# if command
		set add 0
		if {$::ced::c_if(radioOp2) == 1} {set add 8}
		poke_hex_nibble 1 0 [expr {$::ced::c_if(operatornum)+$add}]

		# adjust hex string upon state of operand 1
		poke_hex_byte 3 $::ced::c_if(gprm1)

		# adjust hex string upon state of operand 2
		switch $::ced::c_if(radioOp2) {
			1 {
				# compare with value
				set val $::ced::c_if(value2)
				if {$val == ""} {set val 0}
				poke_hex_word 4 $val
			}
			2 {
				# compare with GPRM
				poke_hex_byte 4 0
				poke_hex_byte 5 $::ced::c_if(gprm2)
			}
			3 {
				# compare with SPRM
				poke_hex_byte 4 0
				poke_hex_byte 5 [expr {$::ced::c_if(sprm2) | 0X80}]
			}
		}
	}
}

#############################################################################
## Update the hexcommand to reflect the IF part of the interface, when the IF
## operand 2 allow only gprm and sprm.

proc c_if_poke2 {} {
	#dprint 3 "> c_if_poke2 {}"
	if {$::ced::c_iflink(link_exists) && $::ced::c_link(check)} return

	if {! $::ced::c_if(check)} {
		# no IF command: remove IF values in hex string
		poke_hex_nibble 1 0 0
		poke_hex_byte 6 0
		poke_hex_byte 7 0
	} else {
		# adjust hex string upon state of operand 1
		poke_hex_nibble 1 0 $::ced::c_if(operatornum)
		poke_hex_byte 6 $::ced::c_if(gprm1)
		# adjust hex string upon state of operand 2
		if {$::ced::c_if(radioOp2) == 2} {
			# compare with GPRM
			poke_hex_byte 7 $::ced::c_if(gprm2)
		} else {
			# compare with SPRM
			poke_hex_byte 7 [expr {$::ced::c_if(sprm2) | 0X80}]
		}
	}
}

#############################################################################
## proc used to destroy the old DVD command interface, and initialize a new one.
## Also, clear all highlights in the hex entry widget.
## Normally called by all the specific DVD command interface procs.

proc create_commandframe {} {
	#dprint 3 "> create_commandframe {}"

	c_hex_entry_refresh
	c_hex_entry_highlight_clear

	destroy $::ced::widget(cmdFrame)
	frame $::ced::widget(cmdFrame) -borderwidth 2 -relief ridge -padx 4 -pady 4
	pack $::ced::widget(cmdFrame) -side top -pady 4 -padx 4 -fill x
}

#############################################################################
#############################################################################
## DVD COMMANDS CREATION PROCEDURES
#############################################################################
#############################################################################
## Following are procs that are called whenever a DVD command is called
## either when the hex entry is validated or by a command menu.
#############################################################################
#############################################################################

#############################################################################
## fake DVD command: illegal_command
## This is a fake DVD command called when the hex entry validation is unable to
## detect a valid DVD command.  Prints out an error message.

proc cmd_illegal_command {cmd msg} {
	#dprint 3 "> cmd_illegal_command {cmd msg}"
	$::ced::widget(CommandName) configure -text ""

	set t "\nERROR: Illegal $cmd command!\n\n$msg\n"

	create_commandframe
	frame $::ced::widget(cmdFrame).f
	pack  $::ced::widget(cmdFrame).f
	label $::ced::widget(cmdFrame).f.lab -text $t
	pack  $::ced::widget(cmdFrame).f.lab -fill x
	frame $::ced::widget(cmdFrame).f.pad -width 460
	pack  $::ced::widget(cmdFrame).f.pad
	bell
}

#############################################################################
## DVD command: NOP

proc cmd_NOP {} {
	#dprint 3 "> cmd_NOP {}"
	$::ced::widget(CommandName) configure -text "NOP"
	set ::hexcommand "00 00 00 00 00 00 00 00"

	set ::ced::c_if(check) false

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	label $::ced::widget(cmdFrame).lab -text "Do nothing." -anchor w
	pack  $::ced::widget(cmdFrame).lab -side top -fill x

	label $::ced::widget(cmdFrame).lab2 -anchor w \
		-text "NOTE: The IF check is totally useless with NOP!"
	pack  $::ced::widget(cmdFrame).lab2 -side top

	adjust_win_size
}
#############################################################################
## DVD command: GoTo

proc cmd_GoTo {} {
	#dprint 3 "> cmd_GoTo {}"
	$::ced::widget(CommandName) configure -text "GoTo"
	set ::hexcommand "00 01 00 00 00 00 00 01"

	set ::ced::c_goto(linenum) [force_range $::ced::c_goto(linenum) 1 128]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Goto line"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_goto(linenum) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 128 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent  -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_goto(linenum)}

	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-128)."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2

	poke_hex_byte 7 $::ced::c_goto(linenum)

	adjust_win_size
}
#############################################################################
## DVD command: Break

proc cmd_Break {} {
	#dprint 3 "> cmd_Break {}"
	$::ced::widget(CommandName) configure -text "Break"
	set ::hexcommand "00 02 00 00 00 00 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Terminate execution of pre, post or cell commands."
	pack  $::ced::widget(cmdFrame).cmd.lab -side left

	adjust_win_size
}
#############################################################################
## DVD command: SetTmpPML
proc cmd_SetTmpPML {} {
	#dprint 3 "> cmd_SetTmpPML {}"
	$::ced::widget(CommandName) configure -text "SetTmpPML"
	set ::hexcommand "00 03 00 00 00 00 01 01"

	set ::ced::c_goto(linenum) [force_range $::ced::c_goto(linenum) 1 128]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set Parental Management Level"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	scale $::ced::widget(cmdFrame).cmd.scale -from 1 -to 8 -variable ::ced::c_TmpPNL(level) \
			-command {::ced::poke_hex_byte 6}
	pack  $::ced::widget(cmdFrame).cmd.scale -side left
	checkbutton $::ced::widget(cmdFrame).cmd.none -text none -onvalue true -offvalue false \
			-variable ::ced::c_settmppml(none) -command ::ced::validate_SetTmpPML
	pack  $::ced::widget(cmdFrame).cmd.none -side left

	frame $::ced::widget(cmdFrame).cmd2
	pack  $::ced::widget(cmdFrame).cmd2 -side top -fill x
	label $::ced::widget(cmdFrame).cmd2.lab2 -text "Then, if successful, goto line"
	pack  $::ced::widget(cmdFrame).cmd2.lab2 -side left
	entry $::ced::widget(cmdFrame).cmd2.ent -textvariable ::ced::c_goto(linenum) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 128 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd2.ent  -side left -pady 4
	bind  $::ced::widget(cmdFrame).cmd2.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_goto(linenum)}
	label $::ced::widget(cmdFrame).cmd2.lab3 -text "(0-128)."
	pack  $::ced::widget(cmdFrame).cmd2.lab3 -side left -padx 2

	validate_SetTmpPML

	adjust_win_size
}
proc validate_SetTmpPML {args} {
	if {$::ced::c_settmppml(none)} {
		scale_state $::ced::widget(cmdFrame).cmd.scale disabled
		#$::ced::widget(cmdFrame).cmd2.lab2 configure -foreground gray
		#$::ced::widget(cmdFrame).cmd2.ent configure -state disabled
		#$::ced::widget(cmdFrame).cmd2.lab3 configure -foreground gray
		::ced::poke_hex_byte 6 15
		#::ced::poke_hex_byte 7 0
	} else {
		scale_state $::ced::widget(cmdFrame).cmd.scale normal
		#$::ced::widget(cmdFrame).cmd2.lab2 configure -foreground black
		#$::ced::widget(cmdFrame).cmd2.ent configure -state normal
		#$::ced::widget(cmdFrame).cmd2.lab3 configure -foreground black
		::ced::poke_hex_byte 6 $::ced::c_TmpPNL(level)
		#::ced::poke_hex_byte 7 $::ced::c_goto(linenum)
	}
	::ced::poke_hex_byte 7 $::ced::c_goto(linenum)
}

#############################################################################
## DVD command: LinkNoLink

proc cmd_LinkNoLink {} {
	#dprint 3 "> cmd_LinkNoLink {}"
	$::ced::widget(CommandName) configure -text "LinkNoLink"
	set ::hexcommand "20 01 00 00 00 00 00 00"

	c_cmds_link_create "Don't link to anything,"
}

#############################################################################
## DVD command: LinkTopCell

proc cmd_LinkTopCell {} {
	#dprint 3 "> cmd_LinkTopCell {}"
	$::ced::widget(CommandName) configure -text "LinkTopCell"
	set ::hexcommand "20 01 00 00 00 00 00 01"

	c_cmds_link_create "Restart the current cell,"
}
#############################################################################
## DVD command: LinkNextCell

proc cmd_LinkNextCell {} {
	#dprint 3 "> cmd_LinkNextCell {}"
	$::ced::widget(CommandName) configure -text "LinkNextCell"
	set ::hexcommand "20 01 00 00 00 00 00 02"

	c_cmds_link_create "Start the next cell in the current PGC,"
}
#############################################################################
## DVD command: LinkPrevCell

proc cmd_LinkPrevCell {} {
	#dprint 3 "> cmd_LinkPrevCell {}"
	$::ced::widget(CommandName) configure -text "LinkPrevCell"
	set ::hexcommand "20 01 00 00 00 00 00 03"

	c_cmds_link_create "Start the previous cell in the current PGC,"
}
#############################################################################
## DVD command: LinkTopPG

proc cmd_LinkTopPG {} {
	#dprint 3 "> cmd_LinkTopPG {}"
	$::ced::widget(CommandName) configure -text "LinkTopPG"
	set ::hexcommand "20 01 00 00 00 00 00 05"

	c_cmds_link_create "Restart the current program,"
}
#############################################################################
## DVD command: LinkNextPG

proc cmd_LinkNextPG {} {
	#dprint 3 "> cmd_LinkNextPG {}"
	$::ced::widget(CommandName) configure -text "LinkNextPG"
	set ::hexcommand "20 01 00 00 00 00 00 06"

	c_cmds_link_create "Start the next program in the current PGC,"
}
#############################################################################
## DVD command: LinkPrevPG

proc cmd_LinkPrevPG {} {
	#dprint 3 "> cmd_LinkPrevPG {}"
	$::ced::widget(CommandName) configure -text "LinkPrevPG"
	set ::hexcommand "20 01 00 00 00 00 00 07"

	c_cmds_link_create "Start the previous program in the current PGC,"
}
#############################################################################
## DVD command: LinkTopPGC

proc cmd_LinkTopPGC {} {
	#dprint 3 "> cmd_LinkTopPGC {}"
	$::ced::widget(CommandName) configure -text "LinkTopPGC"
	set ::hexcommand "20 01 00 00 00 00 00 09"

	c_cmds_link_create "Restart the current PGC,"
}
#############################################################################
## DVD command: LinkNextPGC

proc cmd_LinkNextPGC {} {
	#dprint 3 "> cmd_LinkNextPGC {}"
	$::ced::widget(CommandName) configure -text "LinkNextPGC"
	set ::hexcommand "20 01 00 00 00 00 00 0A"

	c_cmds_link_create "Start the PGC pointed to by the 'next' link,"
}
#############################################################################
## DVD command: LinkPrevPGC

proc cmd_LinkPrevPGC {} {
	#dprint 3 "> cmd_LinkPrevPGC {}"
	$::ced::widget(CommandName) configure -text "LinkPrevPGC"
	set ::hexcommand "20 01 00 00 00 00 00 0B"

	c_cmds_link_create "Start the PGC pointed to by the 'prev' link,"
}
#############################################################################
## DVD command: LinkGoUpPGC

proc cmd_LinkGoUpPGC {} {
	#dprint 3 "> cmd_LinkGoUpPGC {}"
	$::ced::widget(CommandName) configure -text "LinkGoUpPGC"
	set ::hexcommand "20 01 00 00 00 00 00 0C"

	c_cmds_link_create "Start the PGC pointed to by the 'go up' link,"
}
#############################################################################
## DVD command: LinkTailPGC

proc cmd_LinkTailPGC {} {
	#dprint 3 "> cmd_LinkTailPGC {}"
	$::ced::widget(CommandName) configure -text "LinkTailPGC"
	set ::hexcommand "20 01 00 00 00 00 00 0D"

	c_cmds_link_create "Skip to the end of the current PGC,\nexecutes the post-commands,"
}
#############################################################################
## DVD command: RSM

proc cmd_RSM {} {
	#dprint 3 "> cmd_RSM {}"
	$::ced::widget(CommandName) configure -text "RSM"
	set ::hexcommand "20 01 00 00 00 00 00 10"

	c_cmds_link_create "Resume at the location where playback\nwas suspended by a CallSS or a menu call,"
}
#############################################################################
## DVD command: LinkPGCN

proc cmd_LinkPGCN {} {
	#dprint 3 "> cmd_LinkPGCN {}"
	$::ced::widget(CommandName) configure -text "LinkPGCN"
	set ::hexcommand "20 04 00 00 00 00 00 01"

	set ::ced::c_link(pgcn) [force_range $::ced::c_link(pgcn) 1 32767]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to PGC number"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_link(pgcn) -width 6 \
			-validate all -validatecommand "::ced::value_entry_validate 1 32767 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_word 6 $::ced::c_link(pgcn)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-32767)  in the same domain."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left

	poke_hex_word 6 $::ced::c_link(pgcn)

	adjust_win_size
}
#############################################################################
## DVD command: LinkPTT

proc cmd_LinkPTT {} {
	#dprint 3 "> cmd_LinkPTT {}"
	$::ced::widget(CommandName) configure -text "LinkPTT"
	set ::hexcommand "20 05 00 00 00 00 00 01"

	set ::ced::c_link(pttn) [force_range $::ced::c_link(pttn) 1 999]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to Chapter (PTT)"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_link(pttn) -width 5 \
			-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left -padx 2
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_LinkPTT}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-999) in the current title,"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left

	c_highlight_button_create $::ced::widget(cmdFrame).hbcmd
	pack  $::ced::widget(cmdFrame).hbcmd -side left -fill x -pady 4

	poke_LinkPTT
	c_highlight_button_poke

	adjust_win_size
}

#############################################################################
## DVD command: LinkPGN

proc cmd_LinkPGN {} {
	#dprint 3 "> cmd_LinkPGN {}"
	$::ced::widget(CommandName) configure -text "LinkPGN"
	set ::hexcommand "20 06 00 00 00 00 00 01"

	set ::ced::c_link(pgn) [force_range $::ced::c_link(pgn) 1 99]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).pad
	pack  $::ced::widget(cmdFrame).pad -side top -pady 4 -fill x
	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to Program"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_link(pgn) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 99 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left -padx 2
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_link(pgn)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-99)  in the current PGC,"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left

	c_highlight_button_create $::ced::widget(cmdFrame).hbcmd
	pack  $::ced::widget(cmdFrame).hbcmd -side left -fill x -pady 4

	poke_hex_byte 7 $::ced::c_link(pgn)

	adjust_win_size
}

#############################################################################
## DVD command: LinkCN

proc cmd_LinkCN {} {
	#dprint 3 "> cmd_LinkCN {}"
	$::ced::widget(CommandName) configure -text "LinkCN"
	set ::hexcommand "20 07 00 00 00 00 00 01"

	set ::ced::c_link(cell) [force_range $::ced::c_link(cell) 1 255]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 1] -side top -fill x

	frame $::ced::widget(cmdFrame).pad
	pack  $::ced::widget(cmdFrame).pad -side top -pady 4
	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to Cell"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_link(cell) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 255 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 7 $::ced::c_link(cell)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-255)  in the current PGC,"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left

	c_highlight_button_create $::ced::widget(cmdFrame).hbcmd
	pack  $::ced::widget(cmdFrame).hbcmd -side left -fill x -pady 4

	poke_hex_byte 7 $::ced::c_link(cell)

	adjust_win_size
}
#############################################################################
## DVD command: Exit

proc cmd_Exit {} {
	#dprint 3 "> cmd_Exit {}"
	$::ced::widget(CommandName) configure -text "Exit"
	set ::hexcommand "30 01 00 00 00 00 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Stop all playback."
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -pady 6

	adjust_win_size
}
#############################################################################
## DVD command: JumpTT

proc cmd_JumpTT {} {
	#dprint 3 "> cmd_JumpTT {}"
	$::ced::widget(CommandName) configure -text "JumpTT"
	set ::hexcommand "30 02 00 00 00 01 00 00"

	set ::ced::c_jump(title) [force_range $::ced::c_jump(title) 1 99]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to the entry PGC of title"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(title) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 99 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 5 $::ced::c_jump(title)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-99)."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2

	poke_hex_byte 5 $::ced::c_jump(title)

	adjust_win_size
}
#############################################################################
## DVD command: JumpVTS_TT

proc cmd_JumpVTS_TT {} {
	#dprint 3 "> cmd_JumpVTS_TT {}"
	$::ced::widget(CommandName) configure -text "JumpVTS_TT"
	set ::hexcommand "30 03 00 00 00 01 00 00"

	set ::ced::c_jump(title) [force_range $::ced::c_jump(title) 1 99]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to the entry PGC of TTN"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(title) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 99 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-99) , in the same title set (VTS)."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 5 $::ced::c_jump(title)}

	poke_hex_byte 5 $::ced::c_jump(title)

	adjust_win_size
}
#############################################################################
## DVD command: JumpVTS_PTT

proc cmd_JumpVTS_PTT {} {
	#dprint 3 "> cmd_JumpVTS_PTT {}"
	$::ced::widget(CommandName) configure -text "JumpVTS_PTT"
	set ::hexcommand "30 05 00 01 00 01 00 00"

	set ::ced::c_jump(ptt) [force_range $::ced::c_jump(ptt) 1 999]
	set ::ced::c_jump(title) [force_range $::ced::c_jump(title) 1 99]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4

	label $::ced::widget(cmdFrame).cmd.lab2 -text "Jump to TTN"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent2 -textvariable ::ced::c_jump(title) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 1 99 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent2 -side left
	bind  $::ced::widget(cmdFrame).cmd.ent2 <KeyRelease> {::ced::poke_hex_byte 5 $::ced::c_jump(title)}

	label $::ced::widget(cmdFrame).cmd.lab -text "(1-99) , Chapter (PTT)"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(ptt) -width 5 \
			-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_word 2 $::ced::c_jump(ptt)}

	label $::ced::widget(cmdFrame).cmd.lab3 -text "(1-999) , in the same title set (VTS)."
	pack  $::ced::widget(cmdFrame).cmd.lab3 -side left -padx 2

	poke_hex_word 2 $::ced::c_jump(ptt)
	poke_hex_byte 5 $::ced::c_jump(title)

	adjust_win_size
}
#############################################################################
## DVD command: JumpSS-FP

proc cmd_JumpSS-FP {} {
	#dprint 3 "> cmd_JumpSS-FP {}"
	$::ced::widget(CommandName) configure -text "JumpSS-FP"
	set ::hexcommand "30 06 00 00 00 00 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to the First Play PGC."
	pack  $::ced::widget(cmdFrame).cmd.lab  -side left

	adjust_win_size
}
#############################################################################
## DVD command: JumpSS-VMGM-menu

proc cmd_JumpSS-VMGM-menu {} {
	#dprint 3 "> cmd_JumpSS-VMGM-menu {}"
	$::ced::widget(CommandName) configure -text "JumpSS-VMGM-menu"
	set ::hexcommand "30 06 00 00 00 42 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to VMGM Title menu."
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	adjust_win_size
}
#############################################################################
## DVD command: JumpSS-VTSM

proc cmd_JumpSS-VTSM {} {
	#dprint 3 "> cmd_JumpSS-VTSM {}"
	$::ced::widget(CommandName) configure -text "JumpSS-VTSM"
	set ::hexcommand "30 06 00 01 01 83 00 00"

	set ::ced::c_jump(title) [force_range $::ced::c_jump(title) 0 99]
	set ::ced::c_jump(vts)   [force_range $::ced::c_jump(vts)   0 99]
	set ::ced::c_jump(menunumber) [force_range $::ced::c_jump(menunumber) 3 7]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to VTS"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	entry $::ced::widget(cmdFrame).cmd.ent2 -textvariable ::ced::c_jump(vts) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 99 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent2 -side left
	bind  $::ced::widget(cmdFrame).cmd.ent2 <KeyRelease> {::ced::poke_hex_byte 4 $::ced::c_jump(vts)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(0-99) ,"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2
	menubutton $::ced::widget(cmdFrame).cmd.m -textvariable ::ced::c_jump(menulabel) \
			-pady 1 -width 10 -menu $::ced::widget(cmdFrame).cmd.m.mnu -width 13
	pack  $::ced::widget(cmdFrame).cmd.m -side left -pady 2
	menu  $::ced::widget(cmdFrame).cmd.m.mnu -tearoff 0
	foreach i $::ced::c_menus_names_list {
		set num [lindex $i 0]
		set lab "[lindex $i 1] menu"
		$::ced::widget(cmdFrame).cmd.m.mnu add command -label "$lab" \
				-command "set ::ced::c_jump(menulabel) \"$lab\"; set ::ced::c_jump(menunumber) $num; ::ced::poke_hex_nibble 5 1 $num"
	}
	label $::ced::widget(cmdFrame).cmd.lab4 -text "; and set sprm(5) to TTN"
	pack  $::ced::widget(cmdFrame).cmd.lab4 -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(title) -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 99 {%d} {%P} 1"
	::balloon::set_balloon $::ced::widget(cmdFrame).cmd.ent "Remember this TTN\nas last visited"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_byte 3 $::ced::c_jump(title)}
	label $::ced::widget(cmdFrame).cmd.lab5 -text "(1-99)"
	pack  $::ced::widget(cmdFrame).cmd.lab5 -side left -padx 2

	label $::ced::widget(cmdFrame).labmsg -anchor w \
			-text "Note: Setting the VTS to 0 means Last visited VTS.  However, IfoEdit complain about this."
	pack  $::ced::widget(cmdFrame).labmsg -padx 2 -fill x


	poke_hex_byte 3 $::ced::c_jump(title)
	poke_hex_byte 4 $::ced::c_jump(vts)
	poke_hex_nibble 5 1 $::ced::c_jump(menunumber)

	foreach i $::ced::c_menus_names_list {
		if {[lindex $i 0] == $::ced::c_jump(menunumber)} {
			set ::ced::c_jump(menulabel) "[lindex $i 1] menu"
			break
		}
	}
	adjust_win_size
}

#############################################################################
## DVD command: JumpSS-VMGM-pgc

proc cmd_JumpSS-VMGM-pgc {} {
	#dprint 3 "> cmd_JumpSS-VMGM-pgc {}"
	$::ced::widget(CommandName) configure -text "JumpSS-VMGM-pgc"
	set ::hexcommand "30 06 00 01 00 C0 00 00"

	set ::ced::c_jump(pgc) [force_range $::ced::c_jump(pgc) 1 999]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Jump to VMGM PGC"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(pgc) -width 7 \
			-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-999)."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_word 2 $::ced::c_jump(pgc)}

	poke_hex_word 2 $::ced::c_jump(pgc)

	adjust_win_size
}

#############################################################################
## DVD command: CallSS-FP

proc cmd_CallSS-FP {} {
	#dprint 3 "> cmd_CallSS-FP {}"
	$::ced::widget(CommandName) configure -text "CallSS-FP"
	set ::hexcommand "30 08 00 00 01 00 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Call the First Play PGC from a title."
	pack  $::ced::widget(cmdFrame).cmd.lab -side left

	pack [c_resumecell_create $::ced::widget(cmdFrame).resumecell] -side top -fill x

	adjust_win_size
}

#############################################################################
## DVD command: CallSS-VMGM-menu

proc cmd_CallSS-VMGM-menu {} {
	#dprint 3 "> cmd_CallSS-VMGM-menu {}"
	$::ced::widget(CommandName) configure -text "CallSS-VMGM-menu"
	set ::hexcommand "30 08 00 00 01 42 00 00"

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Call the VMGM Title menu from a title."
	pack  $::ced::widget(cmdFrame).cmd.lab -side left

	pack [c_resumecell_create $::ced::widget(cmdFrame).resumecell] -side top -fill x

	adjust_win_size
}

#############################################################################
## DVD command: CallSS-VTSM

proc cmd_CallSS-VTSM {} {
	#dprint 3 "> cmd_CallSS-VTSM {}"
	$::ced::widget(CommandName) configure -text "CallSS-VTSM"
	set ::hexcommand "30 08 00 00 01 80 00 00"

	set ::ced::c_jump(menunumber) [force_range $::ced::c_jump(menunumber) 3 7]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Call the"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	menubutton $::ced::widget(cmdFrame).cmd.m -textvariable ::ced::c_jump(menulabel) \
			-pady 1 -width 10 -menu $::ced::widget(cmdFrame).cmd.m.mnu -width 13
	pack  $::ced::widget(cmdFrame).cmd.m -side left -pady 2
	menu  $::ced::widget(cmdFrame).cmd.m.mnu -tearoff 0
	foreach i $::ced::c_menus_names_list {
		set num [lindex $i 0]
		set lab "[lindex $i 1] menu"
		$::ced::widget(cmdFrame).cmd.m.mnu add command -label "$lab" \
				-command "set ::ced::c_jump(menulabel) \"$lab\"; set ::ced::c_jump(menunumber) $num; ::ced::poke_hex_nibble 5 1 $num"
	}

	label $::ced::widget(cmdFrame).cmd.lab2 -text "of the current VTS."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2

	pack  [c_resumecell_create $::ced::widget(cmdFrame).resumecell] -side top -fill x

	poke_hex_nibble 5 1 $::ced::c_jump(menunumber)

	foreach i $::ced::c_menus_names_list {
		if {[lindex $i 0] == $::ced::c_jump(menunumber)} {
			set ::ced::c_jump(menulabel) "[lindex $i 1] menu"
			break
		}
	}
	adjust_win_size
}

#############################################################################
## DVD command: CallSS-VMGM-pgc

proc cmd_CallSS-VMGM-pgc {} {
	#dprint 3 "> cmd_CallSS-VMGM-pgc {}"
	$::ced::widget(CommandName) configure -text "CallSS-VMGM-pgc"
	set ::hexcommand "30 08 00 00 01 C0 00 00"

	set ::ced::c_jump(pgc) [force_range $::ced::c_jump(pgc) 1 999]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd.lab -text "Call the VMGM PGC"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd.ent -textvariable ::ced::c_jump(pgc) -width 7 \
			-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd.ent -side left
	bind  $::ced::widget(cmdFrame).cmd.ent <KeyRelease> {::ced::poke_hex_word 2 $::ced::c_jump(pgc)}
	label $::ced::widget(cmdFrame).cmd.lab2 -text "(1-999)  from a title."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2

	pack  [c_resumecell_create $::ced::widget(cmdFrame).resumecell] -side top -fill x

	poke_hex_word 2 $::ced::c_jump(pgc)

	adjust_win_size
}

#############################################################################
## DVD command: SetSTN

proc cmd_SetSTN {} {
	#dprint 3 "> cmd_SetSTN {}"

	$::ced::widget(CommandName) configure -text "SetSTN"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "51 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "41 00 00 00 00 00 00 00"
	}

	create_commandframe

	pack  [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x
	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -pady 3 -fill x

	# mode frame
	frame $::ced::widget(cmdFrame).cmd.mode
	pack  $::ced::widget(cmdFrame).cmd.mode -fill x -side top -pady 3
	label $::ced::widget(cmdFrame).cmd.mode.lab -text "Set Audio, Sub-picture and/or Angle to"
	pack  $::ced::widget(cmdFrame).cmd.mode.lab -side left
	radiobutton $::ced::widget(cmdFrame).cmd.mode.r2 -text "constant values" \
			-value 1 -variable ::ced::c_set(sourcemode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.mode.r2 -side left -anchor w
	radiobutton $::ced::widget(cmdFrame).cmd.mode.r1 -text "GPRMs :" \
			-value 2 -variable ::ced::c_set(sourcemode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.mode.r1 -side left -anchor w

	# audio frame
	frame $::ced::widget(cmdFrame).cmd.audio -borderwidth 2 -relief groove
	pack  $::ced::widget(cmdFrame).cmd.audio -side top -fill x -pady 1 -padx 4
	label $::ced::widget(cmdFrame).cmd.audio.lab -text "Audio" -width 9 -anchor w
	pack  $::ced::widget(cmdFrame).cmd.audio.lab -side left
	radiobutton $::ced::widget(cmdFrame).cmd.audio.r1 -text "don't change" \
			-value 0 -variable ::ced::c_setstn(audio_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.audio.r1 -side left
	radiobutton $::ced::widget(cmdFrame).cmd.audio.r2 -text "set to" \
			-value 1 -variable ::ced::c_setstn(audio_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.audio.r2 -side left
	scale $::ced::widget(cmdFrame).cmd.audio.scale -from 0 -to 7 \
			-variable ::ced::c_setstn(audio_value) \
			-command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.audio.scale -side left
	radiobutton $::ced::widget(cmdFrame).cmd.audio.r3 -text "not set" \
			-value 2 -variable ::ced::c_setstn(audio_mode) -command {::ced::validate_SetSTN}
	::balloon::set_balloon $::ced::widget(cmdFrame).cmd.audio.r3 "Mute (no audio)"
	pack  $::ced::widget(cmdFrame).cmd.audio.r3 -side left -anchor w

	# subpic frame
	frame $::ced::widget(cmdFrame).cmd.subpic -borderwidth 2 -relief groove
	pack  $::ced::widget(cmdFrame).cmd.subpic -side top -fill x -pady 1 -padx 4
	label $::ced::widget(cmdFrame).cmd.subpic.lab -text "Sub-picture" -width 9 -anchor w
	pack  $::ced::widget(cmdFrame).cmd.subpic.lab -side left
	radiobutton $::ced::widget(cmdFrame).cmd.subpic.r1 -text "don't change" \
			-value 0 -variable ::ced::c_setstn(subpic_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.subpic.r1 -side left
	radiobutton $::ced::widget(cmdFrame).cmd.subpic.r2 -text "set to" \
			-value 1 -variable ::ced::c_setstn(subpic_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.subpic.r2 -side left
	scale $::ced::widget(cmdFrame).cmd.subpic.scale -from 0 -to 31 \
			-variable ::ced::c_setstn(subpic_value) \
			-command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.subpic.scale -side left
	checkbutton $::ced::widget(cmdFrame).cmd.subpic.onoff -text "on" \
			-variable ::ced::c_setstn(subpic_onoff) -command {::ced::validate_SetSTN} \
			-onvalue true -offvalue false
	::balloon::set_balloon $::ced::widget(cmdFrame).cmd.subpic.onoff "Display the subpicture stream"
	pack  $::ced::widget(cmdFrame).cmd.subpic.onoff -side left
	frame $::ced::widget(cmdFrame).cmd.subpic.fr
	pack  $::ced::widget(cmdFrame).cmd.subpic.fr -side left
	radiobutton $::ced::widget(cmdFrame).cmd.subpic.fr.r3 -text "not set" \
			-value 2 -variable ::ced::c_setstn(subpic_mode) -command {::ced::validate_SetSTN} \
			-bd $::config(min_bd)
	::balloon::set_balloon $::ced::widget(cmdFrame).cmd.subpic.fr.r3 "Don't display any subpicture"
	pack  $::ced::widget(cmdFrame).cmd.subpic.fr.r3 -side top -anchor w
	radiobutton $::ced::widget(cmdFrame).cmd.subpic.fr.r4 -text "force subpic 0" \
			-value 3 -variable ::ced::c_setstn(subpic_mode) -command {::ced::validate_SetSTN} \
			-bd $::config(min_bd)
	::balloon::set_balloon $::ced::widget(cmdFrame).cmd.subpic.fr.r4 "Force subpicture stream 0,\nand display it."
	pack  $::ced::widget(cmdFrame).cmd.subpic.fr.r4 -side top -anchor w

	# angle frame
	frame $::ced::widget(cmdFrame).cmd.angle -borderwidth 2 -relief groove
	pack  $::ced::widget(cmdFrame).cmd.angle -side top -fill x -pady 1 -padx 4
	label $::ced::widget(cmdFrame).cmd.angle.lab -text "Angle" -width 9 -anchor w
	pack  $::ced::widget(cmdFrame).cmd.angle.lab -side left
	radiobutton $::ced::widget(cmdFrame).cmd.angle.r1 -text "don't change" \
			-value 0 -variable ::ced::c_setstn(angle_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.angle.r1 -side left
	radiobutton $::ced::widget(cmdFrame).cmd.angle.r2 -text "set to" \
			-value 1 -variable ::ced::c_setstn(angle_mode) -command {::ced::validate_SetSTN}
	pack  $::ced::widget(cmdFrame).cmd.angle.r2 -side left
	scale $::ced::widget(cmdFrame).cmd.angle.scale -from 1 -to 9 \
			-variable ::ced::c_setstn(angle_value) \
			-command {::ced::validate_SetSTN}
	pack $::ced::widget(cmdFrame).cmd.angle.scale -side left

	c_link_create "And"
	pack $::ced::widget(cmdFrame).lnk -side top -fill x
	c_link_checked
	c_if_checked

	validate_SetSTN

	adjust_win_size
}
proc validate_SetSTN {args} {
	#dprint 3 "> cmd_SetSTN:validate_SetSTN {args}"

	c_if_poke

	if {$::ced::c_set(sourcemode) == 2} {
		if {$::ced::c_setstn(audio_mode) > 1} {set ::ced::c_setstn(audio_mode)  1}
		if {$::ced::c_setstn(subpic_mode)> 1} {set ::ced::c_setstn(subpic_mode) 1}
	}

	if {$::ced::c_setstn(audio_mode)==1} {
		scale_state $::ced::widget(cmdFrame).cmd.audio.scale normal
	} else {
		scale_state $::ced::widget(cmdFrame).cmd.audio.scale disabled
	}
	if {$::ced::c_setstn(subpic_mode)==1} {
		scale_state $::ced::widget(cmdFrame).cmd.subpic.scale normal
	} else {
		scale_state $::ced::widget(cmdFrame).cmd.subpic.scale disabled
	}
	if {$::ced::c_setstn(angle_mode)==1} {
		scale_state $::ced::widget(cmdFrame).cmd.angle.scale normal
	} else {
		scale_state $::ced::widget(cmdFrame).cmd.angle.scale disabled
	}

	if {$::ced::c_set(sourcemode) == 2} {
		poke_hex_nibble 0 0 4
		$::ced::widget(cmdFrame).cmd.audio.scale  configure -to 15 -variable ::ced::c_setstn(audio_gprm)
		$::ced::widget(cmdFrame).cmd.subpic.scale configure -to 15 -variable ::ced::c_setstn(subpic_gprm)
		$::ced::widget(cmdFrame).cmd.angle.scale  configure -from 0 -to 15 -variable ::ced::c_setstn(angle_gprm)
		$::ced::widget(cmdFrame).cmd.audio.r3     configure -state disabled
		$::ced::widget(cmdFrame).cmd.subpic.fr.r3 configure -state disabled
		$::ced::widget(cmdFrame).cmd.subpic.fr.r4 configure -state disabled
		$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state disabled
		if {$::ced::c_setstn(audio_mode)} {
			poke_hex_byte 3 [expr {$::ced::c_setstn(audio_gprm) | 0X80}]
		} else {
			poke_hex_byte 3 0
		}
		if {$::ced::c_setstn(subpic_mode)} {
			poke_hex_byte 4 [expr {$::ced::c_setstn(subpic_gprm)| 0X80}]
		} else {
			poke_hex_byte 4 0
		}
		if {$::ced::c_setstn(angle_mode)} {
			poke_hex_byte 5 [expr {$::ced::c_setstn(angle_gprm) | 0X80}]
		} else {
			poke_hex_byte 5 0
		}
	} else {
		poke_hex_nibble 0 0 5
		$::ced::widget(cmdFrame).cmd.audio.scale  configure -to 7  -variable ::ced::c_setstn(audio_value)
		$::ced::widget(cmdFrame).cmd.subpic.scale configure -to 31 -variable ::ced::c_setstn(subpic_value)
		$::ced::widget(cmdFrame).cmd.angle.scale  configure -from 1 -to 9  -variable ::ced::c_setstn(angle_value)
		$::ced::widget(cmdFrame).cmd.audio.r3     configure -state normal
		$::ced::widget(cmdFrame).cmd.subpic.fr.r3 configure -state normal
		$::ced::widget(cmdFrame).cmd.subpic.fr.r4 configure -state normal
		$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state normal
		if {$::ced::c_setstn(subpic_onoff)} {
			set add 64
		} else {
			set add 0
		}
		switch $::ced::c_setstn(audio_mode) {
			0 {	poke_hex_byte 3 0}
			1 {	poke_hex_byte 3 [expr {$::ced::c_setstn(audio_value) | 0X80}]}
			2 {	poke_hex_byte 3 143}
		}
		switch $::ced::c_setstn(subpic_mode) {
			0 {
				poke_hex_byte 4 0
				$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state disabled
			}
			1 {
				poke_hex_byte 4 [expr {$::ced::c_setstn(subpic_value) + 0X80 + $add}]
				$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state normal
			}
			2 {
				poke_hex_byte 4 190
				$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state disabled
			}
			3 {
				poke_hex_byte 4 255
				$::ced::widget(cmdFrame).cmd.subpic.onoff configure -state disabled
			}
		}
		switch $::ced::c_setstn(angle_mode) {
			0 {	poke_hex_byte 5 0}
			1 {	poke_hex_byte 5 [expr {$::ced::c_setstn(angle_value) | 0X80}]}
		}
	}
	c_link_poke
}

#############################################################################
## DVD command: SetNVTMR

proc cmd_SetNVTMR {} {
	#dprint 3 "> cmd_SetNVTMR {}"
	$::ced::widget(CommandName) configure -text "SetNVTMR"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "52 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "42 00 00 00 00 00 00 00"
	}

	set ::ced::c_setnvtmr(pgc) [force_range $::ced::c_setnvtmr(pgc) 1 999]

	create_commandframe
	pack [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Initialize the Navigation Timer (SPRM 9) to"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_SetNVTMR_validate_and_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4
	label $::ced::widget(cmdFrame).cmd.lab2 -text "seconds."
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left

	frame $::ced::widget(cmdFrame).cmd2
	pack  $::ced::widget(cmdFrame).cmd2 -side top -fill x
	label $::ced::widget(cmdFrame).cmd2.lab -text "When the timer reaches 0 (SPRM 10), jump to Destination PGC"
	pack  $::ced::widget(cmdFrame).cmd2.lab -side left -padx 2
	entry $::ced::widget(cmdFrame).cmd2.e -textvariable ::ced::c_setnvtmr(pgc) -width 6 \
		-validate all -validatecommand "::ced::value_entry_validate 1 999 {%d} {%P} 1"
	pack  $::ced::widget(cmdFrame).cmd2.e -side left -pady 2
	bind  $::ced::widget(cmdFrame).cmd2.e <KeyRelease> {::ced::cmd_SetNVTMR_validate_and_poke}
	label $::ced::widget(cmdFrame).cmd2.lab2 -text "(1-999)."
	pack  $::ced::widget(cmdFrame).cmd2.lab2 -side left -padx 2

	c_link_create "And"
	pack  $::ced::widget(cmdFrame).lnk -side top -fill x

	c_link_checked
	c_if_checked

	cmd_SetNVTMR_validate_and_poke

	adjust_win_size
}

#############################################################################
## DVD command: SetGPRMMD

proc cmd_SetGPRMMD {} {
	#dprint 3 "> cmd_SetGPRMMD {}"

	$::ced::widget(CommandName) configure -text "SetGPRMMD"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "53 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "43 00 00 00 00 00 00 00"
	}

	create_commandframe
	pack  [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set GPRM"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2
	scale $::ced::widget(cmdFrame).cmd.d -from 0 -to 15 \
		-variable ::ced::c_setgprmmd(destination) \
		-command {::ced::cmd_SetGPRMMD_poke}
	pack  $::ced::widget(cmdFrame).cmd.d -side left
	label $::ced::widget(cmdFrame).cmd.lab2 -text "in"
	pack  $::ced::widget(cmdFrame).cmd.lab2 -side left -padx 2
	frame $::ced::widget(cmdFrame).cmd.fm -border 2 -relief groove
	pack  $::ced::widget(cmdFrame).cmd.fm -side left
	radiobutton $::ced::widget(cmdFrame).cmd.fm.r1 -text "counter mode" -value 8 \
		-variable ::ced::c_setgprmmd(mode) -command {::ced::cmd_SetGPRMMD_poke} -anchor w
	pack  $::ced::widget(cmdFrame).cmd.fm.r1 -side top -fill x -anchor w
	radiobutton $::ced::widget(cmdFrame).cmd.fm.r2 -text "register mode" -value 0 \
		-variable ::ced::c_setgprmmd(mode) -command {::ced::cmd_SetGPRMMD_poke} -anchor w
	pack  $::ced::widget(cmdFrame).cmd.fm.r2 -side top -fill x -anchor w
	label $::ced::widget(cmdFrame).cmd.lab3 -text "to "
	pack  $::ced::widget(cmdFrame).cmd.lab3 -side left

	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_SetGPRMMD_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) {} ::ced::c_set(value) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4

	c_link_create "And"
	pack $::ced::widget(cmdFrame).lnk -side top -fill x

	c_link_checked
	c_if_checked

	cmd_SetGPRMMD_poke

	adjust_win_size
}
proc configure_setgprmmd_sprm {text} {
	#dprint 3 "> cmd_SetGPRMMD:configure_setgprmmd_sprm {text}"
	set ::ced::c_set(sprm) [lindex $text 0]
	cmd_SetGPRMMD_poke
}

#############################################################################
## DVD command: SetAMXMD

proc cmd_SetAMXMD {} {
	#dprint 3 "> cmd_SetAMXMD {}"
	$::ced::widget(CommandName) configure -text "SetAMXMD"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "54 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "44 00 00 00 00 00 00 00"
	}

	create_commandframe
	pack  [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set Karaoke audio mixing mode (SPRM 11) to"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_SetAMXMD_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 0 0X1C1C
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4
	pack forget $::ced::widget(cmdFrame).cmd.val.val.e
	pack forget $::ced::widget(cmdFrame).cmd.val.val.l

	frame $::ced::widget(cmdFrame).cmd.val.val.f0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f0 -side left
	label $::ced::widget(cmdFrame).cmd.val.val.f0.l -text "Left  : channel 0" -anchor e
	pack  $::ced::widget(cmdFrame).cmd.val.val.f0.l -fill x -anchor e -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f0.c20 -text "+ channel 2" \
			-variable ::ced::c_setamxmd_mix(20) -onvalue [x2d 400] \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f0.c20 -side top -fill x -anchor e -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f0.c30 -text "+ channel 3" \
			-variable ::ced::c_setamxmd_mix(30) -onvalue [x2d 800] \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f0.c30 -side top -fill x -anchor e -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f0.c40 -text "+ channel 4" \
			-variable ::ced::c_setamxmd_mix(40) -onvalue [x2d 1000] \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f0.c40 -side top -fill x -anchor e -padx 0

	frame $::ced::widget(cmdFrame).cmd.val.val.f1
	pack  $::ced::widget(cmdFrame).cmd.val.val.f1 -side left
	label $::ced::widget(cmdFrame).cmd.val.val.f1.l -text "  Right : channel 1" -anchor e
	pack  $::ced::widget(cmdFrame).cmd.val.val.f1.l -fill x -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f1.c21 -text "+ channel 2" \
			-variable ::ced::c_setamxmd_mix(21) -onvalue 4 \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f1.c21 -side top -fill x -anchor e -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f1.c31 -text "+ channel 3" \
			-variable ::ced::c_setamxmd_mix(31) -onvalue 8 \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f1.c31 -side top -fill x -anchor e -padx 0
	checkbutton $::ced::widget(cmdFrame).cmd.val.val.f1.c41 -text "+ channel 4" \
			-variable ::ced::c_setamxmd_mix(41) -onvalue [x2d 10] \
			-command ::ced::cmd_SetAMXMD_poke -anchor e -padx 0
	pack  $::ced::widget(cmdFrame).cmd.val.val.f1.c41 -side top -fill x -anchor e -padx 0

	c_link_create "And"
	pack  $::ced::widget(cmdFrame).lnk -side top -fill x

	c_link_checked
	c_if_checked

	cmd_SetAMXMD_poke

	adjust_win_size
}
proc validate_SetAMXMD {args} {
	if {$::ced::c_set(sourcemode) == 1} {
		if {[winfo exists $::ced::widget(cmdFrame).cmd.val.val.f0.c20]} {
			$::ced::widget(cmdFrame).cmd.val.val.f0.c20 configure -state normal
			$::ced::widget(cmdFrame).cmd.val.val.f0.c30 configure -state normal
			$::ced::widget(cmdFrame).cmd.val.val.f0.c40 configure -state normal
			$::ced::widget(cmdFrame).cmd.val.val.f1.c21 configure -state normal
			$::ced::widget(cmdFrame).cmd.val.val.f1.c31 configure -state normal
			$::ced::widget(cmdFrame).cmd.val.val.f1.c41 configure -state normal
		}
		set ::ced::c_set(value) [expr \
				{$::ced::c_setamxmd_mix(20) | $::ced::c_setamxmd_mix(30) | \
				$::ced::c_setamxmd_mix(40) | $::ced::c_setamxmd_mix(21) | \
				$::ced::c_setamxmd_mix(31) | $::ced::c_setamxmd_mix(41)}]
	} else {
		if {[winfo exists $::ced::widget(cmdFrame).cmd.val.val.f0.c20]} {
			$::ced::widget(cmdFrame).cmd.val.val.f0.c20 configure -state disabled
			$::ced::widget(cmdFrame).cmd.val.val.f0.c30 configure -state disabled
			$::ced::widget(cmdFrame).cmd.val.val.f0.c40 configure -state disabled
			$::ced::widget(cmdFrame).cmd.val.val.f1.c21 configure -state disabled
			$::ced::widget(cmdFrame).cmd.val.val.f1.c31 configure -state disabled
			$::ced::widget(cmdFrame).cmd.val.val.f1.c41 configure -state disabled
		}
	}
}

#############################################################################
## DVD command: SetHL_BTN

proc cmd_SetHL_BTN {} {
	#dprint 3 "> cmd_SetHL_BTN {}"

	$::ced::widget(CommandName) configure -text "SetHL_BTN"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "56 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "46 00 00 00 00 00 00 00"
	}

	# there is no Highlight Button parameter for this command
	set ::ced::c_link(highlightmode) false

	create_commandframe
	pack  [c_if_create $::ced::widget(cmdFrame) 2] -side top -fill x

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set the Highlighted Button Number (SPRM 8) to"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_SetHL_BTN_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 1 36
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4
	$::ced::widget(cmdFrame).cmd.val.val.r configure -text "button"

	c_link_create "And"
	pack  $::ced::widget(cmdFrame).lnk -side top -fill x

	c_link_checked
	c_if_checked

	cmd_SetHL_BTN_poke

	adjust_win_size
}

#############################################################################
## DVD command: Set

proc cmd_Set {} {
	#dprint 3 "> cmd_Set {}"
	$::ced::widget(CommandName) configure -text "Set"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "70 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "60 00 00 00 00 00 00 00"
	}

	create_commandframe


	# IF
	set in $::ced::widget(cmdFrame)
	frame $in.if ;# -borderwidth 2 -relief groove
	pack  $in.if -side top -fill x
	checkbutton $in.if.ifcheck -command {::ced::c_if_checked; ::ced::cmd_Set_poke} \
			-text "If" -variable ::ced::c_if(check) -onvalue true -offvalue false
	pack  $in.if.ifcheck -side left
	label $in.if.gprmlabel -text {GPRM}
	pack  $in.if.gprmlabel -side left
	scale $in.if.gprm -from 0 -to 15 -variable ::ced::c_if(gprm1) -command ::ced::cmd_Set_poke
	pack  $in.if.gprm -side left

	menubutton $in.if.operator -textvariable ::ced::c_if(operator) -pady 1 -width 2 \
			-menu $in.if.operator.mnu
	pack  $in.if.operator -side left -padx 4
	menu  $in.if.operator.mnu -tearoff 0
	foreach i $::ced::c_compare_operators_list {
		$in.if.operator.mnu add command -label "[lindex $i 1]   [lindex $i 2]" \
			-command "set ::ced::c_if(operator) [lindex $i 2]; set ::ced::c_if(operatornum) [lindex $i 0]; ::ced::cmd_Set_poke"
	}
	c_regval_create $in.if.regval2 ::ced::cmd_Set_poke ::ced::c_if(radioOp2) \
			::ced::c_if(gprm2) ::ced::c_if(sprm2) ::ced::c_if(value2) 0 0XFFFF
	pack  $in.if.regval2 -side left
	label $in.if.then -text "then"
	pack  $in.if.then -side left

	frame $::ced::widget(cmdFrame).cmd
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set GPRM"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left -padx 2

	scale $::ced::widget(cmdFrame).cmd.d -from 0 -to 15 \
		-variable ::ced::c_set(destination) \
		-command {::ced::cmd_Set_poke}
	pack  $::ced::widget(cmdFrame).cmd.d -side left

	menubutton $::ced::widget(cmdFrame).cmd.m -textvariable ::ced::c_set(operator_text) \
			-pady 1 -width 9 -menu $::ced::widget(cmdFrame).cmd.m.mnu
	pack  $::ced::widget(cmdFrame).cmd.m -side left -padx 4
	menu  $::ced::widget(cmdFrame).cmd.m.mnu -tearoff 0
	foreach i $::ced::c_Setoperators {
		if {[lindex $i 0] == "-"} {
			$::ced::widget(cmdFrame).cmd.m.mnu add separator
		} else {
			$::ced::widget(cmdFrame).cmd.m.mnu add command -label "[lindex $i 1]" \
				-command "::ced::cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd \"$i\"; ::ced::cmd_Set_poke"
		}
	}
	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_Set_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4

	c_link_create "And"
	pack  $::ced::widget(cmdFrame).lnk -side top -fill x
	c_link_checked
	c_if_checked

	# validate the Set operand 2 upon state of the Set operator
	foreach i $::ced::c_Setoperators {
		if {$::ced::c_set(operator_value) == [lindex $i 0]} {
			set oplist $i
			break
		}
	}
	cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd $oplist

	cmd_Set_poke
	c_link_poke

	adjust_win_size
}

#############################################################################
## DVD command: SetCLnk

proc cmd_SetCLnk {} {
	#dprint 3 "> cmd_SetCLnk {}"

	$::ced::widget(CommandName) configure -text "SetCLnk"
	if {$::ced::c_set(sourcemode) == 1} {
		set ::hexcommand "90 00 00 00 00 00 00 00"
	} else {
		set ::hexcommand "80 00 00 00 00 00 00 00"
	}

	create_commandframe
	$::ced::widget(cmdFrame) configure -borderwidth 0 -relief flat -padx 0 -pady 0

	set ::ced::c_if(type) 1

	# set
	frame $::ced::widget(cmdFrame).cmd -borderwidth 2 -relief ridge -padx 4
	pack  $::ced::widget(cmdFrame).cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd.lab -text "Set GPRM"
	pack  $::ced::widget(cmdFrame).cmd.lab -side left
	scale $::ced::widget(cmdFrame).cmd.d -from 0 -to 15 \
		-variable ::ced::c_set(destination) \
		-command {::ced::cmd_SetCLnk_poke}
	pack  $::ced::widget(cmdFrame).cmd.d -side left
	menubutton $::ced::widget(cmdFrame).cmd.m -textvariable ::ced::c_set(operator_text) \
			-pady 1 -width 9 -menu $::ced::widget(cmdFrame).cmd.m.mnu
	pack  $::ced::widget(cmdFrame).cmd.m -side left -padx 4
	menu  $::ced::widget(cmdFrame).cmd.m.mnu -tearoff 0
	foreach i $::ced::c_Setoperators {
		if {[lindex $i 0] == "-"} {
			$::ced::widget(cmdFrame).cmd.m.mnu add separator
		} else {
			$::ced::widget(cmdFrame).cmd.m.mnu add command -label "[lindex $i 1]" \
				-command "::ced::cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd \"$i\"; ::ced::cmd_SetCLnk_poke"
		}
	}
	c_regval_create $::ced::widget(cmdFrame).cmd.val ::ced::cmd_SetCLnk_poke ::ced::c_set(sourcemode) \
			::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd.val -side left -pady 4

	# frame 2
	frame $::ced::widget(cmdFrame).cmd2 -borderwidth 2 -relief ridge -padx 4
	pack  $::ced::widget(cmdFrame).cmd2 -side top -fill x -pady 2

	# compare
	frame $::ced::widget(cmdFrame).cmd2.cmp
	pack  $::ced::widget(cmdFrame).cmd2.cmp -side top -fill x -pady 2
	label $::ced::widget(cmdFrame).cmd2.cmp.l1 -text "If  GPRM"
	pack  $::ced::widget(cmdFrame).cmd2.cmp.l1 -side left
	entry $::ced::widget(cmdFrame).cmd2.cmp.e -state readonly -width 2 \
			-textvariable ::ced::c_set(destination)
	pack  $::ced::widget(cmdFrame).cmd2.cmp.e -side left
	menubutton $::ced::widget(cmdFrame).cmd2.cmp.operator -textvariable ::ced::c_if(operator) \
			-pady 1 -width 2 -menu $::ced::widget(cmdFrame).cmd2.cmp.operator.mnu
	pack  $::ced::widget(cmdFrame).cmd2.cmp.operator -side left -padx 4
	menu  $::ced::widget(cmdFrame).cmd2.cmp.operator.mnu -tearoff 0
	foreach i $::ced::c_compare_operators_list {
		$::ced::widget(cmdFrame).cmd2.cmp.operator.mnu add command -label "[lindex $i 1]   [lindex $i 2]" \
				-command "set ::ced::c_if(operator) [lindex $i 2]; set ::ced::c_if(operatornum) [lindex $i 0]; ::ced::cmd_SetCLnk_poke"
	}
	c_regval_create $::ced::widget(cmdFrame).cmd2.cmp.regval2 ::ced::cmd_SetCLnk_poke ::ced::c_if(radioOp2) \
			::ced::c_if(gprm2) ::ced::c_if(sprm2) ::ced::c_if(value2) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd2.cmp.regval2 -side left
	label $::ced::widget(cmdFrame).cmd2.cmp.then -text "then"
	pack  $::ced::widget(cmdFrame).cmd2.cmp.then -side left -padx 2

	# link
	menubutton $::ced::widget(cmdFrame).cmd2.cmp.m -textvariable ::ced::c_link(text) \
			-pady 1 -menu $::ced::widget(cmdFrame).cmd2.cmp.m.mnu -width 11
	pack  $::ced::widget(cmdFrame).cmd2.cmp.m -side left -pady 2
	menu  $::ced::widget(cmdFrame).cmd2.cmp.m.mnu -tearoff 0
	foreach i $::ced::link_commands_list {
		if {[lindex $i 0] == "-"} {
			break
		} else {
			$::ced::widget(cmdFrame).cmd2.cmp.m.mnu add command -label "[lindex $i 2]" \
					-command "::ced::SetCLnk_configure_Link [lindex $i 1]"
		}
	}
	if {$::ced::c_link(k) > 1} {set ::ced::c_link(k) 0}
	if {$::ced::c_link(k) == 0} {set ::ced::c_link(ll) 0}
	SetCLnk_configure_Link $::ced::c_link(ll)

	# validate the Set operand 2 upon state of the Set operator
	foreach i $::ced::c_Setoperators {
		if {$::ced::c_set(operator_value) == [lindex $i 0]} {
			set oplist $i
			break
		}
	}
	cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd $oplist

	cmd_SetCLnk_poke

	adjust_win_size
}
proc SetCLnk_configure_Link {linkvalue} {
	#dprint 3 "> SetCLnk_configure_Link {linkvalue}"

	poke_hex_byte 6 0
	set ::ced::c_link(k)  1
	set ::ced::c_link(ll) $linkvalue
	foreach i $::ced::link_commands_list {
		if {$linkvalue == [lindex $i 1]} {
			set ::ced::c_link(text) [lindex $i 2]
			break
		}
	}
	set text [lindex $::ced::c_link_1_descriptions $linkvalue]
	poke_hex_byte 7 $linkvalue

	destroy $::ced::widget(cmdFrame).cmd2.lnk
	frame $::ced::widget(cmdFrame).cmd2.lnk
	pack  $::ced::widget(cmdFrame).cmd2.lnk -side top -fill x -pady 4
	label $::ced::widget(cmdFrame).cmd2.lnk.l -text $text -justify left -wraplength 190
	pack  $::ced::widget(cmdFrame).cmd2.lnk.l -side left
	frame $::ced::widget(cmdFrame).cmd2.lnk.pad
	pack  $::ced::widget(cmdFrame).cmd2.lnk.pad -side left -padx 2
	c_highlight_button_create  $::ced::widget(cmdFrame).cmd2.lnk.hb
	pack $::ced::widget(cmdFrame).cmd2.lnk.hb -side left

	adjust_win_size
}

#############################################################################
## DVD command: CSetCLnk

proc cmd_CSetCLnk {} {
	#dprint 3 "> cmd_CSetCLnk {}"
	cmd_CSetCLnk_and_CSetLnk "CSetCLnk"
}

#############################################################################
## DVD command: CSetLnk

proc cmd_CSetLnk {} {
	#dprint 3 "> cmd_CSetLnk {}"
	cmd_CSetCLnk_and_CSetLnk "CSetLnk"
}

#############################################################################
## DVD commands: CSetCLnk and CSetLnk
## Because CSetCLnk and CSetLnk use exactly the same widgets, theses commands
## are grouped together.

proc cmd_CSetCLnk_and_CSetLnk {cmdname} {
	#dprint 3 "> cmd_CSetCLnk_and_CSetLnk {cmdname}"

	$::ced::widget(CommandName) configure -text $cmdname
	if {$cmdname == "CSetCLnk"} {
		set ::ced::c_setclnk(cmdvalue) 10
		set linklabel "And"
	} else {
		set ::ced::c_setclnk(cmdvalue) 12
		set linklabel "Then, unconditionnaly,"
	}
	set ::hexcommand "B0 00 00 00 00 00 00 00"	;# byte 1 changed later to A0, B0, C0 or D0

	# create command frame
	create_commandframe
	if {$cmdname == "CSetLnk"} {
		$::ced::widget(cmdFrame) configure -borderwidth 0 -relief flat
	}

	set ::ced::c_if(type)   2
	if {$::ced::c_set(sourcemode) == 1 && $::ced::c_if(radioOp2) == 1} {
		set ::ced::c_if(radioOp2) 2
	}

	# frame 1
	if {$cmdname == "CSetCLnk"} {
		frame $::ced::widget(cmdFrame).cmd1
		pack  $::ced::widget(cmdFrame).cmd1 -side top -fill x
	} else {
		frame $::ced::widget(cmdFrame).cmd1 -borderwidth 2 -relief ridge -pady 4
		pack  $::ced::widget(cmdFrame).cmd1 -side top -fill x -pady 2 -expand 1
	}

	# compare
	frame $::ced::widget(cmdFrame).cmd1.cmp
	pack  $::ced::widget(cmdFrame).cmd1.cmp -side top -fill x

	label $::ced::widget(cmdFrame).cmd1.cmp.if -text " If  GPRM"
	pack  $::ced::widget(cmdFrame).cmd1.cmp.if -side left
	scale $::ced::widget(cmdFrame).cmd1.cmp.gprm -from 0 -to 15 \
			-variable ::ced::c_if(gprm1) -command ::ced::cmd_CSetCLnk_and_CSetLnk_poke
	pack  $::ced::widget(cmdFrame).cmd1.cmp.gprm -side left


	menubutton $::ced::widget(cmdFrame).cmd1.cmp.operator -textvariable ::ced::c_if(operator) \
			-pady 1 -width 2 -menu $::ced::widget(cmdFrame).cmd1.cmp.operator.mnu
	pack  $::ced::widget(cmdFrame).cmd1.cmp.operator -side left -padx 4
	menu  $::ced::widget(cmdFrame).cmd1.cmp.operator.mnu -tearoff 0
	foreach i $::ced::c_compare_operators_list {
		$::ced::widget(cmdFrame).cmd1.cmp.operator.mnu add command -label "[lindex $i 1]   [lindex $i 2]" \
			-command "set ::ced::c_if(operator) [lindex $i 2]; set ::ced::c_if(operatornum) [lindex $i 0]; ::ced::cmd_CSetCLnk_and_CSetLnk_poke"
	}
	c_regval_create $::ced::widget(cmdFrame).cmd1.cmp.regval2 "::ced::cmd_CSetCLnk_and_CSetLnk_poke" \
			::ced::c_if(radioOp2) ::ced::c_if(gprm2) ::ced::c_if(sprm2) ::ced::c_if(value2) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd1.cmp.regval2 -side left
	label $::ced::widget(cmdFrame).cmd1.cmp.then -text "then"
	pack  $::ced::widget(cmdFrame).cmd1.cmp.then -side left -padx 2

	# set
	frame $::ced::widget(cmdFrame).cmd1.cmd
	pack  $::ced::widget(cmdFrame).cmd1.cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd1.cmd.lab -text "Set GPRM"
	pack  $::ced::widget(cmdFrame).cmd1.cmd.lab -side left -padx 2
	scale $::ced::widget(cmdFrame).cmd1.cmd.d -from 0 -to 15 \
		-variable ::ced::c_set(destination) \
		-command "::ced::cmd_CSetCLnk_and_CSetLnk_poke"
	pack  $::ced::widget(cmdFrame).cmd1.cmd.d -side left
	menubutton $::ced::widget(cmdFrame).cmd1.cmd.m -textvariable ::ced::c_set(operator_text) \
			-pady 1 -width 9 -menu $::ced::widget(cmdFrame).cmd1.cmd.m.mnu
	pack  $::ced::widget(cmdFrame).cmd1.cmd.m -side left -padx 4
	menu  $::ced::widget(cmdFrame).cmd1.cmd.m.mnu -tearoff 0
	foreach i $::ced::c_Setoperators {
		if {[lindex $i 0] == "-"} {
			$::ced::widget(cmdFrame).cmd1.cmd.m.mnu add separator
		} else {
			$::ced::widget(cmdFrame).cmd1.cmd.m.mnu add command -label "[lindex $i 1]" \
				-command "::ced::cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd1.cmd \"$i\"; ::ced::cmd_CSetCLnk_and_CSetLnk_poke"
		}
	}
	c_regval_create $::ced::widget(cmdFrame).cmd1.cmd.val "::ced::cmd_CSetCLnk_and_CSetLnk_poke" \
			::ced::c_set(sourcemode) ::ced::c_set(gprm) ::ced::c_set(sprm) ::ced::c_set(value) 0 0XFFFF
	pack  $::ced::widget(cmdFrame).cmd1.cmd.val -side left -pady 4

	# frame 2
	if {$cmdname == "CSetCLnk"} {
		frame $::ced::widget(cmdFrame).cmd2
		pack  $::ced::widget(cmdFrame).cmd2 -side top -fill x
	} else {
		frame $::ced::widget(cmdFrame).cmd2 -borderwidth 2 -relief ridge -pady 4
		pack  $::ced::widget(cmdFrame).cmd2 -side top -fill x -pady 2 -expand 1
	}

	# link
	frame $::ced::widget(cmdFrame).cmd2.cmd
	pack  $::ced::widget(cmdFrame).cmd2.cmd -side top -fill x
	label $::ced::widget(cmdFrame).cmd2.cmd.lab -text $linklabel
	pack  $::ced::widget(cmdFrame).cmd2.cmd.lab -side left -padx 2
	menubutton $::ced::widget(cmdFrame).cmd2.cmd.m -textvariable ::ced::c_link(text) \
			-width 11 -pady 1 -menu $::ced::widget(cmdFrame).cmd2.cmd.m.mnu
	pack  $::ced::widget(cmdFrame).cmd2.cmd.m -side left -pady 2
	menu  $::ced::widget(cmdFrame).cmd2.cmd.m.mnu -tearoff 0
	foreach i $::ced::link_commands_list {
		if {[lindex $i 0] == "-"} {
			break
		} else {
			$::ced::widget(cmdFrame).cmd2.cmd.m.mnu add command -label "[lindex $i 2]" \
				-command "::ced::CSetCLnk_configure_Link $::ced::widget(cmdFrame).cmd2.lnk [lindex $i 0] [lindex $i 1]"
		}
	}
	if {$::ced::c_link(k) > 1} {set ::ced::c_link(k) 0}
	if {$::ced::c_link(k) == 0} {set ::ced::c_link(ll) 0}
	CSetCLnk_configure_Link $::ced::widget(cmdFrame).cmd2.lnk $::ced::c_link(k) $::ced::c_link(ll)

	# validate the Set operand 2 upon state of the Set operator
	foreach i $::ced::c_Setoperators {
		if {$::ced::c_set(operator_value) == [lindex $i 0]} {
			set oplist $i
			break
		}
	}
	cmds_Set_operator_validate $::ced::widget(cmdFrame).cmd1.cmd $oplist

	cmd_CSetCLnk_and_CSetLnk_poke

	adjust_win_size
}
proc CSetCLnk_configure_Link {widget k ll} {
	#dprint 3 "> CSetCLnk_configure_Link $widget $k $ll"
	set ::ced::c_link(k)  $k
	set ::ced::c_link(ll) $ll
	foreach i $::ced::link_commands_list {
		if {$k == [lindex $i 0] && $ll == [lindex $i 1]} {
			set ::ced::c_link(text) "[lindex $i 2]"
			break
		}
	}
	#poke_hex_nibble 1 1 $k	;# WARNING
	poke_hex_byte   7 $ll
	set text [lindex $::ced::c_link_1_descriptions $ll]
	destroy $widget
	frame $widget
	pack  $widget -side top -fill x -pady 4
	label $widget.l -text $text -justify left -wraplength 190
	pack  $widget.l -side left
	frame $widget.pad
	pack  $widget.pad -side left -padx 2
	c_highlight_button_create  $widget.hb
	pack $widget.hb -side left
	c_highlight_button_refresh $widget.hb

	adjust_win_size
}

#############################################################################
#############################################################################
## DVD COMMANDS VALIDATION AND POKE PROCEDURES
#############################################################################
#############################################################################
## Below are specific validation procs for some DVD commands and poke procs
## that poke the hexcommand according to the DVD command interface state.
#############################################################################
#############################################################################



#############################################################################
## poke for SetCLnk

proc cmd_SetCLnk_poke {args} {
	#dprint 3 "> cmd_SetCLnk_poke {args}"
	poke_hex_nibble 1 1 $::ced::c_set(destination)
	poke_hex_nibble 0 1 $::ced::c_set(operator_value)

	# if command
	# if command

	set add 0
	if {$::ced::c_if(radioOp2) == 1} {set add 8}
	poke_hex_nibble 1 0 [expr {$::ced::c_if(operatornum)+$add}]
	switch $::ced::c_if(radioOp2) {
		1 {
			# compare with value
			set val $::ced::c_if(value2)
			if {$val == ""} {set val 0}
			poke_hex_word 4 $val
		}
		2 {
			# compare with GPRM
			poke_hex_byte 4 0
			poke_hex_byte 5 $::ced::c_if(gprm2)
		}
		3 {
			# compare with SPRM
			poke_hex_byte 4 0
			poke_hex_byte 5 [expr {$::ced::c_if(sprm2) | 0X80}]
		}
	}
	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 9
			poke_hex_word 2 $::ced::c_set(value)
		}
		2 {
			poke_hex_nibble 0 0 8
			poke_hex_byte 2 0
			poke_hex_byte 3 $::ced::c_set(gprm)
		}
		3 {
			poke_hex_nibble 0 0 8
			poke_hex_byte 2 0
			poke_hex_byte 3 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}
}

#############################################################################
## poke for Set

proc cmd_Set_poke {args} {
	#dprint 3 "> cmd_Set_poke {args}"

	# if
	if {! $::ced::c_if(check)} {
		# no IF command: remove IF values in hex string
		poke_hex_nibble 1 0 0
		poke_hex_byte 2 0
		poke_hex_word 6 0
	} else {
		# if command
		set add 0
		if {$::ced::c_if(radioOp2) == 1} {set add 8}
		poke_hex_nibble 1 0 [expr {$::ced::c_if(operatornum)+$add}]

		# adjust hex string upon state of operand 1
		poke_hex_byte 2 $::ced::c_if(gprm1)

		# adjust hex string upon state of operand 2
		switch $::ced::c_if(radioOp2) {
			1 {
				# compare with value
				set val $::ced::c_if(value2)
				if {$val == ""} {set val 0}
				poke_hex_word 6 $val
			}
			2 {
				# compare with GPRM
				poke_hex_byte 6 0
				poke_hex_byte 7 $::ced::c_if(gprm2)
			}
			3 {
				# compare with SPRM
				poke_hex_byte 6 0
				poke_hex_byte 7 [expr {$::ced::c_if(sprm2) | 0X80}]
			}
		}
	}

	# Set
	poke_hex_byte 3 $::ced::c_set(destination)
	poke_hex_nibble 0 1 $::ced::c_set(operator_value)
	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 7
			poke_hex_word 4 $::ced::c_set(value)
		}
		2 {
			poke_hex_nibble 0 0 6
			poke_hex_byte 4 0
			poke_hex_byte 5 $::ced::c_set(gprm)
		}
		3 {
			poke_hex_nibble 0 0 6
			poke_hex_byte 4 0
			poke_hex_byte 5 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}

	# Link
	c_link_poke
}

#############################################################################
## validate for Set*

proc cmds_Set_operator_validate {widget option} {
	#dprint 3 "> cmds_Set_operator_validate {widget option}"
	set ::ced::c_set(operator_value)  [lindex $option 0]
	set ::ced::c_set(operator_text)   [lindex $option 1]
	set sprm_ok [lindex $option 2]
	set val_ok  [lindex $option 3]
	set w $widget.val

	if {$sprm_ok} {
		$w.sprm.r configure -state normal
	} else {
		if {$::ced::c_set(sourcemode)==3} {
			set ::ced::c_set(sourcemode) 1
			$w.val.e configure -state normal
			catch {$w.val.bb configure -state normal}
			$w.val.l configure -foreground black
		}
		$w.sprm.r configure -state disabled
		$w.sprm.m configure -state disabled
	}
	if {$val_ok} {
		$w.val.r configure -state normal
	} else {
		if {$::ced::c_set(sourcemode)==1} {
			set ::ced::c_set(sourcemode) 2
			scale_state $w.gprm.s normal
		}
		$w.val.r configure -state disabled
		$w.val.e configure -state disabled
		catch {$w.val.bb configure -state disabled}
		$w.val.l configure -foreground gray
	}
}

#############################################################################
## poke for SetHL_BTN

proc cmd_SetHL_BTN_poke {args} {
	#dprint 3 "> cmd_SetHL_BTN_poke {args}"

	c_if_poke

	switch $::ced::c_set(sourcemode) {
		1 {
			set val $::ced::c_set(value)
			if {$val == "" || $val == "0x" || $val == "0X" || $val == 0} {
				set val 1
			}
			poke_hex_nibble 0 0 5
			poke_hex_word 4 [expr {$val * 1024}]
		}
		2 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 4 0
			poke_hex_byte 5 $::ced::c_set(gprm)
		}
		3 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 4 0
			poke_hex_byte 5 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}
	c_link_poke
}

#############################################################################
## poke for SetAMXMD

proc cmd_SetAMXMD_poke {args} {
	#dprint 3 "> cmd_SetAMXMD_poke {args}"

	validate_SetAMXMD

	c_if_poke

	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 5
			poke_hex_word 4 $::ced::c_set(value)
		}
		2 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 4 0
			poke_hex_byte 5 $::ced::c_set(gprm)
		}
		3 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 4 0
			poke_hex_byte 5 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}
	c_link_poke
}

#############################################################################
## poke for SetGPRMMD

proc cmd_SetGPRMMD_poke {args} {
	#dprint 3 "> cmd_SetGPRMMD_poke {args}"

	c_if_poke

	poke_hex_nibble 5 0 $::ced::c_setgprmmd(mode)
	poke_hex_nibble 5 1 $::ced::c_setgprmmd(destination)

	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 5
			poke_hex_word 2 $::ced::c_set(value)
		}
		2 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 2 0
			poke_hex_byte 3 $::ced::c_set(gprm)
		}
		3 {
			bell
			#poke_hex_nibble 0 0 4
			#poke_hex_byte 2 0
			#poke_hex_byte 3 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}
	c_link_poke
}

#############################################################################
## poke and validate for SetNVTMR

proc cmd_SetNVTMR_validate_and_poke {args} {
	#dprint 3 "> cmd_SetNVTMR_validate_and_poke {args}"

	c_if_poke

	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 5
			poke_hex_word 2 $::ced::c_set(value)
		}
		2 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 2 0
			poke_hex_byte 3 $::ced::c_set(gprm)
		}
		3 {
			poke_hex_nibble 0 0 4
			poke_hex_byte 2 0
			poke_hex_byte 3 [expr {$::ced::c_set(sprm) + 0X80}]
		}
	}
	poke_hex_word 4 $::ced::c_setnvtmr(pgc)
	c_link_poke
}

#############################################################################
## poke and validate for CSetCLnk and CSetLnk

proc cmd_CSetCLnk_and_CSetLnk_poke {args} {
	#dprint 3 "> cmd_CSetCLnk_and_CSetLnk_poke {args}"

	# set
	poke_hex_nibble 1 1 $::ced::c_set(destination)
	poke_hex_nibble 0 1 $::ced::c_set(operator_value)
	switch $::ced::c_set(sourcemode) {
		1 {
			poke_hex_nibble 0 0 [expr {$::ced::c_setclnk(cmdvalue) + 1}]
			poke_hex_word 2 $::ced::c_set(value)
			catch {
				$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.r configure -state disabled
				$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.e configure -state disabled
				catch {$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.bb configure -state disabled}
				$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.l configure -foreground gray
			}
			set ifop1pos 4
		}
		2 {
			poke_hex_nibble 0 0 $::ced::c_setclnk(cmdvalue)
			poke_hex_byte 2 $::ced::c_set(gprm)
			catch {$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.r configure -state normal}
			set ifop1pos 3
			poke_hex_byte 4 0
		}
		3 {
			poke_hex_nibble 0 0 $::ced::c_setclnk(cmdvalue)
			poke_hex_byte 2 [expr {$::ced::c_set(sprm) + 0X80}]
			catch {$::ced::widget(cmdFrame).cmd1.cmp.regval2.val.r configure -state normal}
			set ifop1pos 3
			poke_hex_byte 4 0
		}
	}

	# compare
	if {$::ced::c_if(radioOp2) == 1} {set add 0X8} else {set add 0}
	poke_hex_nibble 1 0 [expr {$::ced::c_if(operatornum) + $add}]
	poke_hex_byte $ifop1pos $::ced::c_if(gprm1)
	switch $::ced::c_if(radioOp2) {
		1 {
			poke_hex_word 4 $::ced::c_if(value2)
			catch {
				$::ced::widget(cmdFrame).cmd1.cmd.val.val.r configure -state disabled
				$::ced::widget(cmdFrame).cmd1.cmd.val.val.e configure -state disabled
				catch {$::ced::widget(cmdFrame).cmd1.cmd.val.val.bb configure -state disabled}
				$::ced::widget(cmdFrame).cmd1.cmd.val.val.l configure -foreground gray
			}
		}
		2 {
			poke_hex_byte 5 $::ced::c_if(gprm2)
			# allow value operand for Set except if operation is SWAP
			if {$::ced::c_set(operator_value) != 2} {
				catch {$::ced::widget(cmdFrame).cmd1.cmd.val.val.r configure -state normal}
			}
		}
		3 {
			poke_hex_byte 5 [expr {$::ced::c_if(sprm2) + 0X80}]
			# allow value operand for Set except if operation is SWAP
			if {$::ced::c_set(operator_value) != 2} {
				catch {$::ced::widget(cmdFrame).cmd1.cmd.val.val.r configure -state normal}
			}
		}
	}
	# link
	poke_hex_byte 7 $::ced::c_link(ll)
}

#############################################################################
## return from the command editor

proc cmdedit_ok {} {
	set ::ced_return 1
	wm withdraw .cmdeditor
	grab release .cmdeditor
}

proc cmdedit_cancel {} {
	set ::ced_return 0
	wm withdraw .cmdeditor
	grab release .cmdeditor
}

proc cmdedit_close {} {
	if {"%W" == ".cmdeditor"} {
		set ::ced_return 0
		wm withdraw .cmdeditor
		grab release .cmdeditor
	}
}

#############################################################################
#############################################################################
## INITIALIZATION PROCEDURES
#############################################################################
#############################################################################


#############################################################################
## Create command editor window, with top rows widgets (commands menus, hex entry).
## Also, initialize most of the widget aliases in the array ::ced::widget().

proc cmdeditor {} {
	#dprint 3 "> cmdeditor {}"

	if {[winfo exists .cmdeditor]} {
		hexcmd_to_gui
		wm deiconify .cmdeditor
		grab set .cmdeditor
		focus .cmdeditor
		return
	}

	toplevel .cmdeditor -class Dialog -menu .cmdeditor.menu
	#bind .cmdeditor	<Destroy> ::ced::cmdedit_close
	wm protocol .cmdeditor WM_DELETE_WINDOW ::ced::cmdedit_close
	bind .cmdeditor	<Escape>  {set ::ced_return 0; wm withdraw .cmdeditor; grab release .cmdeditor}
	wm withdraw .cmdeditor
	#wm geometry .cmdeditor 490x414	;# set at end of proc
	update
	wm resizable .cmdeditor 0 1
	wm title .cmdeditor "PgcEdit - Command Editor"
	wm transient .cmdeditor .
	wm protocol .cmdeditor WM_DELETE_WINDOW {::ced::cmdedit_cancel}
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes .cmdeditor -toolwindow 1}}


	############# menu

	menu .cmdeditor.menu
	set f1 .cmdeditor.menu

	$f1 add cascade -label " Alphabetically " -menu $f1.cm1
	$f1 add cascade -label " Numerically " -menu $f1.cm3
	$f1 add cascade -label " By Type " -menu $f1.cm2
	$f1 add cascade -label " Jump and Call " -menu $f1.cm4

	menu  $f1.cm1
	set cmdnum 0
	foreach label $::ced::c_all_commands_abc {
		incr cmdnum
		if {$cmdnum == 17} {
			set break 1
			set cmdnum 1
		} else {
			set break 0
		}
		$f1.cm1 add command -label $label -command "::ced::cmd_$label" \
				-columnbreak $break
	}

	menu  $f1.cm2
	$f1.cm2 add cascade -label "Flow control" -menu $f1.cm2.f
	menu $f1.cm2.f
	$f1.cm2.f add command -label "No Operation (NOP)" -command "::ced::cmd_NOP"
	$f1.cm2.f add command -label "GoTo line (GoTo)" -command "::ced::cmd_GoTo"
	$f1.cm2.f add command -label "Break (Break)" -command "::ced::cmd_Break"
	$f1.cm2.f add command -label "Parental Management (SetTmpPML)" -command "::ced::cmd_SetTmpPML"
	$f1.cm2.f add command -label "Exit (Exit)" -command "::ced::cmd_Exit"
	$f1.cm2 add cascade -label "Link to" -menu $f1.cm2.l
	menu $f1.cm2.l
	$f1.cm2.l add cascade -label "PGC" -menu $f1.cm2.l.pc
	menu $f1.cm2.l.pc
	$f1.cm2.l.pc add command -label "number (LinkPGCN)" -command "::ced::cmd_LinkPGCN"
	$f1.cm2.l.pc add command -label "Top (LinkTopPGC)" -command "::ced::cmd_LinkTopPGC"
	$f1.cm2.l.pc add command -label "Tail (LinkTailPGC)" -command "::ced::cmd_LinkTailPGC"
	$f1.cm2.l.pc add command -label "Next (LinkNextPGC)" -command "::ced::cmd_LinkNextPGC"
	$f1.cm2.l.pc add command -label "Previous (LinkPrevPGC)" -command "::ced::cmd_LinkPrevPGC"
	$f1.cm2.l.pc add command -label "Go Up (LinkGoUpPGC)" -command "::ced::cmd_LinkGoUpPGC"
	$f1.cm2.l add cascade -label "Program" -menu $f1.cm2.l.pg
	menu $f1.cm2.l.pg
	$f1.cm2.l.pg add command -label "number (LinkPGN)" -command "::ced::cmd_LinkPGN"
	$f1.cm2.l.pg add command -label "Top (LinkTopPG)" -command "::ced::cmd_LinkTopPG"
	$f1.cm2.l.pg add command -label "Next (LinkNextPG)" -command "::ced::cmd_LinkNextPG"
	$f1.cm2.l.pg add command -label "Previous (LinkPrevPG)" -command "::ced::cmd_LinkPrevPG"
	$f1.cm2.l add cascade -label "Cell" -menu $f1.cm2.l.cl
	menu $f1.cm2.l.cl
	$f1.cm2.l.cl add command -label "number (LinkCN)" -command "::ced::cmd_LinkCN"
	$f1.cm2.l.cl add command -label "Top (LinkTopCell)" -command "::ced::cmd_LinkTopCell"
	$f1.cm2.l.cl add command -label "Next (LinkNextCell)" -command "::ced::cmd_LinkNextCell"
	$f1.cm2.l.cl add command -label "Previous (LinkPrevCell)" -command "::ced::cmd_LinkPrevCell"
	$f1.cm2.l add command -label "Chapter (LinkPTT)" -command "::ced::cmd_LinkPTT"
	$f1.cm2.l add separator
	$f1.cm2.l add command -label "Highlight Button (LinkNoLink)" -command "::ced::cmd_LinkNoLink"
	$f1.cm2.l add command -label "Resume (RSM)" -command "::ced::cmd_RSM"
	$f1.cm2 add cascade -label "Jump to" -menu $f1.cm2.j
	$f1.cm2.l add separator
	menu $f1.cm2.j
	$f1.cm2.j add command -label "VMG First Play PGC (JumpSS-FP)" -command "::ced::cmd_JumpSS-FP"
	$f1.cm2.j add command -label "VMGM Title Menu (JumpSS-VMGM-menu)" -command "::ced::cmd_JumpSS-VMGM-menu"
	$f1.cm2.j add command -label "VMGM PGC (JumpSS-VMGM-pgc)" -command "::ced::cmd_JumpSS-VMGM-pgc"
	$f1.cm2.j add command -label "VTSM Menu in this VTS (JumpSS-VTSM)" -command "::ced::cmd_JumpSS-VTSM"
	$f1.cm2.j add command -label "VTST Title (JumpTT)" -command "::ced::cmd_JumpTT"
	$f1.cm2.j add command -label "VTST TTN in this VTS (JumpVTS_TT)" -command "::ced::cmd_JumpVTS_TT"
	$f1.cm2 add cascade -label "Call" -menu $f1.cm2.c
	menu $f1.cm2.c
	$f1.cm2.c add command -label "VMG First Play PGC (CallSS-FP)" -command "::ced::cmd_CallSS-FP"
	$f1.cm2.c add command -label "VMGM Title Menu (CallSS-VMGM-menu)" -command "::ced::cmd_CallSS-VMGM-menu"
	$f1.cm2.c add command -label "VMGM PGC (CallSS-VMGM-pgc)" -command "::ced::cmd_CallSS-VMGM-pgc"
	$f1.cm2.c add command -label "VTSM Menu of this title (CallSS-VTSM)" -command "::ced::cmd_CallSS-VTSM"
	$f1.cm2 add cascade -label "Set" -menu $f1.cm2.s
	menu $f1.cm2.s
	$f1.cm2.s add cascade -label "GPRM" -menu $f1.cm2.s.g
	menu $f1.cm2.s.g
	$f1.cm2.s.g add command -label "Set GPRM (Set)" -command "::ced::cmd_Set"
	$f1.cm2.s.g add command -label "Set and conditionally Link (SetCLnk)" -command "::ced::cmd_SetCLnk"
	$f1.cm2.s.g add command -label "Conditionally Set and Link (CSetCLnk)" -command "::ced::cmd_CSetCLnk"
	$f1.cm2.s.g add command -label "Conditionally Set, Unconditionally Link (CSetLnk)" -command "::ced::cmd_CSetLnk"
	$f1.cm2.s.g add command -label "Counter (SetGPRMMD)" -command "::ced::cmd_SetGPRMMD"
	$f1.cm2.s add cascade -label "SPRM" -menu $f1.cm2.s.s
	menu $f1.cm2.s.s
	$f1.cm2.s.s add command -label "Audio: SPRM 1 (SetSTN)" -command "::ced::cmd_SetSTN"
	$f1.cm2.s.s add command -label "Sub-picture: SPRM 2 (SetSTN)" -command "::ced::cmd_SetSTN"
	$f1.cm2.s.s add command -label "Angle: SPRM 3 (SetSTN)" -command "::ced::cmd_SetSTN"
	$f1.cm2.s.s add command -label "Highlighted button: SPRM 8 (SetHL_BTN)" -command "::ced::cmd_SetHL_BTN"
	$f1.cm2.s.s add command -label "Navigation timer: SPRMs 9,10 (SetNVTMR)" -command "::ced::cmd_SetNVTMR"
	$f1.cm2.s.s add command -label "Karaoke audio mix: SPRM 11 (SetAMXMD)" -command "::ced::cmd_SetAMXMD"
	$f1.cm2.s.s add command -label "Parental level: SPRMs 12,13 (SetTmpPML)" -command "::ced::cmd_SetTmpPML"

	menu  $f1.cm3
	set n 0
	foreach i $::ced::c_all_commands_num {
		if {[lindex $i 0] == "-"} {
			incr n
			$f1.cm3 add cascade -label [lindex $i 1] \
				-menu "$f1.cm3.c$n"
			menu  $f1.cm3.c$n
		} else {
			set cmdstart [expr {[x2d [string range [lindex $i 0] 0 0]]}]
			if {[expr {$cmdstart % 2}]} {
				$f1.cm3.c$n add command -label "$i" -command "set ::ced::c_set(sourcemode) 1; ::ced::cmd_[lindex $i 8]"
			} else {
				$f1.cm3.c$n add command -label "$i" -command "if {\$::ced::c_set(sourcemode)==1} {set ::ced::c_set(sourcemode) 2}; ::ced::cmd_[lindex $i 8]"
			}
		}
	}

	menu  $f1.cm4
	$f1.cm4 add cascade -label "From FP-PGC" -menu $f1.cm4.fp
	menu $f1.cm4.fp
	$f1.cm4.fp add cascade -label "to First-Play PGC" -menu $f1.cm4.fp.t0
	menu $f1.cm4.fp.t0
	$f1.cm4.fp.t0 add command -label "JumpSS-FP" -command "::ced::cmd_JumpSS-FP"
	$f1.cm4.fp add cascade -label "to VMGM Menu" -menu $f1.cm4.fp.t1
	menu $f1.cm4.fp.t1
	$f1.cm4.fp.t1 add command -label "JumpSS-VMGM-menu" -command "::ced::cmd_JumpSS-VMGM-menu"
	$f1.cm4.fp.t1 add command -label "JumpSS-VMGM-pgc" -command "::ced::cmd_JumpSS-VMGM-pgc"
	$f1.cm4.fp add cascade -label "to VTSM Menu" -menu $f1.cm4.fp.t2
	menu $f1.cm4.fp.t2
	$f1.cm4.fp.t2 add command -label "JumpSS-VTSM" -command "::ced::cmd_JumpSS-VTSM"
	$f1.cm4.fp add cascade -label "to VTST Title" -menu $f1.cm4.fp.t3
	menu $f1.cm4.fp.t3
	$f1.cm4.fp.t3 add command -label "JumpTT" -command "::ced::cmd_JumpTT"

	$f1.cm4 add cascade -label "From VMGM menu" -menu $f1.cm4.vmgm
	menu $f1.cm4.vmgm
	$f1.cm4.vmgm add cascade -label "to First-Play PGC" -menu $f1.cm4.vmgm.t1
	menu $f1.cm4.vmgm.t1
	$f1.cm4.vmgm.t1 add command -label "JumpSS-FP" -command "::ced::cmd_JumpSS-FP"
	$f1.cm4.vmgm add cascade -label "to VMGM menu" -menu $f1.cm4.vmgm.t4
	menu $f1.cm4.vmgm.t4
	$f1.cm4.vmgm.t4 add command -label "JumpSS-VMGM-menu" -command "::ced::cmd_JumpSS-VMGM-menu"
	$f1.cm4.vmgm.t4 add command -label "JumpSS-VMGM-pgc" -command "::ced::cmd_JumpSS-VMGM-pgc"
	$f1.cm4.vmgm.t4 add command -label "LinkPGCN" -command "::ced::cmd_LinkPGCN"
	$f1.cm4.vmgm.t4 add command -label "LinkNextPGC" -command "::ced::cmd_LinkNextPGC"
	$f1.cm4.vmgm.t4 add command -label "LinkPrevPGC" -command "::ced::cmd_LinkPrevPGC"
	$f1.cm4.vmgm.t4 add command -label "LinkGoUpPGC" -command "::ced::cmd_LinkGoUpPGC"
	$f1.cm4.vmgm add cascade -label "to VTSM Menu" -menu $f1.cm4.vmgm.t2
	menu $f1.cm4.vmgm.t2
	$f1.cm4.vmgm.t2 add command -label "JumpSS-VTSM" -command "::ced::cmd_JumpSS-VTSM"
	$f1.cm4.vmgm add cascade -label "to VTST Title" -menu $f1.cm4.vmgm.t3
	menu $f1.cm4.vmgm.t3
	$f1.cm4.vmgm.t3 add command -label "JumpTT" -command "::ced::cmd_JumpTT"
	$f1.cm4.vmgm.t3 add command -label "RSM" -command "::ced::cmd_RSM"

	$f1.cm4 add cascade -label "From VTSM menu" -menu $f1.cm4.vtsm
	menu $f1.cm4.vtsm
	$f1.cm4.vtsm add cascade -label "to First-Play PGC" -menu $f1.cm4.vtsm.t1
	menu $f1.cm4.vtsm.t1
	$f1.cm4.vtsm.t1 add command -label "JumpSS-FP" -command "::ced::cmd_JumpSS-FP"
	$f1.cm4.vtsm add cascade -label "to VMGM Menu" -menu $f1.cm4.vtsm.t2
	menu $f1.cm4.vtsm.t2
	$f1.cm4.vtsm.t2 add command -label "JumpSS-VMGM-menu" -command "::ced::cmd_JumpSS-VMGM-menu"
	$f1.cm4.vtsm.t2 add command -label "JumpSS-VMGM-pgc" -command "::ced::cmd_JumpSS-VMGM-pgc"
	$f1.cm4.vtsm add cascade -label "to VTSM Menu in same VTS" -menu $f1.cm4.vtsm.t4
	menu $f1.cm4.vtsm.t4
	$f1.cm4.vtsm.t4 add command -label "JumpSS-VTSM (with VTS=0!)" \
			-command "set ::ced::c_jump(vts) 0; ::ced::cmd_JumpSS-VTSM"
	$f1.cm4.vtsm.t4 add command -label "LinkPGCN" -command "::ced::cmd_LinkPGCN"
	$f1.cm4.vtsm.t4 add command -label "LinkNextPGC" -command "::ced::cmd_LinkNextPGC"
	$f1.cm4.vtsm.t4 add command -label "LinkPrevPGC" -command "::ced::cmd_LinkPrevPGC"
	$f1.cm4.vtsm.t4 add command -label "LinkGoUpPGC" -command "::ced::cmd_LinkGoUpPGC"
	$f1.cm4.vtsm add cascade -label "to VTST Title in same VTS" -menu $f1.cm4.vtsm.t3
	menu $f1.cm4.vtsm.t3
	$f1.cm4.vtsm.t3 add command -label "JumpVTS_TT" -command "::ced::cmd_JumpVTS_TT"
	$f1.cm4.vtsm.t3 add command -label "JumpVTS_PTT" -command "::ced::cmd_JumpVTS_PTT"
	$f1.cm4.vtsm.t3 add command -label "RSM" -command "::ced::cmd_RSM"

	$f1.cm4 add cascade -label "From VTST title" -menu $f1.cm4.vtst
	menu $f1.cm4.vtst
	$f1.cm4.vtst add cascade -label "to First-Play PGC" -menu $f1.cm4.vtst.t1
	menu $f1.cm4.vtst.t1
	$f1.cm4.vtst.t1 add command -label "CallSS-FP" -command "::ced::cmd_CallSS-FP"
	$f1.cm4.vtst add cascade -label "to VMGM Menu" -menu $f1.cm4.vtst.t2
	menu $f1.cm4.vtst.t2
	$f1.cm4.vtst.t2 add command -label "CallSS-VMGM-menu" -command "::ced::cmd_CallSS-VMGM-menu"
	$f1.cm4.vtst.t2 add command -label "CallSS-VMGM-pgc" -command "::ced::cmd_CallSS-VMGM-pgc"
	$f1.cm4.vtst add cascade -label "to VTSM Menu in same VTS" -menu $f1.cm4.vtst.t3
	menu $f1.cm4.vtst.t3
	$f1.cm4.vtst.t3 add command -label "CallSS-VTSM" -command "::ced::cmd_CallSS-VTSM"
	$f1.cm4.vtst add cascade -label "to VTST Title in same VTS" -menu $f1.cm4.vtst.t4
	menu $f1.cm4.vtst.t4
	$f1.cm4.vtst.t4 add command -label "JumpVTS_TT" -command "::ced::cmd_JumpVTS_TT"
	$f1.cm4.vtst.t4 add command -label "JumpVTS_PTT" -command "::ced::cmd_JumpVTS_PTT"
	#$f1.cm4.vtst.t4 add command -label "LinkPGCN" -command "::ced::cmd_LinkPGCN"
	#$f1.cm4.vtst.t4 add command -label "LinkNextPGC" -command "::ced::cmd_LinkNextPGC"
	#$f1.cm4.vtst.t4 add command -label "LinkPrevPGC" -command "::ced::cmd_LinkPrevPGC"
	#$f1.cm4.vtst.t4 add command -label "LinkGoUpPGC" -command "::ced::cmd_LinkGoUpPGC"

	$f1.cm4.vtst add cascade -label "to same VTST Title" -menu $f1.cm4.vtst.t5
	menu $f1.cm4.vtst.t5
	$f1.cm4.vtst.t5 add command -label "LinkPTT" -command "::ced::cmd_LinkPTT"


	frame .cmdeditor.mainframe

	############# frame 2: command label and editable hex string

	set f2 .cmdeditor.mainframe.frame2
	frame $f2
	pack  $f2 -side top -pady 2

	frame $f2.hexcmd
	pack  $f2.hexcmd -pady 4 -side left

	c_hex_entry $f2.hexcmd.hex {::ced::c_hex_entry_poke; ::ced::hexcmd_to_gui}
	pack  $f2.hexcmd.hex -side left
	set ::ced::widget(hex_entry) $f2.hexcmd.hex

	menubutton $f2.hexcmd.m -menu $f2.hexcmd.m.menu -padx 0 -pady 0
	pack  $f2.hexcmd.m -side left -fill y
	bind $f2.hexcmd.m <Button-3> "::ced::c_hex_entry_poke; ::ced::hexcmd_to_gui"
	bind $f2.hexcmd.m <Button-2> "::ced::c_hex_entry_highlight_clear"
	menu  $f2.hexcmd.m.menu -tearoff 0
	$f2.hexcmd.m.menu add command -label "Copy" -accelerator Ctrl-C \
		-command {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) false]}
	$f2.hexcmd.m.menu add command -label "Copy as hex string" -accelerator Ctrl-Shift-C \
		-command {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_ascii $::ced::widget(hex_entry) true]}
	$f2.hexcmd.m.menu add command -label "Copy as binary" -accelerator Alt-C \
		-command {clipboard clear; clipboard append -- [::ced::c_hex_entry_get_binary $::ced::widget(hex_entry)]}
	$f2.hexcmd.m.menu add command -label "Paste" -accelerator Ctrl-V \
		-command {::ced::c_hex_entry_paste $::ced::widget(hex_entry)}
	$f2.hexcmd.m.menu add separator
	$f2.hexcmd.m.menu add command -label "Validate" -accelerator Enter \
		-command {::ced::c_hex_entry_poke; ::ced::hexcmd_to_gui}
	$f2.hexcmd.m.menu add command -label "Undo validation" -accelerator Ctrl-Z \
		-command "::ced::c_hex_entry_undo_validation $f2.hexcmd.hex"
	$f2.hexcmd.m.menu add command -label "Clear validation highlights" \
		-command {::ced::c_hex_entry_highlight_clear}
	$f2.hexcmd.m.menu add command -label "Disassemble without validate" -accelerator Ctrl-D \
		-command {::ced::hexentry_disassemble}

	frame $f2.pad -width 20
	pack  $f2.pad -side left

	if {$::config(web_browser) != ""} {
		button $f2.commandname -font $::config(fixed_width_font) -width 17 -pady 0 -bd 1 \
				-highlightthickness 1 -command {::ced::launch_web_info}
	} else {
		label $f2.commandname -font $::config(fixed_width_font) -width 17 -bd 2 -relief groove
	}
	set ::ced::widget(CommandName) $f2.commandname
	pack  $f2.commandname -fill x -side left

	############# frame 3: disassembled command

	set f3 .cmdeditor.mainframe.frame3
	frame $f3 -borderwidth 2 -relief sunken
	pack  $f3 -side top -padx 4 -pady 2 -fill x
	label $f3.disasm -height 2 -pady 4 -justify left ;#-font $::config(fixed_width_font)
	set ::ced::widget(disasm) $f3.disasm
	pack  $f3.disasm -fill x -expand 1
	bind  $f3.disasm <Button-1> ::ced::hexcmd_disasm
	bind  $f3.disasm <Button-3> ::ced::hexentry_disassemble

	############# frame 4: command interface

	frame .cmdeditor.mainframe.cmdframe
	pack  .cmdeditor.mainframe.cmdframe -side top -padx 2 -pady 4 -fill x -expand 1
	set ::ced::widget(cmdFrame) .cmdeditor.mainframe.cmdframe
	pack .cmdeditor.mainframe -side top -fill both -expand 1

	############# frame 5: OK / Cancel

	frame .cmdeditor.mainframe.okcancel
	pack  .cmdeditor.mainframe.okcancel -side bottom -fill x -pady 4
	frame .cmdeditor.mainframe.okcancel.pad
	pack  .cmdeditor.mainframe.okcancel.pad -side left -fill x -expand 1
	button .cmdeditor.mainframe.okcancel.ok -text OK -width 20 \
			-command ::ced::cmdedit_ok
	pack  .cmdeditor.mainframe.okcancel.ok -side left -padx 4
	button .cmdeditor.mainframe.okcancel.cancel -text Cancel -width 20 \
			-command ::ced::cmdedit_cancel
	pack  .cmdeditor.mainframe.okcancel.cancel -side left -padx 4

	############# configure geometry

	# determine la taille de la window
	set err [catch {set old $::hexcommand}]
	set ::hexcommand "51 07 00 00 00 00 04 01"
	hexcmd_to_gui
	update
	set width  [winfo reqwidth  .cmdeditor]
	set height [winfo reqheight .cmdeditor]
	set ::hexcommand "C1 20 00 00 00 00 00 00"
	hexcmd_to_gui
	update
	if {$width < [winfo reqwidth .cmdeditor]} {
		set width  [winfo reqwidth .cmdeditor]
	}
	if {$height < [winfo reqheight .cmdeditor]} {
		set height [winfo reqheight .cmdeditor]
	}
	set ::hexcommand "53 00 00 00 00 00 00 00"
	hexcmd_to_gui
	update
	if {$width < [winfo reqwidth .cmdeditor]} {
		set width  [winfo reqwidth .cmdeditor]
	}
	if {$height < [winfo reqheight .cmdeditor]} {
		set height [winfo reqheight .cmdeditor]
	}
	if {$::tcl_platform(platform)!="windows"} {incr height 100}
	if {[catch {wm geometry .cmdeditor $::config(ced_win_geometry)}]} {
		wm geometry .cmdeditor ""
	}
	wm minsize  .cmdeditor $width $height
	update
	wm maxsize  .cmdeditor $width $height
	wm minsize  .cmdeditor $width 1
	set ::ced::c_link(check) false
	set ::ced::c_link(highlightmode) false
	if {$err} {
		unset ::hexcommand
	} else {
		set ::hexcommand $old
		hexcmd_to_gui
	}
	# set the wrap length for disasm label
	$::ced::widget(disasm) configure -wraplength [expr {$width - 20}]

	# Establish a trace on the hexcommand to automatically disassemble it
	# when changed.
	#
	# Cree la hexcommand si elle n'existe pas encore
	if {[catch {set dummy $::hexcommand}] > 0} {
		set ::hexcommand "00 00 00 00 00 00 00 00"
	}
	# set the trace
	trace add variable ::hexcommand write ::ced::hexcmd_disasm
}

#############################################################################
## Init global lists used by the interface (mostly in menubuttons)

proc init_cmds_global_lists {} {
	#dprint 3 "> init_cmds_global_lists {}"
	set ::ced::c_all_commands_num {
		{- {00: Flow control}}
		{00 _0 __ __ __ __ __ __  NOP}
		{00 _1 __ __ __ __ __ __  GoTo}
		{00 _2 __ __ __ __ __ __  Break}
		{00 _3 __ __ __ __ __ __  SetTmpPML}
		{- {20: Link}}
		{20 _1 __ __ __ __ __ 00  LinkNoLink}
		{20 _1 __ __ __ __ __ 01  LinkTopCell}
		{20 _1 __ __ __ __ __ 02  LinkNextCell}
		{20 _1 __ __ __ __ __ 03  LinkPrevCell}
		{20 _1 __ __ __ __ __ 05  LinkTopPG}
		{20 _1 __ __ __ __ __ 06  LinkNextPG}
		{20 _1 __ __ __ __ __ 07  LinkPrevPG}
		{20 _1 __ __ __ __ __ 09  LinkTopPGC}
		{20 _1 __ __ __ __ __ 0A  LinkNextPGC}
		{20 _1 __ __ __ __ __ 0B  LinkPrevPGC}
		{20 _1 __ __ __ __ __ 0C  LinkGoUpPGC}
		{20 _1 __ __ __ __ __ 0D  LinkTailPGC}
		{20 _1 __ __ __ __ __ 10  RSM}
		{20 _4 __ __ __ __ __ __  LinkPGCN}
		{20 _5 __ __ __ __ __ __  LinkPTT}
		{20 _6 __ __ __ __ __ __  LinkPGN}
		{20 _7 __ __ __ __ __ __  LinkCN}
		{- {30: Jump and Call}}
		{30 _1 __ __ __ __ __ __  Exit}
		{30 _2 __ __ __ __ __ __  JumpTT}
		{30 _3 __ __ __ __ __ __  JumpVTS_TT}
		{30 _5 __ __ __ __ __ __  JumpVTS_PTT}
		{30 _6 __ __ __ 0_ __ __  JumpSS-FP}
		{30 _6 __ __ __ 4_ __ __  JumpSS-VMGM-menu}
		{30 _6 __ __ __ 8_ __ __  JumpSS-VTSM}
		{30 _6 __ __ __ C_ __ __  JumpSS-VMGM-pgc}
		{30 _8 __ __ __ 0_ __ __  CallSS-FP}
		{30 _8 __ __ __ 4_ __ __  CallSS-VMGM-menu}
		{30 _8 __ __ __ 8_ __ __  CallSS-VTSM}
		{30 _8 __ __ __ C_ __ __  CallSS-VMGM-pgc}
		{- {40-D0: Set}}
		{41 __ __ __ __ __ __ __  SetSTN}
		{42 __ __ __ __ __ __ __  SetNVTMR}
		{43 __ __ __ __ __ __ __  SetGPRMMD}
		{44 __ __ __ __ __ __ __  SetAMXMD}
		{46 __ __ __ __ __ __ __  SetHL_BTN}
		{51 __ __ __ __ __ __ __  SetSTN}
		{52 __ __ __ __ __ __ __  SetNVTMR}
		{53 __ __ __ __ __ __ __  SetGPRMMD}
		{54 __ __ __ __ __ __ __  SetAMXMD}
		{56 __ __ __ __ __ __ __  SetHL_BTN}
		{6_ __ __ __ __ __ __ __  Set}
		{7_ __ __ __ __ __ __ __  Set}
		{8_ __ __ __ __ __ __ __  SetCLnk}
		{9_ __ __ __ __ __ __ __  SetCLnk}
		{A_ __ __ __ __ __ __ __  CSetCLnk}
		{B_ __ __ __ __ __ __ __  CSetCLnk}
		{C_ __ __ __ __ __ __ __  CSetLnk}
		{D_ __ __ __ __ __ __ __  CSetLnk}
	}

	set ::ced::c_all_commands_abc {
		{Break}
		{CallSS-FP}
		{CallSS-VMGM-menu}
		{CallSS-VMGM-pgc}
		{CallSS-VTSM}
		{CSetCLnk}
		{CSetLnk}
		{Exit}
		{GoTo}
		{JumpSS-FP}
		{JumpSS-VMGM-menu}
		{JumpSS-VMGM-pgc}
		{JumpSS-VTSM}
		{JumpTT}
		{JumpVTS_PTT}
		{JumpVTS_TT}
		{LinkCN}
		{LinkGoUpPGC}
		{LinkNextCell}
		{LinkNextPG}
		{LinkNextPGC}
		{LinkNoLink}
		{LinkPGCN}
		{LinkPGN}
		{LinkPrevCell}
		{LinkPrevPG}
		{LinkPrevPGC}
		{LinkPTT}
		{LinkTailPGC}
		{LinkTopCell}
		{LinkTopPG}
		{LinkTopPGC}
		{NOP}
		{RSM}
		{Set}
		{SetAMXMD}
		{SetCLnk}
		{SetGPRMMD}
		{SetHL_BTN}
		{SetNVTMR}
		{SetSTN}
		{SetTmpPML}
	}

	set ::ced::link_commands_list {
		{1 0  LinkNoLink}
		{1 1  LinkTopCell}
		{1 2  LinkNextCell}
		{1 3  LinkPrevCell}
		{1 5  LinkTopPG}
		{1 6  LinkNextPG}
		{1 7  LinkPrevPG}
		{1 9  LinkTopPGC}
		{1 10 LinkNextPGC}
		{1 11 LinkPrevPGC}
		{1 12 LinkGoUpPGC}
		{1 13 LinkTailPGC}
		{1 16 RSM}
		{- -  -}
		{4 0  LinkPGCN}
		{5 0  LinkPTT}
		{6 0  LinkPGN}
		{7 0  LinkCN}
	}

	set ::ced::c_link_1_descriptions {
		"Don't link"
		"Restart the current cell"
		"Start the next cell in the current PGC"
		"Start the previous cell in the current PGC"
		-
		"Restart the current program"
		"Start the next program in the current PGC"
		"Start the previous program in the current PGC"
		-
		"Restart the current PGC"
		"Start the PGC pointed to by the 'next' link"
		"Start the PGC pointed to by the 'prev' link"
		"Start the PGC pointed to by the 'go up' link"
		"Skip to the end of the current PGC, executes the post-commands"
		-
		-
		"Resume at the location where playback was suspended by a CallSS or a menu call"
	}

	set ::ced::c_compare_operators_list {
		{2 EQ ==}
		{3 NE !=}
		{7 LT <}
		{6 LE <=}
		{5 GT >}
		{4 GE >=}
		{1 BC &}
	}

	set ::ced::c_menus_names_list {
		{3 "Root"}
		{4 "Sub-picture"}
		{5 "Audio"}
		{6 "Angle"}
		{7 "Chapter"}
	}

	set ::ced::c_SPRMlist {
		{0  Preferred menu language}
		{1  Audio stream number (ASTN)}
		{2  Sub-picture stream number (SPSTN)}
		{3  Angle number (AGLN)}
		{4  Title number in volume (TITLE)}
		{5  Title number in VTS (VTS_TTN)}
		{6  PGC number (TT_PGCN)}
		{7  Chapter (PTT) number (PTTN)}
		{8  Highlighted button number (HL_BTNN)}
		{9  Navigation timer (NVTMR)}
		{10 PGC jump for nav timer (NV_PGCN)}
		{11 Karaoke audio mixing mode (AMXMD)}
		{12 Parental management country code (CC_PLT)}
		{13 Parental level (PLT)}
		{14 Video preference and current mode}
		{15 Player audio capabilities}
		{16 Preferred audio language}
		{17 Preferred audio language extension}
		{18 Preferred sub-picture language}
		{19 Preferred sub-picture language extension}
		{20 Player region code}
		{21 reserved}
		{22 reserved}
		{23 reserved for extended playback mode}
	}

	set ::ced::c_Setoperators {
		{1  {MOV    =}   1 1}
		{-}
		{3  {ADD    +=}  0 1}
		{4  {SUB    -=}  0 1}
		{5  {MUL    *=}  0 1}
		{6  {DIV     /=} 0 1}
		{7  {MOD   %=}   0 1}
		{8  {RND    ?=}  0 1}
		{-}
		{9  {AND    &=}  1 1}
		{10 {OR      |=} 1 1}
		{11 {XOR    ^=}  1 1}
		{-}
		{2  {SWAP <->}   0 0}
	}
}

#############################################################################
## Init globals storing the params of the DVD commands and used by the interface

proc init_cmds_globals {} {
	#dprint 3 "> init_cmds_globals {}"
	# misc variables set in code:

	# command in hex form like "00 00 00 00 00 00 00 00"
	# ::hexcommand

	# Used for Undo and highlighting of changes after validation
	set ::ced::old_hexcommand ""

	# aliases for widgets
	# ::ced::widget(*)

	# copy of the last disassembled hex command used by hexcmd_disasm
	# to skip the disassembly if the command have not changed.
	set ::ced::old_hex_for_disassemble ""

	# Link and If common default
	# true if the IF and LINK parts are presents and incompatibles
	# set ::ced::c_iflink(link_exists) false ;# set in code

	# IF defaults
	set ::ced::c_if(check) false
	set ::ced::c_if(sprm1) 0
	set ::ced::c_if(gprm1) 0
	set ::ced::c_if(operatornum) 2
	set ::ced::c_if(operator) "=="
	set ::ced::c_if(radioOp2) 1
	set ::ced::c_if(gprm2) 0
	set ::ced::c_if(sprm2) 0
	set ::ced::c_if(value2) 0

	# Goto default
	set ::ced::c_goto(linenum) 3

	# SetTmpPNL default
	set ::ced::c_TmpPNL(level) 1
	set ::ced::c_settmppml(none) false

	# Link commands defaults
	set ::ced::c_link(check) false	;# used with Set* commands
	set ::ced::c_link(k) 1
	set ::ced::c_link(ll) 1
	#set ::ced::c_link(text) "link description"	;# set in code
	set ::ced::c_link(highlightbutton) 0
	set ::ced::c_link(highlightmode) false
	set ::ced::c_link(pgcn) 1
	set ::ced::c_link(pttn) 1
	set ::ced::c_link(pgn) 1
	set ::ced::c_link(cell) 1

	# Jump/Call defaults
	set ::ced::c_jump(title) 1
	set ::ced::c_jump(ptt) 1
	set ::ced::c_jump(vts) 1
	set ::ced::c_jump(pgc) 1
	# Jump/CallSS-VTSM option
	set ::ced::c_jump(menulabel) "Root menu"
	set ::ced::c_jump(menunumber) 3
	# ResumeCell Call defaults
	set ::ced::c_call(resume_mode) true
	set ::ced::c_call(resume_cell) 1

	# Set* defaults
	set ::ced::c_set(destination)    0
	set ::ced::c_set(operator_text)  "=   (MOV)"
	set ::ced::c_set(operator_value)  1
	set ::ced::c_set(sourcemode) 1
	set ::ced::c_set(gprm)  0
	set ::ced::c_set(sprm)  0
	set ::ced::c_set(value) 0

	# SetSTN defaults
	set ::ced::c_setstn(audio_mode)   0
	set ::ced::c_setstn(subpic_mode)  0
	set ::ced::c_setstn(angle_mode)   0
	set ::ced::c_setstn(audio_value)  0
	set ::ced::c_setstn(subpic_value) 0
	set ::ced::c_setstn(angle_value)  1
	set ::ced::c_setstn(audio_gprm)   0
	set ::ced::c_setstn(subpic_gprm)  0
	set ::ced::c_setstn(angle_gprm)   0
	set ::ced::c_setstn(subpic_onoff) true

	# SetAMXMD defaults
	set ::ced::c_setamxmd_mix(20) 0
	set ::ced::c_setamxmd_mix(30) 0
	set ::ced::c_setamxmd_mix(40) 0
	set ::ced::c_setamxmd_mix(21) 0
	set ::ced::c_setamxmd_mix(31) 0
	set ::ced::c_setamxmd_mix(41) 0

	# SetNVTMR defaults
	set ::ced::c_setnvtmr(pgcmode) true
	set ::ced::c_setnvtmr(pgc) 1

	# SetGPRMMD defaults
	set ::ced::c_setgprmmd(destination) 0
	set ::ced::c_setgprmmd(mode) 8

	# SetCLnk & CSetCLnk defaults
	# set ::ced::c_setclnk(cmdvalue) 0	;# set in code
}

#############################################################################
## Initialization

init_cmds_global_lists
init_cmds_globals

}	;# end namespace eval ::ced